#!/bin/bash
#
#  libflame
#  An object-based infrastructure for developing high-performance
#  dense linear algebra libraries.
#
#  Copyright (C) 2008, The University of Texas
#
#  libflame is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of
#  the License, or (at your option) any later version.
#
#  libflame is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with libflame; if you did not receive a copy, see
#  http://www.gnu.org/licenses/.
#
#  For more information, please contact us at flame@cs.utexas.edu or
#  send mail to:
#
#  Field G. Van Zee and/or
#  Robert A. van de Geijn
#  The University of Texas at Austin
#  Department of Computer Sciences
#  1 University Station C0500
#  Austin TX 78712
#

#
# export-for-release.sh
#
# Field G. Van Zee
#

print_usage()
{
	local script_name
	
	# Get the script name
	script_name=${0##*/}
	
	# Echo usage info
	echo " "
	echo " "$script_name
	echo " "
	echo " Field G. Van Zee"
	echo " "
	echo " Export a copy of the FLAME source tree and package it for release"
	echo " under the version vers_str. Note:"
	echo "  o The host argument must be a UTCS host since the UTCS network"
	echo "    filesystem is where the repository resides."
	echo " "
	echo " Usage:"
	echo "   ${script_name} [options] -v vstr host"
	echo " "
	echo " The following arguments are accepted/required:"
	echo " "
	echo "   -r n        revision"
	echo "                 Grab revision n from the repository. If no revision is"
	echo "                 given, the HEAD is exported."
	echo "   -v vstr     version"
	echo "                 Use version vstr to name the revision being exported for"
	echo "                 release. Some examples are \"0.9\" and \"1.0.3b2\"".
	echo " "
	
	# Exit with non-zero exit status
	exit 1
}


main()
{
	repos_path=/u/field/repos/flame
	toplevel_dir=trunk
	release_prefix=libflame
	
	# Process our command line options.
	while getopts ":r:v:" opt; do
		case $opt in
			r  ) revision_str=$OPTARG ;;
			v  ) version_str=$OPTARG ;;
			\? ) print_usage
		esac
	done
	shift $(($OPTIND - 1))
	
	# Check the number of arguments after command line option processing.
	if [ $# != "1" ]; then
		print_usage
	fi
	
	# Set the default revision if necessary.
	if [ "$revision_str" == "" ]; then
		revision_str="HEAD"
	fi
	
	# Make sure that verboseness level is valid
	if [ "$version_str" == "" ]; then
		echo "$0: You must provide a version string with -v."
		exit 1
	fi
	
	# Extract our arguments.
	host=$1

	svn_url="svn+ssh://${host}${repos_path}/${toplevel_dir}"
	
	# Construct release directory name.
	release_dir_name="${release_prefix}-${version_str}"
	
	# Execute the export.
	svn export -r ${revision_str} ${svn_url} ${release_dir_name}
	
	# Tar and gzip.
	tar czf ${release_dir_name}.tar.gz ${release_dir_name}
	
	# Exit peacefully
	return 0
}


# The script's main entry point, passing all parameters given.
main "$@"
