/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"


FLA_Datatype FLASH_Obj_datatype( FLA_Obj H )
{
	return FLA_Obj_datatype( H );
}


int FLASH_Obj_depth( FLA_Obj H )
{
	FLA_Elemtype elemtype;
	FLA_Obj*     buffer_H;
	int          depth = 0;

	// Recurse through the hierarchy to the first leaf node. We initialize
	// the recursion here:
	elemtype = FLA_Obj_elemtype( H );
	buffer_H  = ( FLA_Obj* ) FLA_Obj_buffer( H );

	while ( elemtype == FLA_MATRIX )
	{
		++depth;

		// Get the element type of the top-leftmost underlying object. Also,
		// get a pointer to the first element of the top-leftmost object and
		// assume that it is of type FLA_Obj* in case elemtype is once again
		// FLA_MATRIX.
		elemtype = FLA_Obj_elemtype( buffer_H[0] );
		buffer_H = ( FLA_Obj * ) FLA_Obj_buffer( buffer_H[0] );
	}

	// At this point, the value of depth represents the depth of the matrix
	// hierarchy.
	return depth;
}


int FLASH_Obj_blocksizes( FLA_Obj H, int* blocksizes )
{
	FLA_Elemtype elemtype;
	FLA_Obj*     buffer_H;
	int          depth = 0;

	// Recurse through the hierarchy to the first leaf node. We initialize
	// the recursion here:
	elemtype = FLA_Obj_elemtype( H );
	buffer_H = ( FLA_Obj* ) FLA_Obj_buffer( H );

	while ( elemtype == FLA_MATRIX )
	{
		blocksizes[depth] = FLA_Obj_length( buffer_H[0] );
		++depth;

		// Get the element type of the top-leftmost underlying object. Also,
		// get a pointer to the first element of the top-leftmost object and
		// assume that it is of type FLA_Obj* in case elemtype is once again
		// FLA_MATRIX.
		elemtype = FLA_Obj_elemtype( buffer_H[0] );
		buffer_H = ( FLA_Obj * ) FLA_Obj_buffer( buffer_H[0] );
	}

	// At this point, the first depth elements of blocksizes have been filled
	// with the blocksizes of H's various hierarchical levels. Return the
	// matrix depth as a confirmation of how many blocksizes were found.
	return depth;
}


int FLASH_Obj_scalar_length( FLA_Obj H )
{
	FLA_Obj  HT,              H0,
	         HB,              H1,
	                          H2;
	FLA_Obj* H1p;

	int b = 0;

	if ( FLA_Obj_elemtype( H ) == FLA_SCALAR )
		return FLA_Obj_length( H );

	FLA_Part_2x1( H,    &HT,
	                    &HB,            0, FLA_TOP );

	while ( FLA_Obj_length( HT ) < FLA_Obj_length( H ) )
	{
		FLA_Repart_2x1_to_3x1( HT,                &H0,
		                    /* ** */            /* ** */
		                                          &H1,
		                       HB,                &H2,        1, FLA_BOTTOM );

		/*------------------------------------------------------------*/

		H1p = FLASH_OBJ_PTR_AT( H1 );
		b += H1p->base->m_inner;

		/*------------------------------------------------------------*/

		FLA_Cont_with_3x1_to_2x1( &HT,                H0,
		                                              H1,
		                        /* ** */           /* ** */
		                          &HB,                H2,     FLA_TOP );
  }

  return b;
}


int FLASH_Obj_scalar_width( FLA_Obj H )
{
	FLA_Obj  HL,    HR,       H0,  H1,  H2;
	FLA_Obj* H1p;

	int b = 0;

	if ( FLA_Obj_elemtype( H ) == FLA_SCALAR )
		return FLA_Obj_width( H );

	FLA_Part_1x2( H,    &HL,  &HR,      0, FLA_LEFT );

	while ( FLA_Obj_width( HL ) < FLA_Obj_width( H ) )
	{
		FLA_Repart_1x2_to_1x3( HL,  /**/ HR,        &H0, /**/ &H1, &H2,
		                       1, FLA_RIGHT );

		/*------------------------------------------------------------*/

		H1p = FLASH_OBJ_PTR_AT( H1 );
		b += H1p->base->n_inner;

		/*------------------------------------------------------------*/

		FLA_Cont_with_1x3_to_1x2( &HL,  /**/ &HR,        H0, H1, /**/ H2,
		                          FLA_LEFT );
	}

	return b;
}


FLA_Error FLASH_Obj_create( FLA_Datatype datatype, int m, int n, int depth, int* blocksizes, FLA_Obj* H )
{
	int       i;
	FLA_Obj   flat_matrix;
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	FLA_Error e_val;

	e_val = FLA_Check_valid_datatype( datatype );
	FLA_Check_error_code( e_val );

	e_val = FLA_Check_null_pointer( blocksizes );
	FLA_Check_error_code( e_val );

	e_val = FLA_Check_null_pointer( H );
	FLA_Check_error_code( e_val );

	// A value of depth < 0 should cause an error.

	// Values of m < 1, n < 1 should cause an error. (or < 0?)

	// First depth entries in blocksize should be checked; values < 1 should cause error.
#endif
	
	if ( depth == 0 )
	{
		// Base case: create a single contiguous matrix block.
		FLA_Obj_create( datatype, m, n, H );
	}
	else
	{
		// We need a temporary array the same length as the blocksizes array.
		int* elem_sizes  = ( int * ) FLA_malloc( depth * sizeof( int ) );
		int* depth_sizes = ( int * ) FLA_malloc( depth * sizeof( int ) );
		int* m_offsets = ( int * ) FLA_malloc( depth * sizeof( int ) );
		int* n_offsets = ( int * ) FLA_malloc( depth * sizeof( int ) );
		
		// Compute the number of elements contained along one side of a block,
		// for each blocking level. This is just the product of all blocksizes
		// "internal" to and including the current blocking level.
		elem_sizes[depth-1] = blocksizes[depth-1];
		for( i = depth - 1; i > 0; --i )
		{
			elem_sizes[i-1] = elem_sizes[i] * blocksizes[i-1];
		}

		depth_sizes[depth-1] = 1;
		for ( i = depth - 1; i > 0; --i )
		{
			depth_sizes[i-1] = depth_sizes[i] * blocksizes[i-1];
		}

		for ( i = 0; i < depth; i++ )
		{
			m_offsets[i] = 0;
			n_offsets[i] = 0;
		}
		
		// Create a "flat" matrix object. All leaf-level child objects will refer
		// to various offsets within this object's buffer.
		// Notice that this invocation requests a 1 x m*n flat matrix, whereas the
		// corresponding line in the _create_without_buffer() function requests an
		// m x n matrix. This is simply how Tze originally implemented the
		// functions. There may be a subtle difference between these two
		// methods--one which I can't see at the time of this writing.
		FLA_Obj_create( datatype, 1, m*n, &flat_matrix );
		
		// Recursively create the matrix hierarchy.
		FLASH_Obj_create_hierarchy( datatype, m, n, depth, elem_sizes, flat_matrix, H, 0, depth, depth_sizes, m_offsets, n_offsets );
		
		// Nullify the buffer field within the flat matrix object.
		// We do this to prevent the upcoming FLA_Obj_free() from freeing the memory
		// that was just partitioned up and doled out to the individual leaves
		// of the matrix hierarchy. That is, we want to free the object but not
		// the buffer. Another way to do this would be to call FLA_Obj_free_without_buffer().
		FLA_Obj_attach_buffer( NULL, 1, &flat_matrix );

		// Free the flat_matrix object (but not its buffer).
		FLA_Obj_free_without_buffer( &flat_matrix );
		
		// Free the elem_sizes array.
		FLA_free( elem_sizes );
		FLA_free( depth_sizes );
		FLA_free( m_offsets );
		FLA_free( n_offsets );
	}

	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_create_without_buffer( FLA_Datatype datatype, int m, int n, int depth, int* blocksizes, FLA_Obj* H )
{
	int       i;
	FLA_Obj   flat_matrix;
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	FLA_Error e_val;

	e_val = FLA_Check_valid_datatype( datatype );
	FLA_Check_error_code( e_val );

	// A value of depth < 0 should cause an error.

	// Values of m < 1, n < 1 should cause an error. (or < 0?)

	// First depth entries in blocksize should be checked; values < 1 should cause error.
#endif
	
	if ( depth == 0 )
	{
		// Base case: create a single contiguous matrix block.
		FLA_Obj_create_without_buffer( datatype, m, n, H );
	}
	else
	{
		// We need a temporary array the same length as the blocksizes array.
		int* elem_sizes = ( int * ) FLA_malloc( depth * sizeof( int ) );
		int* depth_sizes = ( int * ) FLA_malloc( depth * sizeof( int ) );
		int* m_offsets = ( int * ) FLA_malloc( depth * sizeof( int ) );
		int* n_offsets = ( int * ) FLA_malloc( depth * sizeof( int ) );

		// Compute the number of elements contained along one side of a block,
		// for each blocking level. This is just the product of all blocksizes
		// "internal" to and including the current blocking level.
		elem_sizes[depth-1] = blocksizes[depth-1];
		for( i = depth - 1; i > 0; --i )
		{
			elem_sizes[i-1] = elem_sizes[i] * blocksizes[i-1];
		}

		depth_sizes[depth-1] = 1;
		for ( i = depth - 1; i > 0; --i )
		{
			depth_sizes[i-1] = depth_sizes[i] * blocksizes[i-1];
		}

		for ( i = 0; i < depth; i++ )
		{
			m_offsets[i] = 0;
			n_offsets[i] = 0;
		}
		
		// Create a "flat" matrix object. All leaf-level child objects will refer
		// to various offsets within this object's buffer.
		// Notice that this invocation requests an m x n matrix, whereas the
		// corresponding line in the _create() function requests a 1 x m*n flat
		// matrix. This is simply how Tze originally implemented the functions.
		// There may be a subtle difference between these two methods--one which
		// I can't see at the time of this writing.
		FLA_Obj_create_without_buffer( datatype, m, n, &flat_matrix );
		
		// Recursively create the matrix hierarchy.
		FLASH_Obj_create_hierarchy( datatype, m, n, depth, elem_sizes, flat_matrix, H, 0, depth, depth_sizes, m_offsets, n_offsets );
		
		// Free the flat_matrix object (but not its buffer).
		FLA_Obj_free_without_buffer( &flat_matrix );
		
		// Free the elem_sizes array.
		FLA_free( elem_sizes );
		FLA_free( depth_sizes );
		FLA_free( m_offsets );
		FLA_free( n_offsets );
	}

	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_create_hierarchy( FLA_Datatype datatype, int m, int n, int depth, int* elem_sizes, FLA_Obj flat_matrix, FLA_Obj* H, int id, int depth_overall, int* depth_sizes, int* m_offsets, int* n_offsets )
{
	int      i, j, b;
	int      next_m, next_n;
	int      num_m, num_n;
	int      m_inner, n_inner;
	int      elem_size_cur;
	FLA_Obj  FL, FR, F0, F1, F2;
	FLA_Obj* buffer_H;

	if ( depth == 0 )
	{
		// If we're asked to create a zero-depth matrix, we interpret that as
		// a request to create leaf-level objects using the remaining portion
		// of the segment of the flat_matrix buffer that was passed in.
		FLA_Obj_create_without_buffer( datatype, m, n, H );
		FLA_Obj_attach_buffer( FLA_Obj_buffer( flat_matrix ), m, H );
#ifdef FLA_ENABLE_SUPERMATRIX
		FLASH_Queue_set_block_size( m * n * FLA_Obj_datatype_size( datatype ) );
#endif
		H->base->id = id;

		for ( i = 0; i < depth_overall; i++ )
		{
			H->base->m_index += m_offsets[i] * depth_sizes[i];
			H->base->n_index += n_offsets[i] * depth_sizes[i];
		}
	}
	else
	{
		// The "current" level's elem_size value. That is, the number of numerical
		// scalar elements along one side of a full block on the current level.
		elem_size_cur = elem_sizes[0];

		// Compute the number of rows and columns in the current hierarchical
		// level of blocking.
		num_m = m / elem_size_cur + ( (m % elem_size_cur) ? 1 : 0 );
		num_n = n / elem_size_cur + ( (n % elem_size_cur) ? 1 : 0 );

		// The total number of scalar elements contained within/below this level
		// of the hierarchy. (The edge cases are handled by the computation of
		// next_m and next_n below, since they are passed in as the new m and n
		// for the next recursive call.)
		m_inner = m;
		n_inner = n;
		
		// Create a matrix whose elements are FLA_Objs for the current level of
		// blocking.
		FLA_Obj_create_ext( datatype, FLA_MATRIX, num_m, num_n, m_inner, n_inner, H );

		if ( depth == depth_overall )
			id = H->base->id;
		else
			H->base->id = id;

		// Grab the buffer from the new hierarchical object. This is an array of
		// FLA_Objs.
		buffer_H = ( FLA_Obj* ) FLA_Obj_buffer( *H );
		
		// Prepare to partition through the flat matrix so we can further allocate
		// segments of it to the various hierarchical sub-matrices. (The second
		// case occurs when the current function is called with a flat_matrix
		// argument that was created without a buffer.)
		if ( FLA_Obj_buffer( flat_matrix ) != NULL )
			FLA_Part_1x2( flat_matrix, &FL, &FR, 0, FLA_LEFT );
		else
			FLA_Obj_create_without_buffer( datatype, 0, 0, &F1 );
		
		for ( j = 0; j < num_n; ++j )
		{
			// Determine the number of elements along the column dimension
			// that will be contained within the submatrix referenced by
			// the (i,j)th FLA_MATRIX element in the current matrix.
			if ( j != num_n-1 || (n % elem_size_cur) == 0 )
				next_n = elem_size_cur;
			else
				next_n = n % elem_size_cur;
			
			n_offsets[depth_overall-depth] = j;
                        
			for ( i = 0; i < num_m; ++i )
			{			
				// Determine the number of elements along the row dimension
				// that will be contained within the submatrix referenced by
				// the (i,j)th FLA_MATRIX element in the current matrix.
				if ( i != num_m-1 || (m % elem_size_cur) == 0 )
					next_m = elem_size_cur;
				else
					next_m = m % elem_size_cur;

				m_offsets[depth_overall-depth] = i;
				
				// Partition the next m*n elements from the flat matrix so we can
				// "attach" them to the hierarchical matrices contained within the
				// (i,j)th FLA_MATRIX object.
				if ( FLA_Obj_buffer( flat_matrix ) != NULL )
				{
					b = min( FLA_Obj_width( FR ), next_m * next_n );
					FLA_Repart_1x2_to_1x3( FL,  /**/ FR,        &F0, /**/ &F1, &F2,
					                       b, FLA_RIGHT );
				}
				
				// Recursively call ourselves, with the appropriate parameters for
				// the next deeper level in the matrix hierarchy.
				FLASH_Obj_create_hierarchy( datatype, next_m, next_n, depth-1, &elem_sizes[1], F1, &buffer_H[j*num_m + i], id, depth_overall, depth_sizes, m_offsets, n_offsets );

				// Continue with the repartitioning.
				if ( FLA_Obj_buffer( flat_matrix ) != NULL )
				{
					FLA_Cont_with_1x3_to_1x2( &FL,  /**/ &FR,        F0, F1, /**/ F2,
					                          FLA_LEFT );
				}
			}
		}

		// Free the temporary flat matrix subpartition object, but only if it was
		// created to begin with. Since it would have been created without a
		// buffer, we must free it in a similar manner.
		if ( FLA_Obj_buffer( flat_matrix ) == NULL )
			FLA_Obj_free_without_buffer( &F1 );
	}
	
	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_create_conf_to( FLA_Trans trans, FLA_Obj H_cur, FLA_Obj* H_new )
{
	FLA_Datatype datatype;
	int          m_H_cur, n_H_cur;
	int          m_H_new, n_H_new;
	int          d_H;
	int*         blocksizes;
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	FLA_Error    e_val;

	e_val = FLA_Check_valid_real_trans( trans );
	FLA_Check_error_code( e_val );
#endif

	// Acquire the numerical datatype, length, width, and depth of the
	// hierarchical matrix object.
	datatype = FLASH_Obj_datatype( H_cur );
	m_H_cur  = FLASH_Obj_scalar_length( H_cur );
	n_H_cur  = FLASH_Obj_scalar_width( H_cur );
	d_H      = FLASH_Obj_depth( H_cur );

	// Allocate a temporary array for the blocksizes, whose length is equal
	// to the object's hierarchical depth.
	blocksizes = ( int * ) FLA_malloc( d_H * sizeof( int ) );

	// Accumulate the blocksizes into a buffer.
	FLASH_Obj_blocksizes( H_cur, blocksizes );

	// Factor in the transposition, if requested.
	if ( trans == FLA_NO_TRANSPOSE )
	{
		m_H_new = m_H_cur;
		n_H_new = n_H_cur;
	}
	else
	{
		m_H_new = n_H_cur;
		n_H_new = m_H_cur;
	}

	// Create the new hierarchical matrix object.
	FLASH_Obj_create( datatype, m_H_new, n_H_new, d_H, blocksizes, H_new ); 

	// Free the temporary blocksize array.
	FLA_free( blocksizes );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_create_hier_conf_to_flat( FLA_Trans trans, FLA_Obj F, int depth, int* blocksizes, FLA_Obj* H )
{
	FLA_Datatype datatype;
	int          m_H, n_H;
	int          m_F, n_F;
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	FLA_Error    e_val;

	e_val = FLA_Check_valid_real_trans( trans );
	FLA_Check_error_code( e_val );

	// A value of depth < 0 should cause an error.

	// First depth entries in blocksize should be checked; values < 1 should cause error.
#endif

	// Acquire the numerical datatype, length, and width of the flat matrix
	// object.
	datatype = FLA_Obj_datatype( F );
	m_F      = FLA_Obj_length( F );
	n_F      = FLA_Obj_width( F );

	// Factor in the transposition, if requested.
	if ( trans == FLA_NO_TRANSPOSE )
	{
		m_H = m_F;
		n_H = n_F;
	}
	else
	{
		m_H = n_F;
		n_H = m_F;
	}

	// Create the hierarchical matrix object.
	FLASH_Obj_create( datatype, m_H, n_H, depth, blocksizes, H );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_create_flat_conf_to_hier( FLA_Trans trans, FLA_Obj H, FLA_Obj* F )
{
	FLA_Datatype datatype;
	int          m_H, n_H;
	int          m_F, n_F;
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	FLA_Error    e_val;

	e_val = FLA_Check_valid_real_trans( trans );
	FLA_Check_error_code( e_val );
#endif

	// Acquire the numerical datatype, length, and width of the
	// hierarchical matrix object.
	datatype = FLASH_Obj_datatype( H );
	m_H      = FLASH_Obj_scalar_length( H );
	n_H      = FLASH_Obj_scalar_width( H );

	// Factor in the transposition, if requested.
	if ( trans == FLA_NO_TRANSPOSE )
	{
		m_F = m_H;
		n_F = n_H;
	}
	else
	{
		m_F = n_H;
		n_F = m_H;
	}

	// Create the flat matrix object.
	FLA_Obj_create( datatype, m_F, n_F, F );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_create_hier_copy_of_flat( FLA_Obj F, int depth, int* blocksizes, FLA_Obj* H )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	// A value of depth < 0 should cause an error.

	// First depth entries in blocksize should be checked; values < 1 should cause error.
#endif

	// Create a hierarchical object conformal to the flat object.
	FLASH_Obj_create_hier_conf_to_flat( FLA_NO_TRANSPOSE, F, depth, blocksizes, H );

	// Initialize the contents of the hierarchical matrix object with the
	// contents of the flat matrix object.
	FLASH_Copy_subobject_to_global( F, 0, 0, *H );
	
	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_create_flat_copy_of_hier( FLA_Obj H, FLA_Obj* F )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING

#endif

	// Create a flat object conformal to the hierarchical object.
	FLASH_Obj_create_flat_conf_to_hier( FLA_NO_TRANSPOSE, H, F );

	// Flatten the hierarchical object's contents into the new flat object.
	FLASH_Copy_global_to_subobject( 0, 0, H, *F );

	return FLA_SUCCESS;
}


void FLASH_Obj_free( FLA_Obj* H )
{
	FLA_Obj* buffer_H;

	// Free the object accordin got whether it contains a hierarchy.
	if ( FLA_Obj_elemtype( *H ) == FLA_MATRIX )
	{
		// Extract a pointer to the data buffer that was parititioned across all
		// leaf-level submatrices.
		buffer_H = ( FLA_Obj * ) FLASH_Obj_extract_buffer( *H );
	
		// Free the data buffer. This works because FLASH_Obj_extract_buffer()
		// returns the starting address of the first element's buffer, which is
		// also the starting address of the entire buffer.
		FLA_free( buffer_H );
	
		// All that remains now is to free the interior of the matrix hierarchy.
		// This includes non-leaf buffers and their corresponding base objects
		// as well as leaf-level base objects.
		FLASH_Obj_free_hierarchy( H );

		// Finally, free the top-level FLASH object itself.
		FLA_Obj_free( H );
	}
	else
	{
		// If the matrix has no hierarchy, treat it like a flat object.
		FLA_Obj_free( H );
	}
}


void FLASH_Obj_free_without_buffer( FLA_Obj* H )
{
	// Free the object according to whether it contains a hierarchy.
	if ( FLA_Obj_elemtype( *H ) == FLA_MATRIX )
	{
		// Skip freeing the numerical data buffer, since the object was
		// presumably created with FLASH_Obj_create_without_buffer().

		// Free the interior of the matrix hierarchy. This includes non-leaf
		// buffers and their corresponding base objects as well as leaf-level
		// base objects.
		FLASH_Obj_free_hierarchy( H );

		// Finally, free the top-level FLASH object itself.
		FLA_Obj_free( H );
	}
	else
	{
		// If the matrix has no hierarchy, treat it like a flat object with
		// no internal data buffer.
		FLA_Obj_free_without_buffer( H );
	}
}


void FLASH_Obj_free_hierarchy( FLA_Obj* H )
{
	int      i,j, m_H, n_H;
	FLA_Obj* buffer_H;
	FLA_Obj* obj_ij_ptr;

	m_H = FLA_Obj_length( *H );
	n_H = FLA_Obj_width( *H );

	// If the element type of H is FLA_SCALAR, then it has no children to free,
	// so we can return. (The base object of H will be freed by the caller,
	// whether that is a recursive instance of the current function,
	// FLASH_Obj_free_hierarchy(), or FLASH_Obj_free() or
	// FLASH_Obj_free_without_buffer()).
	if ( FLA_Obj_elemtype( *H ) == FLA_SCALAR ) return;

	// Since we didn't return, we know that H is of type FLA_MATRIX. Here, we
	// acquire the array of objects contained inside of H.
	buffer_H = ( FLA_Obj* ) FLA_Obj_buffer( *H );

	// For each submatrix in H...
	for ( j = 0; j < n_H; ++j )
	{
		for ( i = 0; i < m_H; ++i )
		{
			obj_ij_ptr = &buffer_H[ j*m_H + i ];

			// Recurse. The object at element (i,j) may also be of element type
			// FLA_MATRIX. If so, we will need to free its children.
			FLASH_Obj_free_hierarchy( obj_ij_ptr );

			// Free the base object at element (i,j). In order to avoid freeing
			// the object's data buffer, which would have already been freed en
			// masse by now if the calling function is FLASH_Obj_free(), we
			// will call FLA_Obj_free_without_buffer().
			FLA_Obj_free_without_buffer( obj_ij_ptr );
		}
	}
}


void* FLASH_Obj_extract_buffer( FLA_Obj H )
{
	FLA_Elemtype elemtype;
	FLA_Obj*     buffer_H;

	// Recurse through the hierarchy to the first leaf node to gain access
	// to the address of the actual numerical data buffer (ie: the "flat"
	// matrix object used in FLASH_Obj_create()). We initialize the search
	// here:
	elemtype = FLA_Obj_elemtype( H );
	buffer_H  = ( FLA_Obj* ) FLA_Obj_buffer( H );

	while ( elemtype == FLA_MATRIX )
	{
		elemtype = FLA_Obj_elemtype( buffer_H[0] );
		buffer_H  = ( FLA_Obj * ) FLA_Obj_buffer( buffer_H[0] );
	}

	// At this point, the value in buffer_H is a pointer to the array that
	// holds the numerical data.
	return ( void* ) buffer_H;
}


FLA_Error FLASH_Obj_flatten( FLA_Obj H, FLA_Obj F )
{
	FLASH_Copy_global_to_subobject( 0, 0, H, F );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_hierarchify( FLA_Obj F, FLA_Obj H )
{
	FLASH_Copy_subobject_to_global( F, 0, 0, H );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_show( char* header, FLA_Obj H, char* elem_format, char* footer )
{
	FLA_Obj F;
	
	// Create a temporary flat object conformal to and copied from the
	// hierarchical object.
	FLASH_Obj_create_flat_copy_of_hier( H, &F );

	// We can now display the temporary matrix in a straightforward manner.
	FLA_Obj_show( header, F, elem_format, footer );

	// Free the temporary flat matrix.
	FLA_Obj_free( &F );
	
	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_attach_buffer( void *buffer, int ldim, FLA_Obj* H )
{
	FLA_Obj      flat_matrix;
	int          m_H, n_H;
	FLA_Datatype datatype;

	// Extract the overall scalar dimensions of the hierarchical matrix and
	// its base numerical datatype. (These fields will be set even if it has
	// a NULL buffer, which presumably it does given that this function was
	// just invoked.)
	m_H      = FLASH_Obj_scalar_length( *H );
	n_H      = FLASH_Obj_scalar_width( *H );
	datatype = FLASH_Obj_datatype( *H );

	// Create a temporary conventional object and attach the given buffer.
	// Segments of this buffer will be partitioned out to the various
	// leaf-level matrices of the hierarchical matrix H.
	FLA_Obj_create_without_buffer( datatype, m_H, n_H, &flat_matrix );
	FLA_Obj_attach_buffer( buffer, ldim, &flat_matrix );

	// Recurse through the hierarchical matrix, assigning segments of
	// flat_matrix to the various leaf-level matrices, similar to what
	// we would do if we were creating the object outright.
	FLASH_Obj_attach_buffer_hierarchy( flat_matrix, H );

	// Free the object (but don't free the buffer!).
	FLA_Obj_free_without_buffer( &flat_matrix );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Obj_attach_buffer_hierarchy( FLA_Obj F, FLA_Obj* H )
{
	FLA_Obj FL,    FR,       F0,  F1,  F2;

	FLA_Obj HL,    HR,       H0,  H1,  H2;

	FLA_Obj F1T,              F01,
	        F1B,              F11,
	                          F21;

	FLA_Obj H1T,              H01,
	        H1B,              H11,
	                          H21;

	FLA_Obj* H11p;

	int b_m, b_n;

	if ( FLA_Obj_elemtype( *H ) == FLA_SCALAR )
	{
		// If we've recursed down to a leaf node, then we can simply attach the
		// matrix buffer to the current leaf-level submatrix.
		FLA_Obj_attach_buffer( FLA_Obj_buffer( F ), FLA_Obj_ldim( F ), H );
	}
	else
	{
		FLA_Part_1x2( *H,    &HL,  &HR,      0, FLA_LEFT );

		FLA_Part_1x2(  F,    &FL,  &FR,      0, FLA_LEFT );

		while ( FLA_Obj_width( HL ) < FLA_Obj_width( *H ) )
		{

			FLA_Repart_1x2_to_1x3( HL,  /**/ HR,        &H0, /**/ &H1, &H2,
			                       1, FLA_RIGHT );

			b_n = FLASH_Obj_scalar_width( H1 );

			FLA_Repart_1x2_to_1x3( FL,  /**/ FR,        &F0, /**/ &F1, &F2,
			                       b_n, FLA_RIGHT );

			/*------------------------------------------------------------*/

			FLA_Part_2x1( H1,    &H1T,
			                     &H1B,            0, FLA_TOP );

			FLA_Part_2x1( F1,    &F1T,
			                     &F1B,            0, FLA_TOP );

			while ( FLA_Obj_length( H1T ) < FLA_Obj_length( H1 ) )
			{

				FLA_Repart_2x1_to_3x1( H1T,               &H01,
				                    /* ** */            /* ** */
				                                          &H11,
				                       H1B,               &H21,        1, FLA_BOTTOM );

				b_m = FLASH_Obj_scalar_length( H11 );

				FLA_Repart_2x1_to_3x1( F1T,               &F01,
				                    /* ** */            /* ** */
				                                          &F11,
				                       F1B,               &F21,      b_m, FLA_BOTTOM );

				/*------------------------------------------------------------*/

				H11p = ( FLA_Obj* ) FLA_Obj_buffer( H11 );

				FLASH_Obj_attach_buffer_hierarchy( F11, H11p );

				/*------------------------------------------------------------*/

				FLA_Cont_with_3x1_to_2x1( &H1T,               H01,
				                                              H11,
				                        /* ** */           /* ** */
				                          &H1B,               H21,     FLA_TOP );

				FLA_Cont_with_3x1_to_2x1( &F1T,               F01,
				                                              F11,
				                        /* ** */           /* ** */
				                          &F1B,               F21,     FLA_TOP );
			}

			/*------------------------------------------------------------*/

			FLA_Cont_with_1x3_to_1x2( &HL,  /**/ &HR,        H0, H1, /**/ H2,
			                          FLA_LEFT );

			FLA_Cont_with_1x3_to_1x2( &FL,  /**/ &FR,        F0, F1, /**/ F2,
			                          FLA_LEFT );

		}
	}

	return FLA_SUCCESS;
}


FLA_Error FLASH_Axpy( FLA_Obj alpha, FLA_Obj X, FLA_Obj Y )
{
	FLA_Obj X_flat;

	// Create a temporary flat copy of the source hierarchicial object X.
	FLASH_Obj_create_flat_copy_of_hier( X, &X_flat );
	
	// Axpy the contents of the temporary flat object to the hierarchical
	// destination object.
	FLASH_Axpy_subobject_to_global( alpha, X_flat, 0, 0, Y );
	
	// Free the temporary flat copy of the source object X.
	FLASH_Obj_free( &X_flat );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Axpy_submatrix_to_global( FLA_Obj alpha, int m, int n, void* buffer, int ldim, int i, int j, FLA_Obj H )
{
	FLA_Obj      flat_matrix;
	FLA_Datatype datatype;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	FLA_Error e_val;

	e_val = FLA_Check_if_scalar( alpha );
	FLA_Check_error_code( e_val );

	e_val = FLA_Check_consistent_object_datatype( alpha, H );
	FLA_Check_error_code( e_val );

	e_val = FLA_Check_submatrix_dims_and_offset( m, n, i, j, H );
	FLA_Check_error_code( e_val );
#endif

	// Acquire the datatype from the hierarchical matrix object.
	datatype = FLASH_Obj_datatype( H );

	// Create a temporary conventional matrix object of the requested datatype
	// and dimensions and attach the given buffer containing the incoming data.
	FLA_Obj_create_without_buffer( datatype, m, n, &flat_matrix );
	FLA_Obj_attach_buffer( buffer, ldim, &flat_matrix );

	// Recurse through H, adding in the corresponding elements of flat_matrix,
	// starting at the (i,j) element offset.
	FLASH_Axpy_hierarchy( FLA_FLAT_TO_HIER, alpha, flat_matrix, i, j, &H );

	// Free the object (but don't free the buffer!).
	FLA_Obj_free_without_buffer( &flat_matrix );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Axpy_global_to_submatrix( FLA_Obj alpha, int i, int j, FLA_Obj H, int m, int n, void* buffer, int ldim )
{
	FLA_Obj      flat_matrix;
	FLA_Datatype datatype;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	FLA_Error e_val;

	e_val = FLA_Check_if_scalar( alpha );
	FLA_Check_error_code( e_val );

	e_val = FLA_Check_consistent_object_datatype( alpha, H );
	FLA_Check_error_code( e_val );

	e_val = FLA_Check_submatrix_dims_and_offset( m, n, i, j, H );
	FLA_Check_error_code( e_val );
#endif

	// Acquire the datatype from the hierarchical matrix object.
	datatype = FLASH_Obj_datatype( H );

	// Create a temporary conventional matrix object of the requested datatype
	// and dimensions and attach the given buffer containing the incoming data.
	FLA_Obj_create_without_buffer( datatype, m, n, &flat_matrix );
	FLA_Obj_attach_buffer( buffer, ldim, &flat_matrix );

	// Recurse through H, adding in the corresponding elements of flat_matrix,
	// starting at the (i,j) element offset.
	FLASH_Axpy_hierarchy( FLA_HIER_TO_FLAT, alpha, flat_matrix, i, j, &H );

	// Free the object (but don't free the buffer!).
	FLA_Obj_free_without_buffer( &flat_matrix );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Axpy_subobject_to_global( FLA_Obj alpha, FLA_Obj F, int i, int j, FLA_Obj H )
{
	FLASH_Axpy_hierarchy( FLA_FLAT_TO_HIER, alpha, F, i, j, &H );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Axpy_global_to_subobject( FLA_Obj alpha, int i, int j, FLA_Obj H, FLA_Obj F )
{
	FLASH_Axpy_hierarchy( FLA_HIER_TO_FLAT, alpha, F, i, j, &H );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Axpy_hierarchy( int direction, FLA_Obj alpha, FLA_Obj F, int i, int j, FLA_Obj* H )
{
	FLA_Error e_val;
	FLASH_Obj_queue queue;

	queue.n_tasks = 0;
	queue.head = NULL;
	queue.tail = NULL;

	e_val = FLASH_Axpy_hierarchy_r( direction, alpha, F, i, j, H, &queue );

#ifdef FLA_ENABLE_MULTITHREADING
	FLASH_Obj_exec( ( void* ) FLASH_Axpy_hierarchy, &queue );
#endif
	return e_val;
}


FLA_Error FLASH_Axpy_hierarchy_r( int direction, FLA_Obj alpha, FLA_Obj F, int i, int j, FLA_Obj* H, FLASH_Obj_queue* queue )
{
	int      m_scal, n_scal;
	int      row_offset, col_offset;
	int      b_m, b_n;
	FLA_Obj* buffer;
	FLA_Obj* h11p;

	FLA_Obj HTL,   HTR,    H00, H01, H02,
	        HBL,   HBR,    H10, H11, H12,
	                       H20, H21, H22;

	FLA_Obj F1T,             F01,
	        F1B,             F11,
	                         F21;

	FLA_Obj hT,              h01,
	        hB,              h11,
	                         h21;

	FLA_Obj FL,    FR,       F0,  F1,  F2;

	FLA_Obj HBR_L,  HBR_R,     h0,  h1,  h2;

	// Once we get down to a submatrix whose elements are scalars, we are down
	// to our base case.
	if ( FLA_Obj_elemtype( *H ) == FLA_SCALAR )
	{
		int m_flat, n_flat;

		// Grab the length and width of the conventional matrix argument.
		m_flat = FLA_Obj_length( F );
		n_flat = FLA_Obj_width( F );

		// Advance to the (i,j) offset of the hierarchical matrix.
		FLA_Part_2x2( *H,   &HTL, &HTR,
		                    &HBL, &HBR,     i, j, FLA_TL );

		// Repartition the destination submatrix, H, such that H11 is conformal
		// to dimensions of the flat conventional matrix F.
		FLA_Repart_2x2_to_3x3( HTL, /**/ HTR,       &H00, /**/ &H01, &H02,
		                    /* ************* */   /* ******************** */
		                                            &H10, /**/ &H11, &H12,
		                       HBL, /**/ HBR,       &H20, /**/ &H21, &H22,
		                       m_flat, n_flat, FLA_BR );

#ifdef FLA_ENABLE_MULTITHREADING
		if ( FLASH_Queue_get_num_threads() > 1 )
		{
			FLASH_Obj_push( direction, alpha, F, H11, queue );
		}	
		else
#endif
		{
			// Depending on which top-level function invoked us, we either add the
			// the source data in the flat matrix to the leaf-level submatrix of the
			// hierarchical matrix, or add the data in the hierarchical submatrix to
			// the appropriate place in the flat matrix.
			if      ( direction == FLA_FLAT_TO_HIER )
			{
				FLA_Axpy( alpha, F, H11 );
			}
			else if ( direction == FLA_HIER_TO_FLAT )
			{
				FLA_Axpy( alpha, H11, F );
			}
		}

		// Return, since we're done for this leaf.
		return FLA_SUCCESS;
	}

	// If the above conditional did not hold, then we proceed.

	// Grab the buffer of H, whose elemtype we know must be FLA_MATRIX. The
	// elements of this buffer are objects.
	buffer = ( FLA_Obj* ) FLA_Obj_buffer( *H );

	// Extract the overall scalar dimensions of the first (top-left) matrix
	// referenced in the hierarchical matrix's buffer.
	m_scal = FLASH_Obj_scalar_length( buffer[0] );
	n_scal = FLASH_Obj_scalar_width( buffer[0] );

	// The i and j offsets given to the current function are relative to the
	// entire matrix, which is represented by H. Since the elemtype of H is
	// FLA_MATRIX, we now compute the number of submatrix elements we will skip
	// in both the m and n dimensions to advance ourselves to the proper
	// hierarchical submatrix within H, the one that contains the destination
	// for the data (or at least the top-left most portion of the data) in the
	// raw buffer matrix. Notice that integer division is performed here.
	b_m = i / m_scal;
	b_n = j / n_scal;

	// We use the indices computed above to advance through H such that the
	// (i,j) element represented by the (i,j) values passed to the current
	// function is situated at the top-left of HBR.
	FLA_Part_2x2( *H,   &HTL, &HTR,
	                    &HBL, &HBR,     b_m, b_n, FLA_TL );

	// Prepare to march laterally through HBR.
	FLA_Part_1x2( HBR,     &HBR_L,  &HBR_R,    0, FLA_LEFT);

	// Prepare to march laterally through the conventional flat matrix.
	FLA_Part_1x2( F,       &FL,  &FR,          0, FLA_LEFT );

	// Iterate laterally through F.
	while ( FLA_Obj_width( FL ) < FLA_Obj_width( F ) )
	{
		// Handle the initial edge case and "interior" cases differently.
		if ( FLA_Obj_width( FL ) == 0 )
		{
			// Compute the width of the matrix that we will recurse into.
			// Since this is the first iteration of the while loop, this
			// blocksize may be smaller than subsequent full blocks if the
			// j offset passed to this routine is non-zero.
			b_n = n_scal - (j % n_scal);

			// Reduce the blocksize if the width of F is smaller than the
			// value computed above.
			b_n = min( FLA_Obj_width( FR ), b_n );

			// If the j offset is non-zero, we must modulo it to the
			// hierarchical blocksize for the element submatrices of H
			// in preparation for our recursion.
			col_offset = j % n_scal;
		}
		else
		{
			// The width of the matrix into which we will recursve may be
			// assumed to be the same as the width of the element submatrices
			// of H. 
			b_n = n_scal;

			// Reduce the blocksize if the width of F is smaller than the
			// element submatrices.
			b_n = min( FLA_Obj_width( FR ), b_n );

			// The column offset will be zero because we will be starting from
			// the left edge of the submatrix.
			col_offset = 0;
		}

		// Expose one column of hierarchical submatrices within HBR.
		FLA_Repart_1x2_to_1x3( HBR_L,  /**/ HBR_R,        &h0, /**/ &h1, &h2,
		                       1, FLA_RIGHT );

		// Expose one panel of F of width b_n, computed above.
		FLA_Repart_1x2_to_1x3( FL,  /**/ FR,        &F0, /**/ &F1, &F2,
		                       b_n, FLA_RIGHT );

		/*------------------------------------------------------------*/

		// Prepare to march vertically through h1 (the first submatrix column
		// of HBR).
		FLA_Part_2x1( h1,    &hT,
		                     &hB,            0, FLA_TOP );

		// Prepare to march vertically through F.
		FLA_Part_2x1( F1,    &F1T,
		                     &F1B,           0, FLA_TOP );

		// Iterate vertically through the current column-panel of F, F1.
		while ( FLA_Obj_length( F1T ) < FLA_Obj_length( F1 ) )
		{

			if ( FLA_Obj_length( F1T ) == 0 )
			{
				// Compute the length of the matrix that we will recurse into.
				// Since this is the first iteration of the while loop, this
				// blocksize may be smaller than subsequent full blocks if the
				// i offset passed to this routine is non-zero.
				b_m = m_scal - (i % m_scal);

	 			// Reduce the blocksize if the length of F is smaller than the
				// value computed above.
				b_m = min( FLA_Obj_length( F1B ), b_m );

				// If the i offset is non-zero, we must modulo it to the
				// hierarchical blocksize for the element submatrices of H
				// in preparation for our recursion.
				row_offset = i % m_scal;
			}
			else
			{
				// The length of the matrix into which we will recursve may be
				// assumed to be the same as the length of the element submatrices
				// of H. 
				b_m = m_scal;

				// Reduce the blocksize if the length of the first panel of
				// F is smaller than the the element submatrices.
				b_m = min( FLA_Obj_length( F1B ), b_m );

				// The column offset will be zero because we will be starting from
				// the top edge of the submatrix.
				row_offset = 0;
			}

			// Expose one hierarchical submatrix within the current column of h1.
			FLA_Repart_2x1_to_3x1( hT,                &h01,
			                    /* ** */            /* ** */
			                                          &h11,
			                       hB,                &h21,        1, FLA_BOTTOM );

			// Expose one block of F1 (the current panel of F) of length b_m,
			// computed above.
			FLA_Repart_2x1_to_3x1( F1T,               &F01,
			                    /* ** */            /* ** */
			                                          &F11,
			                       F1B,               &F21,      b_m, FLA_BOTTOM );

			/*------------------------------------------------------------*/

			// This is a pointer to the FLA_Obj element referenced by h11.
			h11p = ( FLA_Obj* ) FLA_Obj_buffer( h11 );

			// Recurse with new (i,j) offset parameters, a narrowed block of
			// F, and a corresponding hierarchical block of H.
			FLASH_Axpy_hierarchy_r( direction, alpha, F11, row_offset, col_offset, h11p, queue );

			/*------------------------------------------------------------*/

			FLA_Cont_with_3x1_to_2x1( &hT,                h01,
			                                              h11,
			                       /* ** */            /* ** */
			                          &hB,                h21,     FLA_TOP );

			FLA_Cont_with_3x1_to_2x1( &F1T,               F01,
			                                              F11,
			                        /* ** */           /* ** */
			                          &F1B,               F21,     FLA_TOP );
		}

		/*------------------------------------------------------------*/

		FLA_Cont_with_1x3_to_1x2( &FL,  /**/ &FR,        F0, F1, /**/ F2,
		                          FLA_LEFT );

		FLA_Cont_with_1x3_to_1x2( &HBR_L,  /**/ &HBR_R,        h0, h1, /**/ h2,
		                          FLA_LEFT );
	}

	return FLA_SUCCESS;

}


FLA_Error FLASH_Copy( FLA_Obj H_src, FLA_Obj H_dest )
{
	FLA_Obj F;

	// Create a temporary flat copy of the source hierarchicial object.
	FLASH_Obj_create_flat_copy_of_hier( H_src, &F );
	
	// Copy the contents of the temporary flat object to the hierarchical
	// destination object.
	FLASH_Copy_subobject_to_global( F, 0, 0, H_dest );
	
	// Free the temporary flat copy of the source object.
	FLASH_Obj_free( &F );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Copy_submatrix_to_global( int m, int n, void* buffer, int ldim, int i, int j, FLA_Obj H )
{
	FLA_Obj      flat_matrix;
	FLA_Datatype datatype;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	FLA_Error e_val;

	e_val = FLA_Check_submatrix_dims_and_offset( m, n, i, j, H );
	FLA_Check_error_code( e_val );
#endif

	// Acquire the datatype from the hierarchical matrix object.
	datatype = FLASH_Obj_datatype( H );

	// Create a temporary conventional matrix object of the requested datatype
	// and dimensions and attach the given buffer containing the incoming data.
	FLA_Obj_create_without_buffer( datatype, m, n, &flat_matrix );
	FLA_Obj_attach_buffer( buffer, ldim, &flat_matrix );

	// Recurse through H, adding in the corresponding elements of flat_matrix,
	// starting at the (i,j) element offset.
	FLASH_Copy_hierarchy( FLA_FLAT_TO_HIER, flat_matrix, i, j, &H );

	// Free the object (but don't free the buffer!).
	FLA_Obj_free_without_buffer( &flat_matrix );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Copy_global_to_submatrix( int i, int j, FLA_Obj H, int m, int n, void* buffer, int ldim )
{
	FLA_Obj      flat_matrix;
	FLA_Datatype datatype;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	FLA_Error e_val;

	e_val = FLA_Check_submatrix_dims_and_offset( m, n, i, j, H );
	FLA_Check_error_code( e_val );
#endif

	// Acquire the datatype from the hierarchical matrix object.
	datatype = FLASH_Obj_datatype( H );

	// Create a temporary conventional matrix object of the requested datatype
	// and dimensions and attach the given buffer containing the incoming data.
	FLA_Obj_create_without_buffer( datatype, m, n, &flat_matrix );
	FLA_Obj_attach_buffer( buffer, ldim, &flat_matrix );

	// Recurse through H, adding in the corresponding elements of flat_matrix,
	// starting at the (i,j) element offset.
	FLASH_Copy_hierarchy( FLA_HIER_TO_FLAT, flat_matrix, i, j, &H );

	// Free the object (but don't free the buffer!).
	FLA_Obj_free_without_buffer( &flat_matrix );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Copy_subobject_to_global( FLA_Obj F, int i, int j, FLA_Obj H )
{
	FLASH_Copy_hierarchy( FLA_FLAT_TO_HIER, F, i, j, &H );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Copy_global_to_subobject( int i, int j, FLA_Obj H, FLA_Obj F )
{
	FLASH_Copy_hierarchy( FLA_HIER_TO_FLAT, F, i, j, &H );

	return FLA_SUCCESS;
}


FLA_Error FLASH_Copy_hierarchy( int direction, FLA_Obj F, int i, int j, FLA_Obj* H )
{
	FLA_Error e_val;
	FLASH_Obj_queue queue;

	queue.n_tasks = 0;
	queue.head = NULL;
	queue.tail = NULL;

	e_val = FLASH_Copy_hierarchy_r( direction, F, i, j, H, &queue );

#ifdef FLA_ENABLE_MULTITHREADING
	FLASH_Obj_exec( ( void* ) FLASH_Copy_hierarchy, &queue );
#endif
	return e_val;
}


FLA_Error FLASH_Copy_hierarchy_r( int direction, FLA_Obj F, int i, int j, FLA_Obj* H, FLASH_Obj_queue* queue )
{
	int      m_scal, n_scal;
	int      row_offset, col_offset;
	int      b_m, b_n;
	FLA_Obj* buffer;
	FLA_Obj* h11p;

	FLA_Obj HTL,   HTR,    H00, H01, H02,
	        HBL,   HBR,    H10, H11, H12,
	                       H20, H21, H22;

	FLA_Obj F1T,             F01,
	        F1B,             F11,
	                         F21;

	FLA_Obj hT,              h01,
	        hB,              h11,
	                         h21;

	FLA_Obj FL,    FR,       F0,  F1,  F2;

	FLA_Obj HBR_L,  HBR_R,     h0,  h1,  h2;

	// Once we get down to a submatrix whose elements are scalars, we are down
	// to our base case.
	if ( FLA_Obj_elemtype( *H ) == FLA_SCALAR )
	{
		int m_flat, n_flat;

		// Grab the length and width of the conventional matrix argument.
		m_flat = FLA_Obj_length( F );
		n_flat = FLA_Obj_width( F );

		// Advance to the (i,j) offset of the hierarchical matrix.
		FLA_Part_2x2( *H,   &HTL, &HTR,
		                    &HBL, &HBR,     i, j, FLA_TL );

		// Repartition the destination submatrix, H, such that H11 is conformal
		// to dimensions of the flat conventional matrix F.
		FLA_Repart_2x2_to_3x3( HTL, /**/ HTR,       &H00, /**/ &H01, &H02,
		                    /* ************* */   /* ******************** */
		                                            &H10, /**/ &H11, &H12,
		                       HBL, /**/ HBR,       &H20, /**/ &H21, &H22,
		                       m_flat, n_flat, FLA_BR );

#ifdef FLA_ENABLE_MULTITHREADING
		if ( FLASH_Queue_get_num_threads() > 1 )
		{
			FLASH_Obj_push( direction, FLA_ONE, F, H11, queue );
		}	
		else
#endif  
		{
			// Depending on which top-level function invoked us, we either add the
			// the source data in the flat matrix to the leaf-level submatrix of the
			// hierarchical matrix, or add the data in the hierarchical submatrix to
			// the appropriate place in the flat matrix.
			if      ( direction == FLA_FLAT_TO_HIER )
			{
				FLA_Copy( F, H11 );
			}
			else if ( direction == FLA_HIER_TO_FLAT )
			{
				FLA_Copy( H11, F );
			}
		}

		// Return, since we're done for this leaf.
		return FLA_SUCCESS;
	}

	// If the above conditional did not hold, then we proceed.

	// Grab the buffer of H, whose elemtype we know must be FLA_MATRIX. The
	// elements of this buffer are objects.
	buffer = ( FLA_Obj* ) FLA_Obj_buffer( *H );

	// Extract the overall scalar dimensions of the first (top-left) matrix
	// referenced in the hierarchical matrix's buffer.
	m_scal = FLASH_Obj_scalar_length( buffer[0] );
	n_scal = FLASH_Obj_scalar_width( buffer[0] );

	// The i and j offsets given to the current function are relative to the
	// entire matrix, which is represented by H. Since the elemtype of H is
	// FLA_MATRIX, we now compute the number of submatrix elements we will skip
	// in both the m and n dimensions to advance ourselves to the proper
	// hierarchical submatrix within H, the one that contains the destination
	// for the data (or at least the top-left most portion of the data) in the
	// raw buffer matrix. Notice that integer division is performed here.
	b_m = i / m_scal;
	b_n = j / n_scal;

	// We use the indices computed above to advance through H such that the
	// (i,j) element represented by the (i,j) values passed to the current
	// function is situated at the top-left of HBR.
	FLA_Part_2x2( *H,   &HTL, &HTR,
	                    &HBL, &HBR,     b_m, b_n, FLA_TL );

	// Prepare to march laterally through HBR.
	FLA_Part_1x2( HBR,     &HBR_L,  &HBR_R,    0, FLA_LEFT);

	// Prepare to march laterally through the conventional flat matrix.
	FLA_Part_1x2( F,       &FL,  &FR,          0, FLA_LEFT );

	// Iterate laterally through F.
	while ( FLA_Obj_width( FL ) < FLA_Obj_width( F ) )
	{
		// Handle the initial edge case and "interior" cases differently.
		if ( FLA_Obj_width( FL ) == 0 )
		{
			// Compute the width of the matrix that we will recurse into.
			// Since this is the first iteration of the while loop, this
			// blocksize may be smaller than subsequent full blocks if the
			// j offset passed to this routine is non-zero.
			b_n = n_scal - (j % n_scal);

			// Reduce the blocksize if the width of F is smaller than the
			// value computed above.
			b_n = min( FLA_Obj_width( FR ), b_n );

			// If the j offset is non-zero, we must modulo it to the
			// hierarchical blocksize for the element submatrices of H
			// in preparation for our recursion.
			col_offset = j % n_scal;
		}
		else
		{
			// The width of the matrix into which we will recursve may be
			// assumed to be the same as the width of the element submatrices
			// of H. 
			b_n = n_scal;

			// Reduce the blocksize if the width of F is smaller than the
			// element submatrices.
			b_n = min( FLA_Obj_width( FR ), b_n );

			// The column offset will be zero because we will be starting from
			// the left edge of the submatrix.
			col_offset = 0;
		}

		// Expose one column of hierarchical submatrices within HBR.
		FLA_Repart_1x2_to_1x3( HBR_L,  /**/ HBR_R,        &h0, /**/ &h1, &h2,
		                       1, FLA_RIGHT );

		// Expose one panel of F of width b_n, computed above.
		FLA_Repart_1x2_to_1x3( FL,  /**/ FR,        &F0, /**/ &F1, &F2,
		                       b_n, FLA_RIGHT );

		/*------------------------------------------------------------*/

		// Prepare to march vertically through h1 (the first submatrix column
		// of HBR).
		FLA_Part_2x1( h1,    &hT,
		                     &hB,            0, FLA_TOP );

		// Prepare to march vertically through F.
		FLA_Part_2x1( F1,    &F1T,
		                     &F1B,           0, FLA_TOP );

		// Iterate vertically through the current column-panel of F, F1.
		while ( FLA_Obj_length( F1T ) < FLA_Obj_length( F1 ) )
		{

			if ( FLA_Obj_length( F1T ) == 0 )
			{
				// Compute the length of the matrix that we will recurse into.
				// Since this is the first iteration of the while loop, this
				// blocksize may be smaller than subsequent full blocks if the
				// i offset passed to this routine is non-zero.
				b_m = m_scal - (i % m_scal);

	 			// Reduce the blocksize if the length of F is smaller than the
				// value computed above.
				b_m = min( FLA_Obj_length( F1B ), b_m );

				// If the i offset is non-zero, we must modulo it to the
				// hierarchical blocksize for the element submatrices of H
				// in preparation for our recursion.
				row_offset = i % m_scal;
			}
			else
			{
				// The length of the matrix into which we will recursve may be
				// assumed to be the same as the length of the element submatrices
				// of H. 
				b_m = m_scal;

				// Reduce the blocksize if the length of the first panel of
				// F is smaller than the the element submatrices.
				b_m = min( FLA_Obj_length( F1B ), b_m );

				// The column offset will be zero because we will be starting from
				// the top edge of the submatrix.
				row_offset = 0;
			}

			// Expose one hierarchical submatrix within the current column of h1.
			FLA_Repart_2x1_to_3x1( hT,                &h01,
			                    /* ** */            /* ** */
			                                          &h11,
			                       hB,                &h21,        1, FLA_BOTTOM );

			// Expose one block of F1 (the current panel of F) of length b_m,
			// computed above.
			FLA_Repart_2x1_to_3x1( F1T,               &F01,
			                    /* ** */            /* ** */
			                                          &F11,
			                       F1B,               &F21,      b_m, FLA_BOTTOM );

			/*------------------------------------------------------------*/

			// This is a pointer to the FLA_Obj element referenced by h11.
			h11p = ( FLA_Obj* ) FLA_Obj_buffer( h11 );

			// Recurse with new (i,j) offset parameters, a narrowed block of
			// F, and a corresponding hierarchical block of H.
			FLASH_Copy_hierarchy_r( direction, F11, row_offset, col_offset, h11p, queue );

			/*------------------------------------------------------------*/

			FLA_Cont_with_3x1_to_2x1( &hT,                h01,
			                                              h11,
			                       /* ** */            /* ** */
			                          &hB,                h21,     FLA_TOP );

			FLA_Cont_with_3x1_to_2x1( &F1T,               F01,
			                                              F11,
			                        /* ** */           /* ** */
			                          &F1B,               F21,     FLA_TOP );
		}

		/*------------------------------------------------------------*/

		FLA_Cont_with_1x3_to_1x2( &FL,  /**/ &FR,        F0, F1, /**/ F2,
		                          FLA_LEFT );

		FLA_Cont_with_1x3_to_1x2( &HBR_L,  /**/ &HBR_R,        h0, h1, /**/ h2,
		                          FLA_LEFT );
	}

	return FLA_SUCCESS;

}


void FLASH_print_struct( FLA_Obj H )
{
	int      i, j, m_H, n_H;
	FLA_Obj* buffer_temp;

	m_H = FLA_Obj_length( H );
	n_H = FLA_Obj_width( H );

	if ( FLA_Obj_elemtype( H ) == FLA_SCALAR )
		FLASH_print_struct_helper( H, 0 );
	else
	{
		for ( j = 0; j < n_H; ++j )
		{
			for ( i = 0; i < m_H; ++i )
			{
				buffer_temp = ( FLA_Obj* ) FLA_Obj_buffer( H );

				FLASH_print_struct_helper( buffer_temp[ j*m_H + i ], 0 );
			}
		}
	}
}


void FLASH_print_struct_helper( FLA_Obj H, int indent )
{
	int      i, j, k, m_H, n_H;
	FLA_Obj* buffer_temp;

	for ( i = 0; i < indent; ++i )
		fprintf( stdout, "  " );

	if ( FLA_Obj_elemtype( H ) == FLA_SCALAR )
	{
		fprintf( stdout, "LEAF (%3d | ldim %3d | %3d x %3d | addr %p)\n",
		                 FLA_Obj_datatype( H ), FLA_Obj_ldim( H ),
		                 FLA_Obj_length( H ), FLA_Obj_width( H ),
		                 FLA_Obj_buffer( H ) );
		fflush( stdout );
	}
	else
	{
		m_H = FLA_Obj_length( H );
		n_H = FLA_Obj_width( H );
		
		fprintf( stdout, "MATRIX (%dx%d):%d - %p\n",
		                 m_H, n_H,
		                 FLA_Obj_datatype( H ),
		                 FLA_Obj_buffer( H ) );
		fflush( stdout );
		
		for ( j = 0; j < n_H; ++j )
		{
			for ( i = 0; i < m_H; ++i )
			{
				for ( k = 0; k < indent; ++k )
					fprintf( stdout, "  " );
				
				buffer_temp = ( FLA_Obj* ) FLA_Obj_buffer( H );

				FLASH_print_struct_helper( buffer_temp[ j*m_H + i ],
				                           indent + 1);
			}
		}
	}
}


