/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

/* ***************************************************************************

   FLA_Obj_create( )

 *************************************************************************** */

FLA_Error FLA_Obj_create( FLA_Datatype datatype, int m, int n, FLA_Obj *obj )
{
  FLA_Obj_create_ext( datatype, FLA_SCALAR, m, n, m, n, obj );

  return FLA_SUCCESS;
}

void FLA_F2C( fla_obj_create_f ) 
     ( F_INT *datatype, F_INT *m, F_INT *n, F_INT *obj, F_INT *IERROR )
{
  *IERROR = FLA_Obj_create( *( ( FLA_Datatype * ) datatype ), 
                            *( ( int          * ) m        ),
                            *( ( int          * ) n        ),
                             ( ( FLA_Obj      * ) obj      ) );
}

/* ***************************************************************************

   FLA_Obj_create_ext( )

 *************************************************************************** */

FLA_Error FLA_Obj_create_ext( FLA_Datatype datatype, FLA_Elemtype elemtype, int m, int n, int m_inner, int n_inner, FLA_Obj *obj )
{
  size_t buffer_size;
  int    ldim;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;

  e_val = FLA_Check_valid_elemtype( elemtype );
  FLA_Check_error_code( e_val );

  e_val = FLA_Check_valid_datatype( datatype );
  FLA_Check_error_code( e_val );

  e_val = FLA_Check_null_pointer( obj );
  FLA_Check_error_code( e_val );
#endif

  // Populate the fields in the view object.
  obj->m                = m;
  obj->n                = n;
  obj->offm             = 0;
  obj->offn             = 0;

  // Allocate the base object field.
  obj->base             = ( FLA_Base_obj * ) FLA_malloc( sizeof( FLA_Base_obj ) );

  // Populate the fields in the base object.
  obj->base->datatype   = datatype;
  obj->base->elemtype   = elemtype;
  obj->base->m          = m;
  obj->base->n          = n;
  obj->base->m_inner    = m_inner;
  obj->base->n_inner    = n_inner;
  obj->base->id         = ( int ) obj->base;
  obj->base->m_index    = 0;
  obj->base->n_index    = 0;

  // Determine the leading dimension.
  ldim                  = m;
#ifdef FLA_ENABLE_MEMORY_ALIGNMENT
  #ifdef FLA_ENABLE_LDIM_ALIGNMENT
  // We only attempt to align the address if we're creating a scalar object.
  // Aligning leading dimensions of hierarchical objects, whose elements are
  // FLA_Objs, is more difficult, and more prone to wasted padding memory,
  // depending on the size of an FLA_Obj.
  if ( elemtype == FLA_SCALAR )
    ldim                = ( ( m * FLA_Obj_elem_size( *obj ) + 
                            FLA_MEMORY_ALIGNMENT_BOUNDARY - 1 ) / 
                            FLA_MEMORY_ALIGNMENT_BOUNDARY ) *
                          FLA_MEMORY_ALIGNMENT_BOUNDARY /
                          FLA_Obj_elem_size( *obj );
  #endif
#endif
  obj->base->ldim       = ldim;

  // Allocate the base object's element buffer.
  buffer_size           = ( size_t ) ldim *
                          ( size_t ) n *
                          ( size_t ) FLA_Obj_elem_size( *obj );
  obj->base->buffer     = FLA_malloc( buffer_size );

  // Initialize OOC fields (currently unused).
  obj->base->offset     = -1;
  obj->base->fp         = -1;

#ifdef FLA_ENABLE_SUPERMATRIX
  // Initialize SuperMatrix fields.
  obj->base->enqueue = ( FLASH_Enq * ) FLA_malloc( sizeof( FLASH_Enq ) );

  obj->base->enqueue->n_read_tasks   = 0;
  obj->base->enqueue->read_task_head = NULL;
  obj->base->enqueue->read_task_tail = NULL;
  obj->base->enqueue->write_task     = NULL;
#endif

#ifdef FLA_ENABLE_PLAPACK2E 
  obj->base->templ = NULL;
#endif

  return FLA_SUCCESS;
}

/* ***************************************************************************

   FLA_Obj_create_conf_to( )

 *************************************************************************** */

FLA_Error FLA_Obj_create_conf_to( FLA_Trans trans, FLA_Obj obj_old, FLA_Obj *obj )
{
  FLA_Datatype datatype;
  int          m, n, ldim; 
  size_t       buffer_size;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error    e_val;
#endif

  datatype = FLA_Obj_datatype( obj_old );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  e_val = FLA_Check_valid_real_trans( trans );
  FLA_Check_error_code( e_val );

  e_val = FLA_Check_valid_datatype( datatype );
  FLA_Check_error_code( e_val );
#endif

  // Query the dimensions of the existing object.
  if ( trans == FLA_NO_TRANSPOSE )
  {
    m = FLA_Obj_length( obj_old );
    n = FLA_Obj_width( obj_old );
  }
  else
  {
    m = FLA_Obj_width( obj_old );
    n = FLA_Obj_length( obj_old );
  }

  // Populate the fields in the view object.
  obj->m                = m;
  obj->n                = n;
  obj->offm             = 0;
  obj->offn             = 0;

  // Allocate the base object field.
  obj->base             = ( FLA_Base_obj * ) FLA_malloc( sizeof( FLA_Base_obj ) );

  // Populate the fields in the base object.
  obj->base->datatype   = datatype;
  obj->base->elemtype   = FLA_SCALAR;
  obj->base->m          = m;
  obj->base->n          = n;
  obj->base->m_inner    = m;
  obj->base->n_inner    = n;
  obj->base->id         = ( int ) obj->base;
  obj->base->m_index    = 0;
  obj->base->n_index    = 0;

  // Determine the leading dimension.
  ldim                  = m;
#ifdef FLA_ENABLE_MEMORY_ALIGNMENT
  #ifdef FLA_ENABLE_LDIM_ALIGNMENT
  // Note that here we don't guard the ldim computation with a conditional
  // like we do in FLA_Obj_create_ext(), because we assume that the matrix
  // given to this function has an element type of FLA_SCALAR.
  ldim                  = ( ( m * FLA_Obj_elem_size( *obj ) + 
                            FLA_MEMORY_ALIGNMENT_BOUNDARY - 1 ) / 
                            FLA_MEMORY_ALIGNMENT_BOUNDARY ) *
                          FLA_MEMORY_ALIGNMENT_BOUNDARY /
                          FLA_Obj_elem_size( *obj );
  #endif
#endif
  obj->base->ldim       = ldim;

  // Allocate the base object's element buffer.
  buffer_size           = ( size_t ) ldim * 
                          ( size_t ) n *
                          ( size_t ) FLA_Obj_elem_size( *obj );
  obj->base->buffer     = FLA_malloc( buffer_size );

  // Initialize OOC fields (currently unused).
  obj->base->offset     = -1;
  obj->base->fp         = -1;

#ifdef FLA_ENABLE_SUPERMATRIX
  // Initialize SuperMatrix fields.
  obj->base->enqueue = ( FLASH_Enq * ) FLA_malloc( sizeof( FLASH_Enq ) );

  obj->base->enqueue->n_read_tasks   = 0;
  obj->base->enqueue->read_task_head = NULL;
  obj->base->enqueue->read_task_tail = NULL;
  obj->base->enqueue->write_task     = NULL;
#endif

#ifdef FLA_ENABLE_PLAPACK2E 
  obj->base->templ = NULL;
#endif

  return FLA_SUCCESS;
}

void FLA_F2C( fla_obj_create_conf_to_f ) 
     ( F_INT *trans, F_INT *obj_old, F_INT *obj, F_INT *IERROR )
{
  *IERROR = FLA_Obj_create_conf_to( *( ( FLA_Trans * ) trans   ), 
                                    *( ( FLA_Obj   * ) obj_old ),
                                     ( ( FLA_Obj   * ) obj     ) );
}

/* ***************************************************************************

   FLA_Obj_create_without_buffer( )

 *************************************************************************** */

FLA_Error FLA_Obj_create_without_buffer( FLA_Datatype datatype, int m, int n, FLA_Obj *obj )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;

  e_val = FLA_Check_valid_datatype( datatype );
  FLA_Check_error_code( e_val );

  e_val = FLA_Check_null_pointer( obj );
  FLA_Check_error_code( e_val );
#endif

  // Populate the fields in the view object.
  obj->m                = m;
  obj->n                = n;
  obj->offm             = 0;
  obj->offn             = 0;

  // Allocate the base object field.
  obj->base             = ( FLA_Base_obj * ) FLA_malloc( sizeof( FLA_Base_obj ) );

  // Populate the fields in the base object.
  obj->base->datatype   = datatype;
  obj->base->elemtype   = FLA_SCALAR;
  obj->base->m          = m;
  obj->base->n          = n;
  obj->base->m_inner    = m;
  obj->base->n_inner    = n;
  obj->base->id         = ( int ) obj->base;
  obj->base->m_index    = 0;
  obj->base->n_index    = 0;

  // Set the leading dimension.
  obj->base->ldim       = -1;

  // Initialize the base object's element buffer to NULL.
  obj->base->buffer     = NULL;

  // Initialize OOC fields (currently unused).
  obj->base->offset     = -1;
  obj->base->fp         = -1;

#ifdef FLA_ENABLE_SUPERMATRIX
  // Initialize SuperMatrix fields.
  obj->base->enqueue = ( FLASH_Enq * ) FLA_malloc( sizeof( FLASH_Enq ) );

  obj->base->enqueue->n_read_tasks   = 0;
  obj->base->enqueue->read_task_head = NULL;
  obj->base->enqueue->read_task_tail = NULL;
  obj->base->enqueue->write_task     = NULL;
#endif

#ifdef FLA_ENABLE_PLAPACK2E 
  obj->base->templ = NULL;
#endif

  return FLA_SUCCESS;
}

void FLA_F2C( fla_obj_create_w_out_buff_f ) 
     ( F_INT *datatype, F_INT *m, F_INT *n, F_INT *obj, F_INT *IERROR )
{
  *IERROR = FLA_Obj_create_without_buffer( *( ( FLA_Datatype * ) datatype ), 
                                           *( ( int          * ) m        ),
                                           *( ( int          * ) n        ),
                                            ( ( FLA_Obj      * ) obj      ) );
}
 
/* ***************************************************************************

   FLA_Obj_constant_create( )

 *************************************************************************** */

FLA_Error FLA_Obj_constant_create( double const_real, FLA_Obj *obj )
{
  float    *temp_s;
  double   *temp_d;
  scomplex *temp_c;
  dcomplex *temp_z;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error       e_val;

  e_val = FLA_Check_null_pointer( obj );
  FLA_Check_error_code( e_val );
#endif

  FLA_Obj_create( FLA_CONSTANT, 1, 1, obj );

  temp_s       = FLA_FLOAT_PTR( *obj );
  temp_d       = FLA_DOUBLE_PTR( *obj );
  temp_c       = FLA_COMPLEX_PTR( *obj );
  temp_z       = FLA_DOUBLE_COMPLEX_PTR( *obj );

  *temp_s      = ( float ) const_real;
  *temp_d      =           const_real;
  temp_c->real = ( float ) const_real;
  temp_c->imag = ( float ) 0.0;
  temp_z->real =           const_real;
  temp_z->imag =           0.0;

  return FLA_SUCCESS;
}

void FLA_F2C( fla_obj_constant_create_f ) 
     ( F_DOUBLE* const_real, F_INT *obj, F_INT *IERROR )
{
  *IERROR = FLA_Obj_constant_create( *( ( double  * ) const_real ),
                                      ( ( FLA_Obj * ) obj        ) );
}

/* ***************************************************************************

   FLA_Obj_complex_constant_create( )

 *************************************************************************** */

FLA_Error FLA_Obj_complex_constant_create( double const_real, double const_imag, FLA_Obj *obj )
{
  float          *temp_s;
  double         *temp_d;
  scomplex       *temp_c;
  dcomplex *temp_z;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error       e_val;

  e_val = FLA_Check_null_pointer( obj );
  FLA_Check_error_code( e_val );
#endif

  FLA_Obj_create( FLA_CONSTANT, 1, 1, obj );

  temp_s       = FLA_FLOAT_PTR( *obj );
  temp_d       = FLA_DOUBLE_PTR( *obj );
  temp_c       = FLA_COMPLEX_PTR( *obj );
  temp_z       = FLA_DOUBLE_COMPLEX_PTR( *obj );

  *temp_s      = ( float ) const_real;
  *temp_d      =           const_real;
  temp_c->real = ( float ) const_real;
  temp_c->imag = ( float ) const_imag;
  temp_z->real =           const_real;
  temp_z->imag =           const_imag;

  return FLA_SUCCESS;
}

void FLA_F2C( fla_obj_complex_constant_create_f ) 
     ( F_DOUBLE* const_real, F_DOUBLE* const_imag, F_INT *obj, F_INT *IERROR )
{
  *IERROR = FLA_Obj_complex_constant_create( *( ( double  * ) const_real ),
                                             *( ( double  * ) const_imag ),
                                              ( ( FLA_Obj * ) obj        ) );
}

/* ***************************************************************************

   FLA_Obj_attach_buffer( )

 *************************************************************************** */

FLA_Error FLA_Obj_attach_buffer( void *buffer, int lda, FLA_Obj *obj )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;

  e_val = FLA_Check_null_pointer( obj );
  FLA_Check_error_code( e_val );
#endif

  obj->base->buffer = buffer;
  obj->base->ldim   = lda;

  return FLA_SUCCESS;
}

void FLA_F2C( fla_obj_attach_buff_f ) 
     ( F_INT *buffer, F_INT *lda, F_INT *obj, F_INT *IERROR )
{
  *IERROR = FLA_Obj_attach_buffer(  ( ( void    * ) buffer ),
                                   *( ( int     * ) lda    ),
                                    ( ( FLA_Obj * ) obj    ) );
}

/* ***************************************************************************

   FLA_Obj_free( )

 *************************************************************************** */

FLA_Error FLA_Obj_free( FLA_Obj *obj )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;

  e_val = FLA_Check_null_pointer( obj );
  FLA_Check_error_code( e_val );

  e_val = FLA_Check_null_pointer( obj->base );
  FLA_Check_error_code( e_val );
#endif

#ifdef FLA_ENABLE_SUPERMATRIX
  FLA_free( ( void * ) obj->base->enqueue );
#endif

  FLA_free( obj->base->buffer );
  FLA_free( ( void * ) obj->base );
  obj->offm = 0;
  obj->offn = 0;
  obj->m = 0;
  obj->n = 0;

  return FLA_SUCCESS;
}

void FLA_F2C( fla_obj_free_f ) 
     ( F_INT *obj, F_INT *IERROR )
{
  *IERROR = FLA_Obj_free( ( FLA_Obj * ) obj );
}
 
/* ***************************************************************************

   FLA_Obj_free_without_buffer( )

 *************************************************************************** */

FLA_Error FLA_Obj_free_without_buffer( FLA_Obj *obj )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;

  e_val = FLA_Check_null_pointer( obj );
  FLA_Check_error_code( e_val );

  e_val = FLA_Check_null_pointer( obj->base );
  FLA_Check_error_code( e_val );
#endif

#ifdef FLA_ENABLE_SUPERMATRIX
  FLA_free( ( void * ) obj->base->enqueue );
#endif

  FLA_free( ( void * ) obj->base );
  obj->offm = 0;
  obj->offn = 0;
  obj->m = 0;
  obj->n = 0;

  return FLA_SUCCESS;
}

void FLA_F2C( fla_obj_free_w_out_buff_f ) 
     ( F_INT *obj, F_INT *IERROR )
{
  *IERROR = FLA_Obj_free_without_buffer( ( FLA_Obj * ) obj );
}

/* ***************************************************************************

   FLA_Obj_datatype( )

 *************************************************************************** */

FLA_Datatype FLA_Obj_datatype( FLA_Obj obj )
{
  return obj.base->datatype;
}

F_INT FLA_F2C( fla_obj_datatype_f ) ( F_INT *obj )
{
  return FLA_Obj_datatype( *( ( FLA_Obj * ) obj ) );
}

/* ***************************************************************************

   FLA_Obj_elemtype( )

 *************************************************************************** */

FLA_Elemtype FLA_Obj_elemtype( FLA_Obj obj )
{
  return obj.base->elemtype;
}

F_INT FLA_F2C( fla_obj_elemtype_f ) ( F_INT *obj )
{
  return FLA_Obj_elemtype( *( ( FLA_Obj * ) obj ) );
}

/* ***************************************************************************

   FLA_Obj_datatype_size( )

 *************************************************************************** */

int FLA_Obj_datatype_size( FLA_Datatype datatype )
{
  int typesize = 0;

  switch( datatype ){
  case FLA_INT: 
    typesize = sizeof( int );
    break;
  case FLA_FLOAT: 
    typesize = sizeof( float );
    break;
  case FLA_DOUBLE: 
    typesize = sizeof( double );
    break;
  case FLA_COMPLEX: 
    typesize = sizeof( scomplex );
    break;
  case FLA_DOUBLE_COMPLEX: 
    typesize = sizeof( dcomplex );
    break;
  case FLA_CONSTANT: 
    typesize = sizeof( double ) +
               sizeof( double ) + sizeof( double ) + 
               sizeof( scomplex ) + sizeof( dcomplex );
    break;
  case FLA_OBJ:
    // This case is for internal use only! And is needed primarily
    // for fla_init_fortran() when it needs to know how much memory
    // to copy for the global FLA_CONSTANTs.
    typesize = sizeof( FLA_Obj );
    break;
  default:
    FLA_Check_error_code( FLA_INVALID_DATATYPE );
  }

  return typesize;
}

F_INT FLA_F2C( fla_obj_datatype_size_f ) ( F_INT *datatype )
{
  return FLA_Obj_datatype_size( *( ( FLA_Datatype * ) datatype ) );
}

/* ***************************************************************************

   FLA_Obj_elem_size( )

 *************************************************************************** */

int FLA_Obj_elem_size( FLA_Obj obj )
{
  int elem_size = 0;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING 
  FLA_Error e_val;
  
  e_val = FLA_Check_valid_elemtype( FLA_Obj_elemtype( obj ) );
  FLA_Check_error_code( e_val );
  
  e_val = FLA_Check_valid_datatype( FLA_Obj_datatype( obj ) );
  FLA_Check_error_code( e_val );
#endif

  if ( FLA_Obj_elemtype( obj ) == FLA_MATRIX )
  {
    elem_size = sizeof( FLA_Obj );
  }
  else
  {
    switch( FLA_Obj_datatype( obj ) )
    {
      case FLA_INT: 
        elem_size = sizeof( int );
        break;
      case FLA_FLOAT: 
        elem_size = sizeof( float );
        break;
      case FLA_DOUBLE: 
        elem_size = sizeof( double );
        break;
      case FLA_COMPLEX: 
        elem_size = sizeof( scomplex );
        break;
      case FLA_DOUBLE_COMPLEX: 
        elem_size = sizeof( dcomplex );
        break;
      case FLA_CONSTANT: 
        elem_size = sizeof( double ) +
                    sizeof( double ) + sizeof( double ) + 
                    sizeof( scomplex ) + sizeof( dcomplex );
      break;
        case FLA_OBJ:
        // This case is for internal use only! And is needed primarily
        // for fla_init_fortran() when it needs to know how much memory
        // to copy for the global FLA_CONSTANTs.
        elem_size = sizeof( FLA_Obj );
        break;
      default:
        // This default case should never execute.
        FLA_Check_error_code( FLA_INVALID_DATATYPE );
    }
  }

  return elem_size;
}

F_INT FLA_F2C( fla_obj_elem_size_f ) ( F_INT *obj )
{
  return FLA_Obj_elem_size( *( ( FLA_Obj * ) obj ) );
}

/* ***************************************************************************

   FLA_Obj_length( )

 *************************************************************************** */

int FLA_Obj_length( FLA_Obj obj )
{
  return obj.m;
}

F_INT FLA_F2C( fla_obj_length_f ) ( F_INT *obj )
{
  return FLA_Obj_length( *( ( FLA_Obj * ) obj ) );
}

/* ***************************************************************************

   FLA_Obj_width( )

 *************************************************************************** */

int FLA_Obj_width( FLA_Obj obj )
{
  return obj.n;
}

F_INT FLA_F2C( fla_obj_width_f ) ( F_INT *obj )
{
  return FLA_Obj_width( *( ( FLA_Obj * ) obj ) );
}

/* ***************************************************************************

   FLA_Obj_vector_dim( )

 *************************************************************************** */

int FLA_Obj_vector_dim( FLA_Obj obj )
{
  return ( obj.m == 1 ? obj.n : obj.m );
}

F_INT FLA_F2C( fla_obj_vector_dim_f ) ( F_INT *obj )
{
  return FLA_Obj_vector_dim( *( ( FLA_Obj * ) obj ) );
}

/* ***************************************************************************

   FLA_Obj_min_dim( )

 *************************************************************************** */

int FLA_Obj_min_dim( FLA_Obj obj )
{
  return min( obj.m, obj.n );
}

F_INT FLA_F2C( fla_obj_min_dim_f ) ( F_INT *obj )
{
  return FLA_Obj_min_dim( *( ( FLA_Obj * ) obj ) );
}

/* ***************************************************************************

   FLA_Obj_max_dim( )

 *************************************************************************** */

int FLA_Obj_max_dim( FLA_Obj obj )
{
  return max( obj.m, obj.n );
}

F_INT FLA_F2C( fla_obj_max_dim_f ) ( F_INT *obj )
{
  return FLA_Obj_max_dim( *( ( FLA_Obj * ) obj ) );
}

/* ***************************************************************************

   FLA_Obj_ldim( )

 *************************************************************************** */

int FLA_Obj_ldim( FLA_Obj obj )
{
  return (obj.base)->ldim;
}

F_INT FLA_F2C( fla_obj_ldim_f ) ( F_INT *obj )
{
  return FLA_Obj_ldim( *( ( FLA_Obj * ) obj ) );
}

/* ***************************************************************************

   FLA_Obj_buffer( )

 *************************************************************************** */

void* FLA_Obj_buffer( FLA_Obj obj )
{
  void*  buffer;
  size_t elem_size, offm, offn, ldim;
  size_t byte_offset;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING 
  FLA_Error e_val;
  
  e_val = FLA_Check_null_pointer( obj.base );
  FLA_Check_error_code( e_val );
#endif

//  buffer = ( char * ) (obj.base)->buffer + 
//                      elem_size * ( obj.offn * FLA_Obj_ldim( obj ) + 
//                                    obj.offm );

  elem_size = ( size_t ) FLA_Obj_elem_size( obj );
  ldim      = ( size_t ) FLA_Obj_ldim( obj );
  offm      = ( size_t ) obj.offm;
  offn      = ( size_t ) obj.offn;

  byte_offset = elem_size * ( offn * ldim + offm );

  buffer = ( char * ) (obj.base)->buffer + byte_offset;

  return buffer;
}

