/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Conjugate_t( FLA_Uplo uplo, FLA_Obj A )
{
  FLA_Datatype datatype;
  int          i, j;
  int          m_A, n_A, ldim_A;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() )
    FLA_Conjugate_t_check( uplo, A );
#endif

  if ( FLA_Obj_is_real( A ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  switch ( datatype ){

  case FLA_COMPLEX:
  {
    scomplex *buff_A  = ( scomplex * ) FLA_COMPLEX_PTR( A );
    float    *buff_n1 = ( float    * ) FLA_FLOAT_PTR( FLA_MINUS_ONE );

    if ( uplo == FLA_LOWER_TRIANGULAR )
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = j; i < m_A; i++ )
          buff_A[ j*ldim_A + i ].imag *= *buff_n1;
      } 
    }
    else if ( uplo == FLA_UPPER_TRIANGULAR ) 
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j + 1; i++ )
          buff_A[ j*ldim_A + i ].imag *= *buff_n1;
      } 
    }

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A  = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    double   *buff_n1 = ( double   * ) FLA_DOUBLE_PTR( FLA_MINUS_ONE );

    if ( uplo == FLA_LOWER_TRIANGULAR )
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = j; i < m_A; i++ )
          buff_A[ j*ldim_A + i ].imag *= *buff_n1;
      } 
    }
    else if ( uplo == FLA_UPPER_TRIANGULAR ) 
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j + 1; i++ )
          buff_A[ j*ldim_A + i ].imag *= *buff_n1;
      } 
    }

    break;
  }

  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_conjugate_t_f )
     ( F_INT *uplo, F_INT *A, F_INT *IERROR )
{
  *IERROR = FLA_Conjugate_t( *( ( FLA_Uplo * ) uplo ),
                             *( ( FLA_Obj  * ) A    ) );
}

