/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Symmetrize( FLA_Uplo uplo, FLA_Obj A )
{
  FLA_Datatype datatype;
  int          j;
  int          n_A, ldim_A;
  int          ldim_src, inc_src;
  int          ldim_dst, inc_dst;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() )
    FLA_Symmetrize_check( uplo, A );
#endif

  datatype = FLA_Obj_datatype( A );

  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  if ( uplo == FLA_LOWER_TRIANGULAR )
  {
    ldim_src = 1;
    ldim_dst = ldim_A;
    inc_src  = ldim_A;
    inc_dst  = 1;
  }
  else
  {
    ldim_src = ldim_A;
    ldim_dst = 1;
    inc_src  = 1;
    inc_dst  = ldim_A;
  }

  switch ( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A = ( float * ) FLA_FLOAT_PTR( A );

    for ( j = 1; j < n_A; j++ )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_scopy( j,
                   buff_A + j*ldim_src, inc_src,
                   buff_A + j*ldim_dst, inc_dst );
#else
      FLA_C2F( scopy )( &j,
                        buff_A + j*ldim_src, &inc_src,
                        buff_A + j*ldim_dst, &inc_dst );
#endif
    }

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A = ( double * ) FLA_DOUBLE_PTR( A );

    for ( j = 1; j < n_A; j++ )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_dcopy( j,
                   buff_A + j*ldim_src, inc_src,
                   buff_A + j*ldim_dst, inc_dst );
#else
      FLA_C2F( dcopy )( &j,
                        buff_A + j*ldim_src, &inc_src,
                        buff_A + j*ldim_dst, &inc_dst );
#endif
    }

    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A = ( scomplex * ) FLA_COMPLEX_PTR( A );

    for ( j = 1; j < n_A; j++ )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_ccopy( j,
                   buff_A + j*ldim_src, inc_src,
                   buff_A + j*ldim_dst, inc_dst );
#else
      FLA_C2F( ccopy )( &j,
                        buff_A + j*ldim_src, &inc_src,
                        buff_A + j*ldim_dst, &inc_dst );
#endif
    }

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );

    for ( j = 1; j < n_A; j++ )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_zcopy( j,
                   buff_A + j*ldim_src, inc_src,
                   buff_A + j*ldim_dst, inc_dst );
#else
      FLA_C2F( zcopy )( &j,
                        buff_A + j*ldim_src, &inc_src,
                        buff_A + j*ldim_dst, &inc_dst );
#endif
    }

    break;
  }

  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_symmetrize_f )
     ( F_INT *uplo, F_INT *A, F_INT *IERROR )
{
  *IERROR = FLA_Symmetrize( *( ( FLA_Uplo * ) uplo ),
                            *( ( FLA_Obj  * ) A    ) );
}

