/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

#define sign( x ) ( (x) < 0 ? -1.0 : 1.0 )

FLA_Error FLA_Househ2( FLA_Obj chi_1, FLA_Obj x2, FLA_Obj beta )
{    
/*
  Compute Householder transform ( I - beta / 1  \ ( 1 u2^T ) ):
                                           \ u2 /

  Compute u2 and beta such that
  ( I - beta / 1  \ ( 1 u2^T ) ) / chi_1 \ = / -+ || x ||_2 \
             \ u2 /              \  x2   /   \       0      /
  where 
      x    =  / chi_1 \ 
              \   x2  /
      beta = 2 / ( 1 + u2^T u2 )

  Upon completion, -+ || x ||_2 has overwritten chi_1
                   and u2 has overwritten x2
*/
  int x2_stride; 
  int x2_length;
  int i_one = 1;
  int i_two = 2; 

  switch ( FLA_Obj_datatype( chi_1 ) )
  {

  case FLA_DOUBLE:
  {
    double *chi_1_buff = ( double * ) FLA_DOUBLE_PTR( chi_1 );
    double *x2_buff    = ( double * ) FLA_DOUBLE_PTR( x2 );
    double *beta_buff  = ( double * ) FLA_DOUBLE_PTR( beta );
    double *zero_buff  = ( double * ) FLA_DOUBLE_PTR( FLA_ZERO );
    double y[ 2 ];
    double temp;
    
    if ( FLA_Obj_length( x2 ) == 1 )
    {
      x2_stride = FLA_Obj_ldim( x2 );
      x2_length = FLA_Obj_width( x2 );
    }
    else
    {
      x2_stride = 1;
      x2_length = FLA_Obj_length( x2 );
    }

    if ( x2_length == 0 )
    {
      *beta_buff = *zero_buff;
      return FLA_SUCCESS;
    }
    else if ( x2_length < 0 )
    {
      /*
        Negative length of x_2 encountered. If Hess was invoked, check ilo, ihi
        for correct index base--may have to subtract 1 from Fortran ilo/ihi
      */
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
      FLA_Check_error_code( FLA_EXPECTED_NONNEGATIVE_VALUE );
#endif
    }
      
    /* 
      y = /    chi_1    \
          \ || x_2 ||_2 /  ( notice: || y ||_2 = || x ||_2 )   
    */

    y[ 0 ] = *chi_1_buff;

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    y[ 1 ] = cblas_dnrm2( x2_length, x2_buff, x2_stride );
#else
    y[ 1 ] = FLA_C2F( dnrm2 )( &x2_length, x2_buff, &x2_stride );
#endif

    if ( y[ 1 ] == *zero_buff )
    {
      /* Rotation is identity */
      *beta_buff = *zero_buff;
    }
    else
    {

      /* 
        temp  = || / chi_1 \ ||
                || \  x_2  / ||_2   
      */

#ifdef FLA_ENABLE_CBLAS_INTERFACE
      temp = cblas_dnrm2( i_two, y, i_one );
#else
      temp = FLA_C2F( dnrm2 )( &i_two, y, &i_one );
#endif

      /* 
        y = / chi_1 + sign( chi_1 ) * temp  \
            \         x_2                   / 
      */

      if ( sign( y[ 0 ] ) < 1.0 )
      {
        y[ 0 ] -= temp; 
        *chi_1_buff = temp;
      }
      else
      {
        y[ 0 ] += temp; 
        *chi_1_buff = -temp;
      }

      /* 
        y = /     1        \
            \ eta_2/eta_1  /  ( normalize y to have "1" first entry ) 
      */

      y[ 1 ] = y[ 1 ] / y[ 0 ];

      /* 
        Overwrite x_2 with x_2/eta_2 
      */

      y[ 0 ] = 1.0 / y[ 0 ];
      
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_dscal( x2_length, y[ 0 ], x2_buff, x2_stride );
#else
      FLA_C2F( dscal )( &x2_length, &y[ 0 ], x2_buff, &x2_stride );
#endif

      y[ 0 ] = 1.0;

      /* 
        temp  = y^T y 
      */

#ifdef FLA_ENABLE_CBLAS_INTERFACE
      temp = cblas_ddot( i_two, y, i_one, y, i_one );
#else
      temp = FLA_C2F( ddot )( &i_two, y, &i_one, y, &i_one );
#endif

      *beta_buff = 2.0 / temp;
    }
    
    break;
  }

  default:
  {
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
      FLA_Check_error_code( FLA_NOT_YET_IMPLEMENTED );
#endif
  }
  }

  return FLA_SUCCESS;
}
