/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

#define sign( x ) ( (x) < 0 ? -1.0 : 1.0 )

FLA_Error FLA_Househ2_UT_opt( int n, double *chi_1, double *x2, int inc_x2, double *tau )
/*
  Compute Householder transform ( I - 1/tau / 1  \ ( 1 u2^T ) ):
                                            \ u2 /

  Compute u2 and tau such that
  ( I - 1/tau / 1  \ ( 1 u2^T ) ) / chi_1 \ = / -+ || x ||_2 \
              \ u2 /              \  x2   /   \       0      /
  where 
      x    =  / chi_1 \ 
              \   x2  /
      tau = ( 1 + u2^T u2 )/2

  Upon completion, -+ || x ||_2 has overwritten chi_1
                   and u2 has overwritten x2
*/
{    
  int 
    i_one = 1, 
    i_two = 2, 
    stride_x2, 
    length_x2;

  double
    y[ 2 ], temp,
    d_zero = 0.0,
    d_half = 1.0 / 2.0;
    
  length_x2 = n;
  stride_x2 = inc_x2;

  if ( length_x2 == 0 ) {
    *tau = d_half;
    *chi_1 = - *chi_1;
    return FLA_SUCCESS;
  }
  else if ( length_x2 < 0 ) {
    /*
      Negative length of x_2 encountered. If Hess was invoked, check ilo, ihi
      for correct index base--may have to subtract 1 from Fortran ilo/ihi 
    */
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
    FLA_Check_error_code( FLA_EXPECTED_NONNEGATIVE_VALUE );
#endif
  }
    
  /* 
    y = /    chi_1    \
        \ || x_2 ||_2 /  ( notice: || y ||_2 = || x ||_2 )   
  */

  y[ 0 ] = *chi_1;

#ifdef FLA_ENABLE_CBLAS_INTERFACE
  y[ 1 ] = cblas_dnrm2( length_x2, x2, stride_x2 );
#else
  y[ 1 ] = FLA_C2F( dnrm2 )( &length_x2, x2, &stride_x2 );
#endif

  if ( y[ 1 ] == d_zero ) {
    *tau = d_half;
    *chi_1 = - *chi_1;
  }
  else {

    /* 
      temp  = || / chi_1 \ ||
              || \  x_2  / ||_2   
    */

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    temp = cblas_dnrm2( i_two, y, i_one );
#else
    temp = FLA_C2F( dnrm2 )( &i_two, y, &i_one );
#endif

    /* 
      y = / chi_1 + sign( chi_1 ) * temp  \
          \         x_2                   / 
    */

    if ( sign( y[ 0 ] ) < 1 ) {
      y[ 0 ] -= temp; 
      *chi_1 = temp;
    }
    else {
      y[ 0 ] += temp; 
      *chi_1 = -temp;
    }

    /* 
      y = /     1        \
          \ eta_2/eta_1  /  ( normalize y to have "1" first entry ) 
    */

    y[ 1 ] = y[ 1 ] / y[ 0 ];

    /* 
      Overwrite x_2 with x_2/eta_2 
    */

    y[ 0 ] = 1.0 / y[ 0 ];

#ifdef FLA_ENABLE_CBLAS_INTERFACE    
    cblas_dscal( length_x2, y[ 0 ], x2, stride_x2 );
#else
    FLA_C2F( dscal )( &length_x2, &y[ 0 ], x2, &stride_x2 );
#endif

    y[ 0 ] = 1.0;

    /* 
      temp  = y^T y 
    */

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    temp = cblas_ddot( i_two, y, i_one, y, i_one );
#else
    temp = FLA_C2F( ddot )( &i_two, y, &i_one, y, &i_one );
#endif

    *tau = temp / 2.0;
  }

  return FLA_SUCCESS;
}
