/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Axpy( FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  FLA_Datatype datatype;
  int          j, n_iter;
  int          num_elem;
  int          m_A, n_A, inc_A, ldim_A;
  int          m_B,      inc_B, ldim_B;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Axpy_check( alpha, A, B );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  m_B      = FLA_Obj_length( B );
  ldim_B   = FLA_Obj_ldim( B );

  if ( FLA_Obj_is_vector( A ) )
  {
    inc_A    = ( m_A == 1 ? ldim_A : 1 );
    inc_B    = ( m_B == 1 ? ldim_B : 1 );
    n_iter   = 1;
    num_elem = FLA_Obj_max_dim( A );
  }
  else
  {
    inc_A    = 1;
    inc_B    = 1;
    n_iter   = n_A;
    num_elem = m_A;
  }

  switch ( datatype ){

  case FLA_FLOAT:
  {
    float *buff_alpha = ( float * ) FLA_FLOAT_PTR( alpha );
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_B     = ( float * ) FLA_FLOAT_PTR( B );

    for ( j = 0; j < n_iter; ++j )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_saxpy( num_elem,
                   *buff_alpha, 
                   buff_A + j*ldim_A, inc_A, 
                   buff_B + j*ldim_B, inc_B );
#else
      FLA_C2F( saxpy )( &num_elem,
                        buff_alpha, 
                        buff_A + j*ldim_A, &inc_A, 
                        buff_B + j*ldim_B, &inc_B );
#endif
    }

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_alpha = ( double * ) FLA_DOUBLE_PTR( alpha );
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_B     = ( double * ) FLA_DOUBLE_PTR( B );

    for ( j = 0; j < n_iter; ++j )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_daxpy( num_elem,
                   *buff_alpha, 
                   buff_A + j*ldim_A, inc_A, 
                   buff_B + j*ldim_B, inc_B );
#else
      FLA_C2F( daxpy )( &num_elem,
                        buff_alpha, 
                        buff_A + j*ldim_A, &inc_A, 
                        buff_B + j*ldim_B, &inc_B );
#endif
    }

    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
    scomplex *buff_A =     ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_B =     ( scomplex * ) FLA_COMPLEX_PTR( B );

    for ( j = 0; j < n_iter; ++j )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_caxpy( num_elem,
                   *buff_alpha, 
                   buff_A + j*ldim_A, inc_A, 
                   buff_B + j*ldim_B, inc_B );
#else
      FLA_C2F( caxpy )( &num_elem,
                        buff_alpha, 
                        buff_A + j*ldim_A, &inc_A, 
                        buff_B + j*ldim_B, &inc_B );
#endif
    }

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_B     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( B );

    for ( j = 0; j < n_iter; ++j )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_zaxpy( num_elem,
                   *buff_alpha, 
                   buff_A + j*ldim_A, inc_A, 
                   buff_B + j*ldim_B, inc_B );
#else
      FLA_C2F( zaxpy )( &num_elem,
                        buff_alpha, 
                        buff_A + j*ldim_A, &inc_A, 
                        buff_B + j*ldim_B, &inc_B );
#endif
    }

    break;
  }

  }
  
  return FLA_SUCCESS;
}

void FLA_F2C( fla_axpy_f ) 
     ( F_INT *alpha, F_INT *A, F_INT *B, F_INT *IERROR )
{
  *IERROR = FLA_Axpy( *( ( FLA_Obj * ) alpha ), 
                      *( ( FLA_Obj * ) A     ), 
                      *( ( FLA_Obj * ) B     ) );
}

