/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Dot_c( FLA_Conj conj, FLA_Obj x, FLA_Obj y, FLA_Obj rho )
{
  FLA_Datatype datatype;
  int          num_elem;
  int          m_x, inc_x, ldim_x;
  int          m_y, inc_y, ldim_y;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Dot_c_check( conj, x, y, rho );
#endif

  if ( FLA_Obj_has_zero_dim( x ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( x );

  m_x      = FLA_Obj_length( x );
  ldim_x   = FLA_Obj_ldim( x );

  m_y      = FLA_Obj_length( y );
  ldim_y   = FLA_Obj_ldim( y ); 

  inc_x    = ( m_x == 1 ? ldim_x : 1 );
  inc_y    = ( m_y == 1 ? ldim_y : 1 );
  num_elem = FLA_Obj_max_dim( x );


  switch ( datatype ){

  case FLA_FLOAT:
  {
    float *buff_x   = ( float * ) FLA_FLOAT_PTR( x );
    float *buff_y   = ( float * ) FLA_FLOAT_PTR( y );
    float *buff_rho = ( float * ) FLA_FLOAT_PTR( rho );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    *buff_rho = 
      cblas_sdot( num_elem, 
                  buff_x, inc_x,
                  buff_y, inc_y );
#else
    *buff_rho = 
      FLA_C2F( sdot ) ( &num_elem, 
                        buff_x, &inc_x,
                        buff_y, &inc_y );
#endif

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_x   = ( double * ) FLA_DOUBLE_PTR( x );
    double *buff_y   = ( double * ) FLA_DOUBLE_PTR( y );
    double *buff_rho = ( double * ) FLA_DOUBLE_PTR( rho );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    *buff_rho = 
      cblas_ddot( num_elem, 
                  buff_x, inc_x,
                  buff_y, inc_y );
#else
    *buff_rho = 
      FLA_C2F( ddot ) ( &num_elem, 
                        buff_x, &inc_x,
                        buff_y, &inc_y );
#endif

    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_x   = ( scomplex * ) FLA_COMPLEX_PTR( x );
    scomplex *buff_y   = ( scomplex * ) FLA_COMPLEX_PTR( y );
    scomplex *buff_rho = ( scomplex * ) FLA_COMPLEX_PTR( rho );

    if ( conj == FLA_NO_CONJUGATE )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_cdotu_sub( num_elem, 
                       buff_x, inc_x,
                       buff_y, inc_y, buff_rho );
#else
      FLA_F2C( fla_cdotu ) ( &num_elem, 
                             buff_x, &inc_x,
                             buff_y, &inc_y, buff_rho );
#endif
    }
    else
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_cdotc_sub( num_elem, 
                       buff_x, inc_x,
                       buff_y, inc_y, buff_rho );
#else
      FLA_F2C( fla_cdotc ) ( &num_elem, 
                             buff_x, &inc_x,
                             buff_y, &inc_y, buff_rho );
#endif
    }

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_x   = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( x );
    dcomplex *buff_y   = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( y );
    dcomplex *buff_rho = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( rho );

    if ( conj == FLA_NO_CONJUGATE )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_zdotu_sub( num_elem, 
                       buff_x, inc_x,
                       buff_y, inc_y, buff_rho );
#else
      FLA_F2C( fla_zdotu ) ( &num_elem, 
                             buff_x, &inc_x,
                             buff_y, &inc_y, buff_rho );
#endif
    }
    else
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_zdotc_sub( num_elem, 
                       buff_x, inc_x,
                       buff_y, inc_y, buff_rho );
#else
      FLA_F2C( fla_zdotc ) ( &num_elem, 
                             buff_x, &inc_x,
                             buff_y, &inc_y, buff_rho );
#endif
    }

    break;
  }

  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_dot_c_f ) 
     ( F_INT *conj, F_INT *x, F_INT *y, F_INT *rho, F_INT *IERROR )
{
  *IERROR = FLA_Dot_c( *( ( FLA_Conj * ) conj ), 
                      *( ( FLA_Obj  * ) x    ), 
                      *( ( FLA_Obj  * ) y    ),
                      *( ( FLA_Obj  * ) rho  ) );
}

