/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Dot_cs( FLA_Conj conj, FLA_Obj alpha, FLA_Obj x, FLA_Obj y, FLA_Obj beta, FLA_Obj rho )
{
  FLA_Datatype datatype;
  int          num_elem;
  int          m_x, inc_x, ldim_x;
  int          m_y, inc_y, ldim_y;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Dot_cs_check( conj, alpha, x, y, beta, rho );
#endif

  if ( FLA_Obj_has_zero_dim( x ) )
  {
    FLA_Scal( beta, rho );
    return FLA_SUCCESS;
  }

  datatype = FLA_Obj_datatype( x );

  m_x      = FLA_Obj_length( x );
  ldim_x   = FLA_Obj_ldim( x );

  m_y      = FLA_Obj_length( y );
  ldim_y   = FLA_Obj_ldim( y ); 

  inc_x    = ( m_x == 1 ? ldim_x : 1 );
  inc_y    = ( m_y == 1 ? ldim_y : 1 );
  num_elem = FLA_Obj_max_dim( x );


  switch ( datatype ){
  
  case FLA_FLOAT:
  {
    float *buff_x      = ( float * ) FLA_FLOAT_PTR( x );
    float *buff_y      = ( float * ) FLA_FLOAT_PTR( y );
    float *buff_rho    = ( float * ) FLA_FLOAT_PTR( rho );
    float *buff_alpha  = ( float * ) FLA_FLOAT_PTR( alpha );
    float *buff_beta   = ( float * ) FLA_FLOAT_PTR( beta );
    float  temp;

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    temp = cblas_sdot( num_elem, 
                       buff_x, inc_x, 
                       buff_y, inc_y ); 
#else
    temp = FLA_C2F( sdot ) ( &num_elem, 
                             buff_x, &inc_x, 
                             buff_y, &inc_y ); 
#endif

    *buff_rho = (*buff_alpha) * temp + (*buff_beta) * (*buff_rho);

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_x      = ( double * ) FLA_DOUBLE_PTR( x );
    double *buff_y      = ( double * ) FLA_DOUBLE_PTR( y );
    double *buff_rho    = ( double * ) FLA_DOUBLE_PTR( rho );
    double *buff_alpha  = ( double * ) FLA_DOUBLE_PTR( alpha );
    double *buff_beta   = ( double * ) FLA_DOUBLE_PTR( beta );
    double  temp;

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    temp = cblas_ddot( num_elem, 
                       buff_x, inc_x, 
                       buff_y, inc_y ); 
#else
    temp = FLA_C2F( ddot ) ( &num_elem, 
                             buff_x, &inc_x, 
                             buff_y, &inc_y ); 
#endif

    *buff_rho = (*buff_alpha) * temp + (*buff_beta) * (*buff_rho);

    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_x      = ( scomplex * ) FLA_COMPLEX_PTR( x );
    scomplex *buff_y      = ( scomplex * ) FLA_COMPLEX_PTR( y );
    scomplex *buff_rho    = ( scomplex * ) FLA_COMPLEX_PTR( rho );
    scomplex *buff_alpha  = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
    scomplex *buff_beta   = ( scomplex * ) FLA_COMPLEX_PTR( beta );
    scomplex  temp, temp_rho;

    if ( conj == FLA_NO_CONJUGATE )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_cdotu_sub( num_elem, 
                       buff_x, inc_x, 
                       buff_y, inc_y, &temp ); 
#else
      FLA_F2C( fla_cdotu ) ( &num_elem, 
                             buff_x, &inc_x, 
                             buff_y, &inc_y, &temp ); 
#endif
    }
    else
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_cdotc_sub( num_elem, 
                       buff_x, inc_x, 
                       buff_y, inc_y, &temp ); 
#else
      FLA_F2C( fla_cdotc ) ( &num_elem, 
                             buff_x, &inc_x, 
                             buff_y, &inc_y, &temp ); 
#endif
    }

    temp_rho.real = buff_alpha->real * temp.real      - buff_alpha->imag * temp.imag +
                    buff_beta->real  * buff_rho->real - buff_beta->imag  * buff_rho->imag;

    temp_rho.imag = buff_alpha->real * temp.imag      + buff_alpha->imag * temp.real +
                    buff_beta->real  * buff_rho->imag + buff_beta->imag  * buff_rho->real;

    buff_rho->real = temp_rho.real;
    buff_rho->imag = temp_rho.imag;

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_x      = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( x );
    dcomplex *buff_y      = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( y );
    dcomplex *buff_rho    = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( rho );
    dcomplex *buff_alpha  = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
    dcomplex *buff_beta   = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( beta );
    dcomplex  temp, temp_rho;

    if ( conj == FLA_NO_CONJUGATE )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_zdotu_sub( num_elem, 
                       buff_x, inc_x, 
                       buff_y, inc_y, &temp ); 
#else
      FLA_F2C( fla_zdotu ) ( &num_elem, 
                             buff_x, &inc_x, 
                             buff_y, &inc_y, &temp ); 
#endif
    }
    else
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_zdotc_sub( num_elem, 
                       buff_x, inc_x, 
                       buff_y, inc_y, &temp ); 
#else
      FLA_F2C( fla_zdotc ) ( &num_elem, 
                             buff_x, &inc_x, 
                             buff_y, &inc_y, &temp ); 
#endif
    }

    temp_rho.real = buff_alpha->real * temp.real      - buff_alpha->imag * temp.imag +
                    buff_beta->real  * buff_rho->real - buff_beta->imag  * buff_rho->imag;

    temp_rho.imag = buff_alpha->real * temp.imag      + buff_alpha->imag * temp.real +
                    buff_beta->real  * buff_rho->imag + buff_beta->imag  * buff_rho->real;

    buff_rho->real = temp_rho.real;
    buff_rho->imag = temp_rho.imag;

    break;
  }

  }
  
  return FLA_SUCCESS;
}

void FLA_F2C( fla_dot_cs_f ) 
     ( F_INT *conj, F_INT *alpha, F_INT *x, F_INT *y, F_INT *beta, F_INT *rho, F_INT *IERROR )
{
  *IERROR = FLA_Dot_cs( *( ( FLA_Conj * ) conj  ),
                        *( ( FLA_Obj  * ) alpha ),
                        *( ( FLA_Obj  * ) x     ),
                        *( ( FLA_Obj  * ) y     ),
                        *( ( FLA_Obj  * ) beta  ),
                        *( ( FLA_Obj  * ) rho   ) );
}

