/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Scal_c( FLA_Conj conjalpha, FLA_Obj alpha, FLA_Obj A )
{
  FLA_Datatype datatype, dt_alpha;
  int          j, n_iter;
  int          num_elem;
  int          m_A, n_A, inc_A, ldim_A;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() )
    FLA_Scal_c_check( conjalpha, alpha, A );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  if ( FLA_Obj_equals( alpha, FLA_ONE ) )
  {
    return FLA_SUCCESS;
  }
  else if ( FLA_Obj_equals( alpha, FLA_ZERO ) )
  {
    FLA_Obj_set_to_scalar( FLA_ZERO, A );
    return FLA_SUCCESS;
  }
  else if ( FLA_Obj_equals( alpha, FLA_MINUS_ONE ) )
  {
    FLA_Negate( A );
    return FLA_SUCCESS;
  }

  dt_alpha = FLA_Obj_datatype( alpha );
  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  if ( FLA_Obj_is_vector( A ) )
  {
    inc_A    = ( m_A == 1 ? ldim_A : 1 );
    n_iter   = 1;
    num_elem = FLA_Obj_max_dim( A );
  }
  else
  {
    inc_A    = 1;
    n_iter   = n_A;
    num_elem = m_A;
  }


  switch ( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_alpha = ( float * ) FLA_FLOAT_PTR( alpha );

    for ( j = 0; j < n_iter; ++j )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_sscal( num_elem,
                   *buff_alpha,
                   buff_A + j*ldim_A, inc_A );
#else
      FLA_C2F( sscal )( &num_elem,
                        buff_alpha,
                        buff_A + j*ldim_A, &inc_A );
#endif
    }

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_alpha = ( double * ) FLA_DOUBLE_PTR( alpha );

    for ( j = 0; j < n_iter; ++j )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_dscal( num_elem,
                   *buff_alpha,
                   buff_A + j*ldim_A, inc_A );
#else
      FLA_C2F( dscal )( &num_elem,
                        buff_alpha,
                        buff_A + j*ldim_A, &inc_A );
#endif
    }

    break;
  }

  case FLA_COMPLEX:
  {
    if ( dt_alpha == FLA_COMPLEX )
    {
      scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
      scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
      scomplex  temp_alpha = *buff_alpha;

      if ( conjalpha == FLA_CONJUGATE )
        temp_alpha.imag *= -1.0F;

      for ( j = 0; j < n_iter; ++j )
      {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
        cblas_cscal( num_elem,
                     &temp_alpha,
                     buff_A + j*ldim_A, inc_A );
#else
        FLA_C2F( cscal )( &num_elem,
                          &temp_alpha,
                          buff_A + j*ldim_A, &inc_A );
#endif
      }
    }
    else if ( dt_alpha == FLA_FLOAT )
    {
      scomplex *buff_A     = ( scomplex * ) FLA_FLOAT_PTR( A );
      float    *buff_alpha = ( float    * ) FLA_FLOAT_PTR( alpha );

      for ( j = 0; j < n_iter; ++j )
      {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
        cblas_csscal( num_elem,
                      *buff_alpha,
                      buff_A + j*ldim_A, inc_A );
#else
        FLA_C2F( csscal )( &num_elem,
                           buff_alpha,
                           buff_A + j*ldim_A, &inc_A );
#endif
      }
    }

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    if ( dt_alpha == FLA_DOUBLE_COMPLEX )
    {
      dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
      dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
      dcomplex  temp_alpha = *buff_alpha;

      if ( conjalpha == FLA_CONJUGATE )
        temp_alpha.imag *= -1.0;

      for ( j = 0; j < n_iter; ++j )
      {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
        cblas_zscal( num_elem,
                     &temp_alpha,
                     buff_A + j*ldim_A, inc_A );
#else
        FLA_C2F( zscal )( &num_elem,
                          &temp_alpha,
                          buff_A + j*ldim_A, &inc_A );
#endif
      }
    }
    else if ( dt_alpha == FLA_DOUBLE )
    {
      dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
      double   *buff_alpha = ( double   * ) FLA_DOUBLE_PTR( alpha );

      for ( j = 0; j < n_iter; ++j )
      {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
        cblas_zdscal( num_elem,
                      *buff_alpha,
                      buff_A + j*ldim_A, inc_A );
#else
        FLA_C2F( zdscal )( &num_elem,
                           buff_alpha,
                           buff_A + j*ldim_A, &inc_A );
#endif
      }
    }

    break;
  }

  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_scal_c_f )
     ( F_INT *conjalpha, F_INT *alpha, F_INT *A, F_INT *IERROR )
{
  *IERROR = FLA_Scal_c( *( ( FLA_Conj * ) conjalpha ),
                        *( ( FLA_Obj  * ) alpha     ),
                        *( ( FLA_Obj  * ) A         ) );
}

