/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Ger_c( FLA_Conj conjx, FLA_Conj conjy, FLA_Obj alpha, FLA_Obj x, FLA_Obj y, FLA_Obj A )
{
  FLA_Datatype datatype;
  int          m_A, n_A, ldim_A;
  int          m_x, inc_x, ldim_x;
  int          m_y, inc_y, ldim_y;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
  CBLAS_ORDER     cblas_order = CblasColMajor;
#endif
  FLA_Obj      x_copy;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Ger_c_check( conjx, conjy, alpha, x, y, A );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  m_x      = FLA_Obj_length( x );
  ldim_x   = FLA_Obj_ldim( x );

  m_y      = FLA_Obj_length( y );
  ldim_y   = FLA_Obj_ldim( y );

  inc_x    = ( m_x == 1 ? ldim_x : 1 );
  inc_y    = ( m_y == 1 ? ldim_y : 1 );

  if ( conjx == FLA_CONJUGATE )
  {
    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, x, &x_copy );
    FLA_Copy_t( FLA_CONJ_NO_TRANSPOSE, x, x_copy );

    m_x    = FLA_Obj_length( x_copy );
    ldim_x = FLA_Obj_ldim( x_copy );
    inc_x  = ( m_x == 1 ? ldim_x : 1 );
  }
  else
  {
    x_copy = x;
  }


  switch( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_x     = ( float * ) FLA_FLOAT_PTR( x );
    float *buff_y     = ( float * ) FLA_FLOAT_PTR( y );
    float *buff_alpha = ( float * ) FLA_FLOAT_PTR( alpha );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_sger( cblas_order,
                m_A, n_A,
                *buff_alpha,
                buff_x, inc_x,
                buff_y, inc_y,
                buff_A, ldim_A ); 
#else
    FLA_C2F( sger )( &m_A, &n_A,
                     buff_alpha,
                     buff_x, &inc_x,
                     buff_y, &inc_y,
                     buff_A, &ldim_A ); 
#endif
    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_x     = ( double * ) FLA_DOUBLE_PTR( x );
    double *buff_y     = ( double * ) FLA_DOUBLE_PTR( y );
    double *buff_alpha = ( double * ) FLA_DOUBLE_PTR( alpha );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_dger( cblas_order,
                m_A, n_A,
                *buff_alpha,
                buff_x, inc_x,
                buff_y, inc_y,
                buff_A, ldim_A ); 
#else
    FLA_C2F( dger )( &m_A, &n_A,
                     buff_alpha,
                     buff_x, &inc_x,
                     buff_y, &inc_y,
                     buff_A, &ldim_A ); 
#endif
    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_x     = ( scomplex * ) FLA_COMPLEX_PTR( x_copy );
    scomplex *buff_y     = ( scomplex * ) FLA_COMPLEX_PTR( y );
    scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );

    if ( conjy == FLA_NO_CONJUGATE )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_cgeru( cblas_order,
                   m_A, n_A,
                   buff_alpha,
                   buff_x, inc_x,
                   buff_y, inc_y,
                   buff_A, ldim_A ); 
#else
      FLA_C2F( cgeru )( &m_A, &n_A,
                        buff_alpha,
                        buff_x, &inc_x,
                        buff_y, &inc_y,
                        buff_A, &ldim_A ); 
#endif
    }
    else
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_cgerc( cblas_order,
                   m_A, n_A,
                   buff_alpha,
                   buff_x, inc_x,
                   buff_y, inc_y,
                   buff_A, ldim_A ); 
#else
      FLA_C2F( cgerc )( &m_A, &n_A,
                        buff_alpha,
                        buff_x, &inc_x,
                        buff_y, &inc_y,
                        buff_A, &ldim_A ); 
#endif
    }

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_x     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( x_copy );
    dcomplex *buff_y     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( y );
    dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );

    if ( conjy == FLA_NO_CONJUGATE )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_zgeru( cblas_order,
                   m_A, n_A,
                   buff_alpha,
                   buff_x, inc_x,
                   buff_y, inc_y,
                   buff_A, ldim_A ); 
#else
      FLA_C2F( zgeru )( &m_A, &n_A,
                        buff_alpha,
                        buff_x, &inc_x,
                        buff_y, &inc_y,
                        buff_A, &ldim_A ); 
#endif
    }
    else
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_zgerc( cblas_order,
                   m_A, n_A,
                   buff_alpha,
                   buff_x, inc_x,
                   buff_y, inc_y,
                   buff_A, ldim_A ); 
#else
      FLA_C2F( zgerc )( &m_A, &n_A,
                        buff_alpha,
                        buff_x, &inc_x,
                        buff_y, &inc_y,
                        buff_A, &ldim_A ); 
#endif
    }

    break;
  }

  }

  if ( conjx == FLA_CONJUGATE )
    FLA_Obj_free( &x_copy );
  
  return FLA_SUCCESS;
}


void FLA_F2C( fla_ger_c_f ) 
     ( F_INT *conjx, F_INT *conjy, F_INT *alpha, F_INT *x, F_INT *y, F_INT *A, F_INT *IERROR )
{
  *IERROR = FLA_Ger_c( *( ( FLA_Conj * ) conjx ),
                       *( ( FLA_Conj * ) conjy ),
                       *( ( FLA_Obj  * ) alpha ),
                       *( ( FLA_Obj  * ) x     ),
                       *( ( FLA_Obj  * ) y     ),
                       *( ( FLA_Obj  * ) A     ) );
}

