/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Her2_c( FLA_Uplo uplo, FLA_Conj conjxy, FLA_Obj alpha, FLA_Obj x, FLA_Obj y, FLA_Obj A )
{
  FLA_Datatype datatype;
  int          m_A, ldim_A;
  int          m_x, ldim_x, inc_x;
  int          m_y, ldim_y, inc_y;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
  CBLAS_ORDER  cblas_order = CblasColMajor;
  CBLAS_UPLO   blas_uplo;   
#else
  char         blas_uplo;
#endif
  FLA_Obj      x_copy, y_copy;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Her2_c_check( uplo, conjxy, alpha, x, y, A );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  ldim_A   = FLA_Obj_ldim( A );

  m_x      = FLA_Obj_length( x );
  ldim_x   = FLA_Obj_ldim( x );

  m_y      = FLA_Obj_length( y );
  ldim_y   = FLA_Obj_ldim( y );

  inc_x    = ( m_x == 1 ? ldim_x : 1 );
  inc_y    = ( m_y == 1 ? ldim_y : 1 );

  if ( conjxy == FLA_CONJUGATE )
  {
    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, x, &x_copy );
    FLA_Copy_t( FLA_CONJ_NO_TRANSPOSE, x, x_copy );

    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, y, &y_copy );
    FLA_Copy_t( FLA_CONJ_NO_TRANSPOSE, y, y_copy );

    m_x      = FLA_Obj_length( x_copy );
    ldim_x   = FLA_Obj_ldim( x_copy );
    inc_x    = ( m_x == 1 ? ldim_x : 1 );

    m_y      = FLA_Obj_length( y_copy );
    ldim_y   = FLA_Obj_ldim( y_copy );
    inc_y    = ( m_y == 1 ? ldim_y : 1 );
  }
  else
  {
    x_copy = x;
    y_copy = y;
  }

  FLA_Param_map_to_blas_uplo( uplo, &blas_uplo );


  switch( datatype ){

  case FLA_COMPLEX:
  {
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_x     = ( scomplex * ) FLA_COMPLEX_PTR( x_copy );
    scomplex *buff_y     = ( scomplex * ) FLA_COMPLEX_PTR( y_copy );
    scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
    scomplex  alphac;

    alphac       = *buff_alpha;
    alphac.imag *= -1.0F;

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_cher2( cblas_order,
                 blas_uplo,
                 m_A,
                 &alphac,
                 buff_x, inc_x,
                 buff_y, inc_y,
                 buff_A, ldim_A ); 
#else
    FLA_C2F( cher2 )( &blas_uplo,
                      &m_A,
                      &alphac,
                      buff_x, &inc_x,
                      buff_y, &inc_y,
                      buff_A, &ldim_A ); 
#endif
    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_x     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( x_copy );
    dcomplex *buff_y     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( y_copy );
    dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
    dcomplex  alphac;

    alphac       = *buff_alpha;
    alphac.imag *= -1.0;

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_zher2( cblas_order,
                 blas_uplo,
                 m_A,
                 &alphac,
                 buff_x, inc_x,
                 buff_y, inc_y,
                 buff_A, ldim_A ); 
#else
    FLA_C2F( zher2 )( &blas_uplo,
                      &m_A,
                      &alphac,
                      buff_x, &inc_x,
                      buff_y, &inc_y,
                      buff_A, &ldim_A ); 
#endif
    break;
  }

  }
  
  if ( conjxy == FLA_CONJUGATE )
  {
    FLA_Obj_free( &x_copy );
    FLA_Obj_free( &y_copy );
  }

  return FLA_SUCCESS;
}


void FLA_F2C( fla_her2_c_f ) 
     ( F_INT *uplo, F_INT *conjxy, F_INT *alpha, F_INT *x, F_INT *y, F_INT *A, F_INT *IERROR )
{
  *IERROR = FLA_Her2_c( *( ( FLA_Uplo * ) uplo   ), 
                        *( ( FLA_Conj * ) conjxy ), 
                        *( ( FLA_Obj  * ) alpha  ), 
                        *( ( FLA_Obj  * ) x      ),
                        *( ( FLA_Obj  * ) y      ),
                        *( ( FLA_Obj  * ) A      ) );
}

