/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Hemm_external( FLA_Side side, FLA_Uplo uplo, FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  FLA_Datatype datatype;
  int          ldim_A;
  int          ldim_B; 
  int          m_C, n_C, ldim_C;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
  CBLAS_ORDER  cblas_order = CblasColMajor;
  CBLAS_SIDE   blas_side;
  CBLAS_UPLO   blas_uplo;
#else
  char         blas_side;
  char         blas_uplo; 
#endif

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Hemm_check( side, uplo, alpha, A, B, beta, C );
#endif

  if ( FLA_Obj_has_zero_dim( C ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  ldim_A   = FLA_Obj_ldim( A );

  ldim_B   = FLA_Obj_ldim( B );

  m_C      = FLA_Obj_length( C );
  n_C      = FLA_Obj_width( C );
  ldim_C   = FLA_Obj_ldim( C );

  FLA_Param_map_to_blas_side( side, &blas_side );
  FLA_Param_map_to_blas_uplo( uplo, &blas_uplo );
  

  switch( datatype ){

  case FLA_COMPLEX:
  {
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_B     = ( scomplex * ) FLA_COMPLEX_PTR( B );
    scomplex *buff_C     = ( scomplex * ) FLA_COMPLEX_PTR( C );
    scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
    scomplex *buff_beta  = ( scomplex * ) FLA_COMPLEX_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_chemm( cblas_order,
                 blas_side,
                 blas_uplo,
                 m_C,
                 n_C,
                 buff_alpha,
                 buff_A, ldim_A, 
                 buff_B, ldim_B, 
                 buff_beta,  
                 buff_C, ldim_C );
#else
    FLA_C2F( chemm )( &blas_side,
                      &blas_uplo,
                      &m_C,
                      &n_C,
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_B, &ldim_B, 
                      buff_beta,  
                      buff_C, &ldim_C );
#endif
    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_B     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( B );
    dcomplex *buff_C     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( C );
    dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
    dcomplex *buff_beta  = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_zhemm( cblas_order,
                 blas_side,
                 blas_uplo,
                 m_C,
                 n_C,
                 buff_alpha,
                 buff_A, ldim_A,
                 buff_B, ldim_B,
                 buff_beta,
                 buff_C, ldim_C );
#else
    FLA_C2F( zhemm )( &blas_side,
                      &blas_uplo,
                      &m_C,
                      &n_C,
                      buff_alpha,
                      buff_A, &ldim_A,
                      buff_B, &ldim_B,
                      buff_beta,
                      buff_C, &ldim_C );
#endif
    break;
  }

  }
 
  return FLA_SUCCESS;
}

void FLA_F2C( fla_hemm_external_f ) 
     ( F_INT *side, F_INT *uplo, F_INT *alpha, F_INT *A, F_INT *B, F_INT *beta, F_INT *C, F_INT *IERROR )
{
  *IERROR = FLA_Hemm_external( *( ( FLA_Side * ) side  ),
                               *( ( FLA_Uplo * ) uplo  ),
                               *( ( FLA_Obj  * ) alpha ),
                               *( ( FLA_Obj  * ) A     ),
                               *( ( FLA_Obj  * ) B     ),
                               *( ( FLA_Obj  * ) beta  ),
                               *( ( FLA_Obj  * ) C     ) );
}

FLA_Error FLA_Hemm_ll_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Hemm_external( FLA_LEFT, FLA_LOWER_TRIANGULAR, alpha, A, B, beta, C );
}

FLA_Error FLA_Hemm_lu_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Hemm_external( FLA_LEFT, FLA_UPPER_TRIANGULAR, alpha, A, B, beta, C );
}

FLA_Error FLA_Hemm_rl_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Hemm_external( FLA_RIGHT, FLA_LOWER_TRIANGULAR, alpha, A, B, beta, C );
}

FLA_Error FLA_Hemm_ru_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Hemm_external( FLA_RIGHT, FLA_UPPER_TRIANGULAR, alpha, A, B, beta, C );
}

