/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Syrk_external( FLA_Uplo uplo, FLA_Trans trans, FLA_Obj alpha, FLA_Obj A, FLA_Obj beta, FLA_Obj C )
{
  FLA_Datatype datatype;
  int          k_A;
  int          m_A, n_A, ldim_A;
  int          m_C, ldim_C;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
  CBLAS_ORDER     cblas_order = CblasColMajor;
  CBLAS_UPLO      blas_uplo;
  CBLAS_TRANSPOSE blas_trans;
#else
  char         blas_uplo; 
  char         blas_trans;
#endif

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Syrk_check( uplo, trans, alpha, A, beta, C );
#endif

  if ( FLA_Obj_has_zero_dim( C ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  m_C      = FLA_Obj_length( C );
  ldim_C   = FLA_Obj_ldim( C );

  if ( trans == FLA_NO_TRANSPOSE )
    k_A = n_A;
  else
    k_A = m_A;

  FLA_Param_map_to_blas_uplo( uplo, &blas_uplo );
  FLA_Param_map_to_blas_trans( trans, &blas_trans );


  switch( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_C     = ( float * ) FLA_FLOAT_PTR( C );
    float *buff_alpha = ( float * ) FLA_FLOAT_PTR( alpha );
    float *buff_beta  = ( float * ) FLA_FLOAT_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_ssyrk( cblas_order,
                 blas_uplo, 
                 blas_trans,
                 m_C,
                 k_A, 
                 *buff_alpha,
                 buff_A, ldim_A, 
                 *buff_beta,  
                 buff_C, ldim_C );
#else
    FLA_C2F( ssyrk )( &blas_uplo, 
                      &blas_trans,
                      &m_C,
                      &k_A, 
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_beta,  
                      buff_C, &ldim_C );
#endif
    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_C     = ( double * ) FLA_DOUBLE_PTR( C );
    double *buff_alpha = ( double * ) FLA_DOUBLE_PTR( alpha );
    double *buff_beta  = ( double * ) FLA_DOUBLE_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_dsyrk( cblas_order,
                 blas_uplo, 
                 blas_trans,
                 m_C,
                 k_A, 
                 *buff_alpha,
                 buff_A, ldim_A, 
                 *buff_beta,  
                 buff_C, ldim_C );
#else
    FLA_C2F( dsyrk )( &blas_uplo, 
                      &blas_trans,
                      &m_C,
                      &k_A, 
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_beta,  
                      buff_C, &ldim_C );
#endif
    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_C     = ( scomplex * ) FLA_COMPLEX_PTR( C );
    scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
    scomplex *buff_beta  = ( scomplex * ) FLA_COMPLEX_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_csyrk( cblas_order,
                 blas_uplo, 
                 blas_trans,
                 m_C,
                 k_A, 
                 buff_alpha,
                 buff_A, ldim_A, 
                 buff_beta,  
                 buff_C, ldim_C );
#else
    FLA_C2F( csyrk )( &blas_uplo, 
                      &blas_trans,
                      &m_C,
                      &k_A, 
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_beta,  
                      buff_C, &ldim_C );
#endif
    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_C     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( C );
    dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
    dcomplex *buff_beta  = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_zsyrk( cblas_order,
                 blas_uplo, 
                 blas_trans,
                 m_C,
                 k_A, 
                 buff_alpha,
                 buff_A, ldim_A, 
                 buff_beta,  
                 buff_C, ldim_C );
#else
    FLA_C2F( zsyrk )( &blas_uplo, 
                      &blas_trans,
                      &m_C,
                      &k_A, 
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_beta,  
                      buff_C, &ldim_C );
#endif
    break;
  }

  }
  
  return FLA_SUCCESS;
}

void FLA_F2C( fla_syrk_external_f ) 
     ( F_INT *uplo, F_INT *trans, F_INT *alpha, F_INT *A, F_INT *beta, F_INT *C, F_INT *IERROR )
{
  *IERROR = FLA_Syrk_external( *( ( FLA_Uplo  * ) uplo  ),
                               *( ( FLA_Trans * ) trans ), 
                               *( ( FLA_Obj   * ) alpha ),
                               *( ( FLA_Obj   * ) A     ),
                               *( ( FLA_Obj   * ) beta  ),
                               *( ( FLA_Obj   * ) C     ) );
}

FLA_Error FLA_Syrk_ln_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Syrk_external( FLA_LOWER_TRIANGULAR, FLA_NO_TRANSPOSE, alpha, A, beta, C );
}

FLA_Error FLA_Syrk_lt_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Syrk_external( FLA_LOWER_TRIANGULAR, FLA_TRANSPOSE, alpha, A, beta, C );
}

FLA_Error FLA_Syrk_un_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Syrk_external( FLA_UPPER_TRIANGULAR, FLA_NO_TRANSPOSE, alpha, A, beta, C );
}

FLA_Error FLA_Syrk_ut_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Syrk_external( FLA_UPPER_TRIANGULAR, FLA_TRANSPOSE, alpha, A, beta, C );
}

