/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Accum_S_UT( FLA_Direct direct, FLA_Store storev, FLA_Obj A, FLA_Obj t, FLA_Obj S )
{
  FLA_Datatype datatype;
  int          m_H, ldim_A;
  int          k_S, ldim_S;
  char         lapack_direct; 
  char         lapack_storev;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() )
    FLA_Accum_S_check( direct, storev, A, t, S );
#endif

  datatype = FLA_Obj_datatype( A );

  if ( storev == FLA_COLUMNWISE )
    m_H    = FLA_Obj_length( A );
  else
    m_H    = FLA_Obj_width( A );

  k_S      = FLA_Obj_width( S );
  ldim_A   = FLA_Obj_ldim( A );
  ldim_S   = FLA_Obj_ldim( S );

  FLA_Param_map_to_lapack_direct( direct, &lapack_direct );
  FLA_Param_map_to_lapack_storev( storev, &lapack_storev );


  switch( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_t = ( float * ) FLA_FLOAT_PTR( t );
    float *buff_S = ( float * ) FLA_FLOAT_PTR( S );

    FLA_C2F( slarftut )( &lapack_direct,
                         &lapack_storev,
                         &m_H,
                         &k_S,
                         buff_A, &ldim_A,
                         buff_t,
                         buff_S, &ldim_S );

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_t = ( double * ) FLA_DOUBLE_PTR( t );
    double *buff_S = ( double * ) FLA_DOUBLE_PTR( S );

    FLA_C2F( dlarftut )( &lapack_direct,
                         &lapack_storev,
                         &m_H,
                         &k_S,
                         buff_A, &ldim_A,
                         buff_t,
                         buff_S, &ldim_S );

    break;
  } 

  case FLA_COMPLEX:
  {
    scomplex *buff_A = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_t = ( scomplex * ) FLA_COMPLEX_PTR( t );
    scomplex *buff_S = ( scomplex * ) FLA_COMPLEX_PTR( S );

    FLA_C2F( clarftut )( &lapack_direct,
                         &lapack_storev,
                         &m_H,
                         &k_S,
                         buff_A, &ldim_A,
                         buff_t,
                         buff_S, &ldim_S );

    break;
  } 

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_t = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( t );
    dcomplex *buff_S = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( S );

    FLA_C2F( zlarftut )( &lapack_direct,
                         &lapack_storev,
                         &m_H,
                         &k_S,
                         buff_A, &ldim_A,
                         buff_t,
                         buff_S, &ldim_S );

    break;
  } 

  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_accum_s_ut_f )
     ( F_INT *direct, F_INT *storev, F_INT *A, F_INT *t, F_INT *S, F_INT *IERROR )
{
  *IERROR = FLA_Accum_S_UT( *( ( FLA_Direct * ) direct ),
                            *( ( FLA_Store  * ) storev ),
                            *( ( FLA_Obj    * ) A      ), 
                            *( ( FLA_Obj    * ) t      ), 
                            *( ( FLA_Obj    * ) S      ) );
}

