/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

extern int* fla_lu_pivots_temp;

FLA_Error FLA_Swap_rows( FLA_Obj A, int k1, int k2, FLA_Obj ipiv )
{
  FLA_Datatype datatype;
  int          n_A, ldim_A;
  int          m_ipiv, n_ipiv, ldim_ipiv;
  int          inc_ipiv;
  int*         buff_ipiv;
  int          i;
  int          k1_1, k2_1;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() )
    FLA_Swap_rows_check( A, k1, k2, ipiv );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  n_A       = FLA_Obj_width( A );
  ldim_A    = FLA_Obj_ldim( A );

  m_ipiv    = FLA_Obj_length( ipiv );
  n_ipiv    = FLA_Obj_width( ipiv );
  ldim_ipiv = FLA_Obj_ldim( ipiv );

  buff_ipiv = FLA_INT_PTR( ipiv );

  if      ( m_ipiv == 1 ) inc_ipiv = ldim_ipiv;
  else if ( n_ipiv == 1 ) inc_ipiv = 1;

  // Convert zero-based FLAME indices to one-based for LAPACK.
  k1_1      = k1 + 1;
  k2_1      = k2 + 1;

  // Translate FLAME pivot indices to LAPACK-compatible indices. It is
  // important to note that this conversion, unlike the one done by
  // FLA_Shift_pivots_to(), is NOT in-place, but rather done separately
  // in an internal temporary libflame buffer.
  for ( i = 0; i < m_ipiv; i++ )
  {
    fla_lu_pivots_temp[ i ] = buff_ipiv[ i ] + i + 1;
  }

  switch ( datatype ){

  case FLA_FLOAT:
  {
    float* buff_A = ( float * ) FLA_FLOAT_PTR( A );

    FLA_C2F( slaswp )( &n_A,
                       buff_A, &ldim_A,
                       &k1_1, 
                       &k2_1,
                       fla_lu_pivots_temp,
                       &inc_ipiv );
    break;
  }

  case FLA_DOUBLE:
  {
    double* buff_A = ( double * ) FLA_DOUBLE_PTR( A );

    FLA_C2F( dlaswp )( &n_A,
                       buff_A, &ldim_A,
                       &k1_1, 
                       &k2_1,
                       fla_lu_pivots_temp,
                       &inc_ipiv );
    break;
  }

  case FLA_COMPLEX:
  {
    scomplex* buff_A = ( scomplex * ) FLA_COMPLEX_PTR( A );

    FLA_C2F( claswp )( &n_A,
                       buff_A, &ldim_A,
                       &k1_1, 
                       &k2_1,
                       fla_lu_pivots_temp,
                       &inc_ipiv );
    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex* buff_A = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );

    FLA_C2F( zlaswp )( &n_A,
                       buff_A, &ldim_A,
                       &k1_1, 
                       &k2_1,
                       fla_lu_pivots_temp,
                       &inc_ipiv );
    break;
  }

  }
  
  return FLA_SUCCESS;
}

void FLA_F2C( fla_swap_rows_f ) 
     ( F_INT *A, F_INT* k1, F_INT* k2, F_INT *ipiv, F_INT *IERROR )
{
  *IERROR = FLA_Swap_rows( *( ( FLA_Obj * ) A   ),
                           *( ( int     * ) k1  ), 
                           *( ( int     * ) k2  ), 
                           *( ( FLA_Obj * ) ipiv) );
}

