/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Local_asum_check( PLA_Obj x, 
				PLA_Obj asum_x )
/*----------------------------------------------------------------------------
  
   PLA_Local_asum_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( x, asum_x, "PLA_Local_asum: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_axpy_check( PLA_Obj alpha, 
				PLA_Obj A, 
				PLA_Obj B )
/*----------------------------------------------------------------------------

   PLA_Local_axpy_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( alpha, B, "PLA_Local_axpy: objects not on same processor" );
      PLA_Check_consistent_owners( A,     B, "PLA_Local_axpy: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_axpy_x_check( PLA_Trans trans,
				  PLA_Obj   alpha, 
				  PLA_Obj   A, 
				  PLA_Obj   B )
/*----------------------------------------------------------------------------

   PLA_Local_axpy_x_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( alpha, B, "PLA_Local_axpy_x: objects not on same processor" );
      PLA_Check_consistent_owners( A,     B, "PLA_Local_axpy_x: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_axpy_y_check( PLA_Trans trans,
				  PLA_Obj   alpha, 
				  PLA_Obj   A, 
				  PLA_Obj   beta, 
				  PLA_Obj   B )
/*----------------------------------------------------------------------------

   PLA_Local_axpy_y_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( alpha, B, "PLA_Local_axpy_y: objects not on same processor" );
      PLA_Check_consistent_owners( A,     B, "PLA_Local_axpy_y: objects not on same processor" );
      PLA_Check_consistent_owners( beta,  B, "PLA_Local_axpy_y: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_copy_check( PLA_Obj A,
				PLA_Obj B )
/*----------------------------------------------------------------------------

   PLA_Local_copy_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( A, B, "PLA_Local_copy: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_copy_x_check( PLA_Trans trans,
				  PLA_Obj   A,
				  PLA_Obj   B )
/*----------------------------------------------------------------------------

   PLA_Local_copy_x_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( A, B, "PLA_Local_copy_x: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_dot_check( PLA_Obj x,
			       PLA_Obj y,
			       PLA_Obj rho )
/*----------------------------------------------------------------------------

   PLA_Local_dot_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( x, rho, "PLA_Local_dot: objects not on same processor" );
      PLA_Check_consistent_owners( y, rho, "PLA_Local_dot: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_dot_x_check( PLA_Obj alpha,
				 PLA_Obj x,
				 PLA_Obj y,
				 PLA_Obj beta,
				 PLA_Obj rho )
/*----------------------------------------------------------------------------

   PLA_Local_dot_x_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( alpha, rho, "PLA_Local_dot_x: objects not on same processor" );
      PLA_Check_consistent_owners( x,     rho, "PLA_Local_dot_x: objects not on same processor" );
      PLA_Check_consistent_owners( y,     rho, "PLA_Local_dot_x: objects not on same processor" );
      PLA_Check_consistent_owners( beta,  rho, "PLA_Local_dot_x: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_dotc_check( PLA_Obj x,
				PLA_Obj y,
				PLA_Obj rho )
/*----------------------------------------------------------------------------

   PLA_Local_dotc_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( x, rho, "PLA_Local_dotc: objects not on same processor" );
      PLA_Check_consistent_owners( y, rho, "PLA_Local_dotc: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_iamax_check( PLA_Obj x,
				 PLA_Obj index )
/*----------------------------------------------------------------------------

   PLA_Local_iamax_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( x, index, "PLA_Local_iamax: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_inv_scal_check( PLA_Obj alpha,
				    PLA_Obj A )
/*----------------------------------------------------------------------------

   PLA_Local_inv_scal_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( alpha, A, "PLA_Local_inv_scal: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_nrm2_check( PLA_Obj x,
				PLA_Obj nrm_x )
/*----------------------------------------------------------------------------

   PLA_Local_nrm2_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( x, nrm_x, "PLA_Local_nrm2: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_scal_check( PLA_Obj alpha,
				PLA_Obj A )
/*----------------------------------------------------------------------------

   PLA_Local_scal_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( alpha, A, "PLA_Local_scal: objects not on same processor" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_scal_x_check( PLA_Uplo uplo,
				  PLA_Obj  alpha,
				  PLA_Obj  A )
/*----------------------------------------------------------------------------

   PLA_Local_scal_x_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( alpha, A, "PLA_Local_scal_x: objects not on same processor" );
   }
   
   return PLA_SUCCESS; 
}


PLA_Error PLA_Local_swap_check( PLA_Obj A,
				PLA_Obj B )
/*----------------------------------------------------------------------------

   PLA_Local_swap_check

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_consistent_owners( A, B, "PLA_Local_swap: objects not on same processor" );
   }
   
   return PLA_SUCCESS; 
}
