/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Obj_create_check( PLA_Objtype  objtype,
				PLA_Datatype datatype, 
				int          m, 
				int          n, 
				int          alignM, 
				int          alignN, 
				int          ownerRow, 
				int          ownerCol, 
				PLA_Temp     templ )
/*----------------------------------------------------------------------------

Purpose :  Check creating an object

IN      objtype      type of object
IN      datatype     datatype
IN      m            global length
IN      n            global width
IN      alignM       alignment to template
IN      alignN       alignment to template
IN      ownerRow     owner of the row
IN      ownerCol     owner of the column
IN      templ        template

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_valid_objtype ( objtype,  "Invalid objtype" );
      PLA_Check_valid_datatype( datatype, "Invalid datatype" );
      
      if ( templ == NULL )
	 PLA_Abort( "Template is NULL", __LINE__, __FILE__ );
      
      if ( m < 1 || n < 1 )
	 PLA_Abort( "Empty object", __LINE__, __FILE__ );
      
      if ( alignM < 0 || alignN < 0 )
	 PLA_Abort( "Malformed alignment", __LINE__, __FILE__ );
      
      if ( ownerRow != PLA_ALL_ROWS && 
	   ownerRow < 0             &&
	   ownerRow >= PLA_Temp_comm_col_size( templ ) )
	 PLA_Abort( "Malformed owning row", __LINE__, __FILE__ );
      
      if ( ownerCol != PLA_ALL_COLS && 
	   ownerCol < 0             &&
	   ownerCol >= PLA_Temp_comm_row_size( templ ) )
	 PLA_Abort( "Malformed owning column", __LINE__, __FILE__ );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_set_element_check( PLA_Obj alpha )
/*----------------------------------------------------------------------------

Purpose :  Check object set element

IN      alpha      object must be a scalar

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_if_scalar( alpha, "PLA_Obj_set_element: alpha must be 1x1" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_add_to_element_check( PLA_Obj alpha )
/*----------------------------------------------------------------------------

Purpose :  Check object add to element

IN      alpha      object must be a scalar

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_if_scalar( alpha, "PLA_Obj_add_to_element: alpha must be 1x1" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_get_local_contents_check( PLA_Trans trans,
					    PLA_Obj   obj,
					    int       m,
					    int       n,
					    void      *A,
					    int       ldA )
/*----------------------------------------------------------------------------

Purpose :   Check getting local contents from object to local buffer

IN      trans        option to transpose
IN      obj          object
IN      m            length to extract
IN      n            width to extract
IN      A            buffer to extract data
IN      ldA          leading dimension of buffer

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      if ( trans == PLA_NO_TRANSPOSE )
      {
	 if ( PLA_Obj_length( obj ) != m || 
	      PLA_Obj_width ( obj ) != n )
	    PLA_Abort( "Dimensions do not match", __LINE__, __FILE__ );
      }
      else if ( trans == PLA_TRANSPOSE )
      {
	 if ( PLA_Obj_length( obj ) != n || 
	      PLA_Obj_width ( obj ) != m )
	    PLA_Abort( "Dimensions do not match", __LINE__, __FILE__ );
      }
      else
	 PLA_Abort( "Undefined transpose option", __LINE__, __FILE__ );
   }
   
   return PLA_SUCCESS;
}
