/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#ifndef PLA_MAIN_PROTOTYPES_H
#define PLA_MAIN_PROTOTYPES_H


/**************** PLA_Error.c ****************/

int PLA_printf( char *string, ... );
void PLA_Abort( char *s1, int line, char *file );


/**************** PLA_Memory.c ****************/

void * PLA_malloc( int size );
void PLA_free( void *buffer );


/**************** PLA_Check.c ****************/

PLA_Bool PLA_Error_check( );
void PLA_Error_check_set( PLA_Bool val );
PLA_Bool PLA_Check_parameters( );
void PLA_Check_parameters_set( PLA_Bool val );

void PLA_Check_valid_side( PLA_Side side, char *message );
void PLA_Check_valid_quadrant( PLA_Quadrant quadrant, char *message );
void PLA_Check_valid_uplo( PLA_Uplo uplo, char *message );
void PLA_Check_valid_trans( PLA_Trans trans, char *message );
void PLA_Check_valid_diag( PLA_Diag diag, char *message );
void PLA_Check_valid_conj( PLA_Conj conj, char *message );
void PLA_Check_valid_datatype( PLA_Datatype datatype, char *message );
void PLA_Check_real_object( PLA_Obj A, char *message );
void PLA_Check_complex_object( PLA_Obj A, char *message );
void PLA_Check_if_square( PLA_Obj A, char *message );
void PLA_Check_if_scalar( PLA_Obj alpha, char *message );
void PLA_Check_if_vector( PLA_Obj A, char *message );
void PLA_Check_consistent_datatype( PLA_Datatype datatype, PLA_Obj A, char *message );
void PLA_Check_consistent_dims( PLA_Trans trans, PLA_Obj A, PLA_Obj B, char *message );

void PLA_Check_valid_objtype( PLA_Objtype objtype, char *message );
void PLA_Check_valid_op( PLA_Op op, char *message );
void PLA_Check_if_local_object( PLA_Obj A, char *message );
void PLA_Check_consistent_owners( PLA_Obj A, PLA_Obj B, char *message );
void PLA_Check_consistent_templates( PLA_Obj A, PLA_Obj B, char *message );


/**************** PLA_Init.c ****************/

PLA_Error PLA_Init( );
PLA_Error PLA_Finalize( );
PLA_Bool PLA_Initialized( );


/**************** PLA_Temp.c ****************/

PLA_Error PLA_Temp_create( MPI_Comm comm, int nprows, int npcols, int mb, PLA_Distr colDistr, int nb, PLA_Distr rowDistr, PLA_Temp *templ );
PLA_Error PLA_Temp_free( PLA_Temp *templ );
PLA_Error PLA_Temp_comm_all_info( PLA_Temp templ, MPI_Comm *comm, int *me, int *nprocs );
MPI_Comm PLA_Temp_comm_all( PLA_Temp templ );
int PLA_Temp_comm_all_rank( PLA_Temp templ );
int PLA_Temp_comm_all_size( PLA_Temp templ ); 
PLA_Error PLA_Temp_comm_row_info( PLA_Temp templ, MPI_Comm *commRow, int *mycol, int *npcols );
MPI_Comm PLA_Temp_comm_row( PLA_Temp templ );
int PLA_Temp_comm_row_rank( PLA_Temp templ );
int PLA_Temp_comm_row_size( PLA_Temp templ );
PLA_Error PLA_Temp_comm_col_info( PLA_Temp templ, MPI_Comm *commCol, int *myrow, int *nprows );
MPI_Comm PLA_Temp_comm_col( PLA_Temp templ );
int PLA_Temp_comm_col_rank( PLA_Temp templ );
int PLA_Temp_comm_col_size( PLA_Temp templ );
int PLA_Temp_mb( PLA_Temp templ );
PLA_Distr PLA_Temp_col_distr( PLA_Temp templ );
int PLA_Temp_nb( PLA_Temp templ );
PLA_Distr PLA_Temp_row_distr( PLA_Temp templ );


/**************** PLA_Obj.c ****************/

PLA_Error PLA_Matrix_create( PLA_Datatype datatype, int m, int n, int alignM, int alignN, PLA_Temp templ, PLA_Obj *obj );
PLA_Error PLA_Mrowvector_create( PLA_Datatype datatype, int m, int n, int alignN, PLA_Temp templ, PLA_Obj *obj );
PLA_Error PLA_Mcolvector_create( PLA_Datatype datatype, int m, int n, int alignM, PLA_Temp templ, PLA_Obj *obj );
PLA_Error PLA_Mscalar_create( PLA_Datatype datatype, int m, int n, int ownerRow, int ownerCol, PLA_Temp templ, PLA_Obj *obj );
PLA_Error PLA_Pmrowvector_create( PLA_Datatype datatype, int m, int n, int alignN, int ownerRow, PLA_Temp templ, PLA_Obj *obj );
PLA_Error PLA_Pmcolvector_create( PLA_Datatype datatype, int m, int n, int alignM, int ownerCol, PLA_Temp templ, PLA_Obj *obj );
PLA_Error PLA_Constant_complex_create( double real, double imag, PLA_Obj *obj );
PLA_Error PLA_Constant_create( double real, PLA_Obj *obj );
PLA_Error PLA_Matrix_create_conf_to( PLA_Trans trans, PLA_Obj obj, PLA_Obj *newObj );
PLA_Error PLA_Mrowvector_create_conf_to( PLA_Trans trans, PLA_Obj obj, int numVects, PLA_Obj *newObj );
PLA_Error PLA_Mcolvector_create_conf_to( PLA_Trans trans, PLA_Obj obj, int numVects, PLA_Obj *newObj );
PLA_Error PLA_Mscalar_create_conf_to( PLA_Trans trans, PLA_Obj obj, int ownerRow, int ownerCol, PLA_Obj *newObj );
PLA_Error PLA_Pmrowvector_create_conf_to( PLA_Trans trans, PLA_Obj obj, int numVects, int ownerRow, PLA_Obj *newObj );
PLA_Error PLA_Pmcolvector_create_conf_to( PLA_Trans trans, PLA_Obj obj, int numVects, int ownerCol, PLA_Obj *newObj );
PLA_Error PLA_Obj_free( PLA_Obj *obj ); 
PLA_Objtype PLA_Obj_objtype( PLA_Obj obj );
PLA_Datatype PLA_Obj_datatype( PLA_Obj obj );
PLA_Temp PLA_Obj_template( PLA_Obj obj );
PLA_Error PLA_Obj_global_info( PLA_Obj obj, int *m, int *n, int *alignM, int *alignN, int *ownerRow, int *ownerCol );
int PLA_Obj_length( PLA_Obj obj );
int PLA_Obj_width( PLA_Obj obj );
int PLA_Obj_align_m( PLA_Obj obj );
int PLA_Obj_align_n( PLA_Obj obj );
int PLA_Obj_owner_row( PLA_Obj obj );
int PLA_Obj_owner_col( PLA_Obj obj );
int PLA_Obj_local_length( PLA_Obj obj );
int PLA_Obj_local_width( PLA_Obj obj );
int PLA_Obj_local_ldim( PLA_Obj obj );
PLA_Error PLA_Obj_local_info( PLA_Obj obj, int *localm, int *localn, int *ldim, void **buffer );

PLA_Bool PLA_Obj_is_valid_state( PLA_Obj obj );


/**************** PLA_Misc.c ****************/

PLA_Error PLA_Obj_set_to( PLA_Obj obj, void *value );
PLA_Error PLA_Obj_set_to_minus_one( PLA_Obj obj );
PLA_Error PLA_Obj_set_to_zero( PLA_Obj obj );
PLA_Error PLA_Obj_set_to_one( PLA_Obj obj );
PLA_Error PLA_Obj_set_element( PLA_Obj alpha, void *value );
PLA_Error PLA_Obj_add_to_element( PLA_Obj alpha, void *value );
PLA_Error PLA_Obj_add_to_diagonal( PLA_Obj A, void *value );

PLA_Error PLA_Obj_get_local_contents( PLA_Trans trans, PLA_Obj obj, int m, int n, void *A, int ldA );

int PLA_Obj_owner( PLA_Obj obj, int *myrow, int *mycol );
int PLA_Obj_owner_all( PLA_Obj obj );
PLA_Bool PLA_Obj_is_owner( PLA_Obj obj );

int PLA_Obj_split_size( PLA_Obj obj, PLA_Side side );


/**************** PLA_View.c ****************/

PLA_Error PLA_Part_2x2( PLA_Obj A,  PLA_Obj *ATL, PLA_Obj *ATR, 
		                    PLA_Obj *ABL, PLA_Obj *ABR, 
			int mb, int nb, PLA_Quadrant quadrant );
PLA_Error PLA_Merge_2x2( PLA_Obj ATL, PLA_Obj ATR, 
			 PLA_Obj ABL, PLA_Obj ABR, 
			 PLA_Obj *A );
PLA_Error PLA_Repart_2x2_to_3x3( PLA_Obj ATL, PLA_Obj ATR,  PLA_Obj *A00, PLA_Obj *A01, PLA_Obj *A02,
				                            PLA_Obj *A10, PLA_Obj *A11, PLA_Obj *A12,
				 PLA_Obj ABL, PLA_Obj ABR,  PLA_Obj *A20, PLA_Obj *A21, PLA_Obj *A22,
				 int mb, int nb, PLA_Quadrant quadrant );
PLA_Error PLA_Cont_with_3x3_to_2x2( PLA_Obj *ATL, PLA_Obj *ATR,  PLA_Obj A00, PLA_Obj A01, PLA_Obj A02,
				                                 PLA_Obj A10, PLA_Obj A11, PLA_Obj A12,
				    PLA_Obj *ABL, PLA_Obj *ABR,  PLA_Obj A20, PLA_Obj A21, PLA_Obj A22,
				    PLA_Quadrant quadrant );
PLA_Error PLA_Part_2x1( PLA_Obj A,  PLA_Obj *AT,
			            PLA_Obj *AB,
			int mb, PLA_Side side );
PLA_Error PLA_Merge_2x1( PLA_Obj AT,
			 PLA_Obj AB,
			 PLA_Obj *A );
PLA_Error PLA_Repart_2x1_to_3x1( PLA_Obj AT,  PLA_Obj *A0,
				              PLA_Obj *A1,
				 PLA_Obj AB,  PLA_Obj *A2,
				 int mb, PLA_Side side );
PLA_Error PLA_Cont_with_3x1_to_2x1( PLA_Obj *AT,  PLA_Obj A0,
			                          PLA_Obj A1,
				    PLA_Obj *AB,  PLA_Obj A2,
				    PLA_Side side );
PLA_Error PLA_Part_1x2( PLA_Obj A,
			PLA_Obj *AL, PLA_Obj *AR,
			int nb, PLA_Side );
PLA_Error PLA_Merge_1x2( PLA_Obj AL, PLA_Obj AR,
			 PLA_Obj *A );
PLA_Error PLA_Repart_1x2_to_1x3( PLA_Obj  AL,              PLA_Obj  AR,
				 PLA_Obj *A0, PLA_Obj *A1, PLA_Obj *A2,
				 int, PLA_Side side );
PLA_Error PLA_Cont_with_1x3_to_1x2( PLA_Obj *AL,             PLA_Obj *AR,
				    PLA_Obj  A0, PLA_Obj A1, PLA_Obj  A2,
				    PLA_Side side );


/**************** PLA_Check_main.c ****************/

PLA_Error PLA_Obj_create_check( PLA_Objtype objtype, PLA_Datatype datatype, int m, int n, int alignM, int alignN, int ownerRow, int ownerCol, PLA_Temp templ );
PLA_Error PLA_Obj_set_element_check( PLA_Obj alpha );
PLA_Error PLA_Obj_add_to_element_check( PLA_Obj alpha );
PLA_Error PLA_Obj_get_local_contents_check( PLA_Trans trans, PLA_Obj obj, int m, int n, void *A, int ldA );


#endif
