/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Bool PLA_initialized = FALSE;

extern int PLA_memcounter;


// statically create a global template
PLA_Template PLA_TEMPLATE_WORLD;

// global pointer to global template
PLA_Temp PLA_TEMP_WORLD = &PLA_TEMPLATE_WORLD;


// global objects of constants
PLA_Obj PLA_MINUS_ONE;
PLA_Obj PLA_ZERO;
PLA_Obj PLA_ONE;
PLA_Obj PLA_TWO;
PLA_Obj PLA_i;


// global constants
float
   PLA_s_minus_one,
   PLA_s_zero,
   PLA_s_one;
double
   PLA_d_minus_one,
   PLA_d_zero,
   PLA_d_one;
complex
   PLA_c_minus_one,
   PLA_c_zero,
   PLA_c_one;
dcomplex
   PLA_z_minus_one,
   PLA_z_zero,
   PLA_z_one;
int
   PLA_i_minus_one,
   PLA_i_zero,
   PLA_i_one;


PLA_Error PLA_Temp_create_without_buffer( MPI_Comm, 
					  int, int, 
					  int, PLA_Distr, 
					  int, PLA_Distr, 
					  PLA_Temp );

PLA_Error PLA_Temp_free_without_buffer( PLA_Temp * );


PLA_Error PLA_Init( )
/*----------------------------------------------------------------------------

Purpose :  Initialize PLAPACK2e
         
----------------------------------------------------------------------------*/
{
   if ( PLA_initialized )
      return PLA_FAILURE;
   
   PLA_initialized = TRUE;
   PLA_memcounter  = 0; 

   PLA_setup_enter( );
   
   PLA_Temp_create_without_buffer( MPI_COMM_WORLD,
				   1, 1,          
				   PLA_TEMP_WORLD_NB,
				   PLA_ROW_MAJOR,    
				   PLA_TEMP_WORLD_NB,
				   PLA_COL_MAJOR,    
				   PLA_TEMP_WORLD ); 
   
   PLA_Constant_create( -1.0, &PLA_MINUS_ONE );
   PLA_Constant_create(  0.0, &PLA_ZERO );
   PLA_Constant_create(  1.0, &PLA_ONE );
   PLA_Constant_create(  2.0, &PLA_TWO );
   PLA_Constant_complex_create( 0.0, 1.0, &PLA_i ); 
   
   PLA_s_minus_one      = -1.0;
   PLA_s_zero           =  0.0;
   PLA_s_one            =  1.0;
   
   PLA_d_minus_one      = -1.0;
   PLA_d_zero           =  0.0;
   PLA_d_one            =  1.0;
   
   PLA_c_minus_one.real = -1.0;
   PLA_c_minus_one.imag =  0.0;
   PLA_c_zero.real      =  0.0;
   PLA_c_zero.imag      =  0.0;
   PLA_c_one.real       =  1.0;
   PLA_c_one.imag       =  0.0;
   
   PLA_z_minus_one.real = -1.0;
   PLA_z_minus_one.imag =  0.0;
   PLA_z_zero.real      =  0.0;
   PLA_z_zero.imag      =  0.0;
   PLA_z_one.real       =  1.0;
   PLA_z_one.imag       =  0.0;
   
   PLA_i_minus_one      = -1;
   PLA_i_zero           =  0;
   PLA_i_one            =  1;
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Finalize( )
/*----------------------------------------------------------------------------

Purpose :  Exit PLAPACK2e environment

----------------------------------------------------------------------------*/
{
   if ( !PLA_initialized )
      return PLA_FAILURE;
   
   PLA_Obj_free( &PLA_MINUS_ONE );  
   PLA_Obj_free( &PLA_ZERO );
   PLA_Obj_free( &PLA_ONE );
   PLA_Obj_free( &PLA_TWO );
   PLA_Obj_free( &PLA_i );

   PLA_Temp_free_without_buffer( &PLA_TEMP_WORLD );
  
   PLA_setup_exit( );   
   
#ifdef FLA_ENABLE_MEM_COUNTER
   PLA_printf( "PLAPACK2e memory counter: %d\n", PLA_memcounter );
#endif
   PLA_initialized = FALSE;
   
   return PLA_SUCCESS;
}


PLA_Bool PLA_Initialized( )
/*----------------------------------------------------------------------------

Purpose :  Check if PLAPACK2e is initialized

Return :   initialized or not

----------------------------------------------------------------------------*/
{
   return PLA_initialized;
}
