/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Obj_set_to( PLA_Obj obj,
                          void    *value )
/*----------------------------------------------------------------------------

Purpose :   Set the object all to a value

IN      obj        object
IN      value      address of value

----------------------------------------------------------------------------*/
{
   int
      i, j, 
      m, n,
      ldim;
   
   m = PLA_Obj_local_length ( obj );
   n = PLA_Obj_local_width  ( obj );
   ldim = PLA_Obj_local_ldim( obj );
   
   
   switch ( PLA_Obj_datatype( obj ) )
   {
      case PLA_FLOAT: 
      {
         float
            *buff       = PLA_FLOAT_PTR( obj ),
            *buff_value = (float *) value;
         
         for ( j = 0; j < n; j++ )
            for ( i = 0; i < m; i++ )
               buff[ j * ldim + i ] = *buff_value;
         
         break;
      }
      case PLA_DOUBLE: 
      {
         double 
            *buff       = PLA_DOUBLE_PTR( obj ),
            *buff_value = (double *) value;
         
         for ( j = 0; j < n; j++ )
            for ( i = 0; i < m; i++ )
               buff[ j * ldim + i ] = *buff_value;
         
         break;
      }
      case PLA_COMPLEX: 
      {
         scomplex
            *buff       = PLA_COMPLEX_PTR( obj ),
            *buff_value = ( scomplex *) value;
         
         for ( j = 0; j < n; j++ )
            for ( i = 0; i < m; i++ )
               buff[ j * ldim + i ] = *buff_value;
         
         break;
      }
      case PLA_DOUBLE_COMPLEX: 
      {
         dcomplex
            *buff       = PLA_DOUBLE_COMPLEX_PTR( obj ),
            *buff_value = (dcomplex *) value;
         
         for ( j = 0; j < n; j++ )
            for ( i = 0; i < m; i++ )
               buff[ j * ldim + i ] = *buff_value;
         
         break;
      }
      case PLA_INT: 
      {
         int
            *buff       = PLA_INT_PTR( obj ),
            *buff_value = (int *) value;
         
         for ( j = 0; j < n; j++ )
            for ( i = 0; i < m; i++ )
               buff[ j * ldim + i ] = *buff_value;
         
         break;
      }
      case PLA_CONSTANT:
         PLA_Abort( "Cannot change constant", __LINE__, __FILE__ );
         break;
      default:
         PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
         break;
   }  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_set_to_minus_one( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Set the object all to minus one

IN      obj        object

----------------------------------------------------------------------------*/
{
   switch ( PLA_Obj_datatype( obj ) )
   {
      case PLA_FLOAT: 
               PLA_Obj_set_to( obj, &PLA_s_minus_one );
         break;
      case PLA_DOUBLE: 
               PLA_Obj_set_to( obj, &PLA_d_minus_one );
         break;
      case PLA_COMPLEX: 
               PLA_Obj_set_to( obj, &PLA_c_minus_one );
         break;
      case PLA_DOUBLE_COMPLEX: 
         PLA_Obj_set_to( obj, &PLA_z_minus_one );
         break;
      case PLA_INT: 
               PLA_Obj_set_to( obj, &PLA_i_minus_one );
         break;
      case PLA_CONSTANT:
         PLA_Abort( "Cannot change constant", __LINE__, __FILE__ );
         break;
      default:
         PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
         break;
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_set_to_zero( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Set the object all to zero

IN      obj        object

----------------------------------------------------------------------------*/
{
   switch ( PLA_Obj_datatype( obj ) )
   {
      case PLA_FLOAT: 
         PLA_Obj_set_to( obj, &PLA_s_zero );
         break;
      case PLA_DOUBLE: 
         PLA_Obj_set_to( obj, &PLA_d_zero );
         break;
      case PLA_COMPLEX: 
         PLA_Obj_set_to( obj, &PLA_c_zero );
         break;
      case PLA_DOUBLE_COMPLEX: 
         PLA_Obj_set_to( obj, &PLA_z_zero );
         break;
      case PLA_INT: 
         PLA_Obj_set_to( obj, &PLA_i_zero );
         break;
      case PLA_CONSTANT:
         PLA_Abort( "Cannot change constant", __LINE__, __FILE__ );
         break;
      default:
         PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
         break;
   } 
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_set_to_one( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Set the object all to one

IN      obj        object

----------------------------------------------------------------------------*/
{
   switch ( PLA_Obj_datatype( obj ) )
   {
      case PLA_FLOAT: 
         PLA_Obj_set_to( obj, &PLA_s_one );
         break;
      case PLA_DOUBLE: 
         PLA_Obj_set_to( obj, &PLA_d_one );
         break;
      case PLA_COMPLEX: 
         PLA_Obj_set_to( obj, &PLA_c_one );
         break;
      case PLA_DOUBLE_COMPLEX: 
         PLA_Obj_set_to( obj, &PLA_z_one );
         break;
      case PLA_INT: 
         PLA_Obj_set_to( obj, &PLA_i_one );
         break;
      case PLA_CONSTANT:
         PLA_Abort( "Cannot change constant", __LINE__, __FILE__ );
         break;
      default:
         PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
         break;
   } 
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_set_element( PLA_Obj alpha,
                               void    *value )
/*----------------------------------------------------------------------------

Purpose :   Set the scalar to the indicated value
           
IN      alpha      scalar object
IN      value      address of value

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Obj_set_element_check( alpha );
   
   if ( PLA_Obj_is_owner( alpha ) )
   {
      switch ( PLA_Obj_datatype( alpha ) )
      {
         case PLA_FLOAT: 
         {
            float
               *buff       = PLA_FLOAT_PTR( alpha ), 
               *buff_value = (float *) value;    
            
            if ( buff != NULL )
               buff[0] = *buff_value;
            
            break;
         }
         case PLA_DOUBLE: 
         {
            double 
               *buff       = PLA_DOUBLE_PTR( alpha ),
               *buff_value = (double *) value;    
            
            if ( buff != NULL )
               buff[0] = *buff_value;
            
            break;
         }
         case PLA_COMPLEX: 
         {
            scomplex
               *buff       = PLA_COMPLEX_PTR( alpha ), 
               *buff_value = ( scomplex *) value;    
            
            if ( buff != NULL )
               buff[0] = *buff_value;
            
            break;
         }
         case PLA_DOUBLE_COMPLEX: 
         {
            dcomplex
               *buff       = PLA_DOUBLE_COMPLEX_PTR( alpha ), 
               *buff_value = (dcomplex *) value;    
            
            if ( buff != NULL )
               buff[0] = *buff_value;
            
            break;
         }
         case PLA_INT: 
         {
            int
               *buff       = PLA_INT_PTR( alpha ), 
               *buff_value = (int *) value;    
            
            if ( buff != NULL )
               buff[0] = *buff_value;
            
            break;
         }
         case PLA_CONSTANT:
            PLA_Abort( "Cannot change constant", __LINE__, __FILE__ );
            break;
         default:
            PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
            break;
      } 
   }  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_add_to_element( PLA_Obj alpha,
                                  void    *value )
/*----------------------------------------------------------------------------

Purpose :   Add to the scalar with the indicated value

IN      alpha      scalar object
IN      value      address of value

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Obj_add_to_element_check( alpha );
   
   if ( PLA_Obj_is_owner( alpha ) )
   {
      switch ( PLA_Obj_datatype( alpha ) )
      {
         case PLA_FLOAT: 
         {
            float
               *buff       = PLA_FLOAT_PTR( alpha ), 
               *buff_value = (float *) value;    
            
            if ( buff != NULL )
               buff[0] += *buff_value;
            
            break;
         }
         case PLA_DOUBLE: 
         {
            double 
               *buff       = PLA_DOUBLE_PTR( alpha ), 
               *buff_value = (double *) value;    
            
            if ( buff != NULL )
               buff[0] += *buff_value;
            
            break;
         }
         case PLA_COMPLEX: 
         {
            scomplex
               *buff       = PLA_COMPLEX_PTR( alpha ), 
               *buff_value = ( scomplex *) value;    
            
            if ( buff != NULL )
            {
               buff[0].real += (*buff_value).real;
               buff[0].imag += (*buff_value).imag;
            }
            
            break;
         }
         case PLA_DOUBLE_COMPLEX: 
         {
            dcomplex
               *buff       = PLA_DOUBLE_COMPLEX_PTR( alpha ), 
               *buff_value = (dcomplex *) value;    
            
            if ( buff != NULL )
            {
               buff[0].real += (*buff_value).real;
               buff[0].imag += (*buff_value).imag;
            }
            
            break;
          }
         case PLA_INT: 
         {
            int
               *buff       = PLA_INT_PTR( alpha ), 
               *buff_value = (int *) value;    
            
            if ( buff != NULL )
               buff[0] += *buff_value;
            
            break;
         }
         case PLA_CONSTANT:
            PLA_Abort( "Cannot change constant", __LINE__, __FILE__ );
            break;
         default:
            PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
            break;
      }   
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_add_to_diagonal( void    *value,
                                   PLA_Obj A       )
/*----------------------------------------------------------------------------

Purpose :   Add value to the diagonal of the object

IN      value      address of value
IN      A          object

----------------------------------------------------------------------------*/
{
   PLA_Obj
      ATL,   ATR,      A00,  a01,     A02,
      ABL,   ABR,      a10t, alpha11, a12t,
                       A20,  a21,     A22;
   
   PLA_Part_2x2( A,    &ATL, &ATR,
                 &ABL, &ABR,     0, 0, PLA_TL );
   
   while ( min( PLA_Obj_length( ABR ), PLA_Obj_width( ABR ) ) != 0 )
   {
      PLA_Repart_2x2_to_3x3( ATL, /**/ ATR,      &A00,  /**/ &a01,     &A02,
                          /* ************* */ /* ************************* */
                                                 &a10t, /**/ &alpha11, &a12t,
                             ABL, /**/ ABR,      &A20,  /**/ &a21,     &A22,
                             1, 1, PLA_BR );

      PLA_Obj_add_to_element( alpha11, value );
      
      PLA_Cont_with_3x3_to_2x2( &ATL, /**/ &ATR,      A00,  a01,     /**/ A02,
                                                       a10t, alpha11, /**/ a12t,
                             /* *************** */ /* ********************** */
                                &ABL, /**/ &ABR,      A20,  a21,     /**/ A22,
                                PLA_TL );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_get_local_contents( PLA_Trans trans,
                                      PLA_Obj   obj,
                                      int       m,
                                      int       n,
                                      void      *A,
                                      int       ldA )
/*----------------------------------------------------------------------------

Purpose :   Extract local contents from object to local buffer

IN      trans        option to transpose
IN      obj          object
IN      m            length to extract
IN      n            width to extract
OUT     A            buffer to extract data
IN      ldA          leading dimension of buffer

----------------------------------------------------------------------------*/
{
   int
      i, j,
      ldim_obj = PLA_Obj_local_ldim( obj );
   
   if ( PLA_Error_check( ) )
      PLA_Obj_get_local_contents_check( trans, obj, m, n, A, ldA );
   
   if ( trans == PLA_NO_TRANSPOSE )
   {
      switch ( PLA_Obj_datatype( obj ) )
      {
         case PLA_FLOAT:
         {
            float 
               *buff_obj = PLA_FLOAT_PTR( obj ),
               *buff_A   = (float *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + j * ldim_obj + i );
            
            break;
         }
         case PLA_DOUBLE:
         {
            double
               *buff_obj = PLA_DOUBLE_PTR( obj ),
               *buff_A   = (double *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + j * ldim_obj + i );
            
            break;
         }
         case PLA_COMPLEX:
         {
            scomplex
               *buff_obj = PLA_COMPLEX_PTR( obj ),
               *buff_A   = ( scomplex *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + j * ldim_obj + i );
            
            break;
         }
         case PLA_DOUBLE_COMPLEX:
         {
            dcomplex
               *buff_obj = PLA_DOUBLE_COMPLEX_PTR( obj ),
               *buff_A   = (dcomplex *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + j * ldim_obj + i );
            
            break;
         }
         case PLA_INT:
         {
            int
               *buff_obj = PLA_INT_PTR( obj ),
               *buff_A   = (int *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + j * ldim_obj + i );
            
            break;
         }
         case PLA_CONSTANT:
            PLA_Abort( "Cannot extract from constant", __LINE__, __FILE__ );
            break;
         default:
            PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
            break;
      }
   }
   else if ( trans == PLA_TRANSPOSE )
   {   
      switch ( PLA_Obj_datatype( obj ) )
      {
         case PLA_FLOAT:
         {
            float 
               *buff_obj = PLA_FLOAT_PTR( obj ),
               *buff_A   = (float *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + i * ldim_obj + j );
            
            break;
         }
         case PLA_DOUBLE:
         {
            double
               *buff_obj = PLA_DOUBLE_PTR( obj ),
               *buff_A   = (double *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + i * ldim_obj + j );
            
            break;
         }
         case PLA_COMPLEX:
         {
            scomplex
               *buff_obj = PLA_COMPLEX_PTR( obj ),
               *buff_A   = ( scomplex *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + i * ldim_obj + j );
            
            break;
         }
         case PLA_DOUBLE_COMPLEX:
         {
            dcomplex
               *buff_obj = PLA_DOUBLE_COMPLEX_PTR( obj ),
               *buff_A   = (dcomplex *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + i * ldim_obj + j );
            
            break;
         }
         case PLA_INT:
         {
            int
               *buff_obj = PLA_INT_PTR( obj ),
               *buff_A   = (int *) A;
            
            for ( j = 0; j < n; j++ )
               for ( i = 0; i < m; i++ )
                  *( buff_A + j * ldA + i ) = *( buff_obj + i * ldim_obj + j );
            
            break;
         }
         case PLA_CONSTANT:
            PLA_Abort( "Cannot extract from constant", __LINE__, __FILE__ );
            break;
         default:
            PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
            break;
      }
   }
   else
      PLA_Abort( "Undefined transpose option", __LINE__, __FILE__ );
   
   return PLA_SUCCESS;
}




PLA_Error PLA_Obj_compute_owner( PLA_Obj obj,
                                 int     *myrow,
                                 int     *mycol )                         
/*----------------------------------------------------------------------------

Purpose :   Compute the owner row and column of the first element in object

IN      obj          object
OUT     myrow        rank of owning row
OUT     mycol        rank of owning column

----------------------------------------------------------------------------*/
{
   switch ( PLA_Obj_objtype( obj ) )
   {
      case PLA_MATRIX:
         *myrow = ( obj.base->templ->colDistr == PLA_ROW_MAJOR ?
                    ( ((obj.offm + obj.base->alignM) / obj.base->templ->mb) %
                      obj.base->templ->nprows ) :
                    ( ((obj.offm + obj.base->alignM) / obj.base->templ->mb) /
                      obj.base->templ->npcols ) % obj.base->templ->nprows );
         
         *mycol = ( obj.base->templ->rowDistr == PLA_COL_MAJOR ?
                    ( ((obj.offn + obj.base->alignN) / obj.base->templ->nb) %
                      obj.base->templ->npcols ) :
                    ( ((obj.offn + obj.base->alignN) / obj.base->templ->nb) /
                      obj.base->templ->nprows ) % obj.base->templ->npcols );
         break;
         
      case PLA_MROWVECTOR:
         if ( obj.base->templ->rowDistr == PLA_COL_MAJOR )
         {
            *myrow = ( ((obj.offn + obj.base->alignN) / obj.base->templ->nb) %
                       obj.base->templ->nprows );
            *mycol = ( ((obj.offn + obj.base->alignN) / obj.base->templ->nb) /
                       obj.base->templ->nprows % obj.base->templ->npcols );
         }
         else
         {
            *myrow = ( ((obj.offn + obj.base->alignN) / obj.base->templ->nb) /
                       obj.base->templ->npcols % obj.base->templ->nprows );
            *mycol = ( ((obj.offn + obj.base->alignN) / obj.base->templ->nb) %
                       obj.base->templ->npcols );
         }
         break;
         
      case PLA_MCOLVECTOR:
         if ( obj.base->templ->colDistr == PLA_COL_MAJOR )
         {
            *myrow = ( ((obj.offm + obj.base->alignM) / obj.base->templ->mb) %
                       obj.base->templ->nprows );
            *mycol = ( ((obj.offm + obj.base->alignM) / obj.base->templ->mb) /
                       obj.base->templ->nprows % obj.base->templ->npcols );
         }
         else
         {
            *myrow = ( ((obj.offm + obj.base->alignM) / obj.base->templ->mb) /
                       obj.base->templ->npcols % obj.base->templ->nprows );
            *mycol = ( ((obj.offm + obj.base->alignM) / obj.base->templ->mb) %
                       obj.base->templ->npcols );
         }
         break;
         
      case PLA_MSCALAR:
         *myrow = obj.base->ownerRow;
         *mycol = obj.base->ownerCol;
         break;
         
      case PLA_PMROWVECTOR:
         *myrow = obj.base->ownerRow;
         *mycol = ( ((obj.offn + obj.base->alignN) / obj.base->templ->nb) %
                    obj.base->templ->npcols );
         break;
         
      case PLA_PMCOLVECTOR:
         *myrow = ( ((obj.offm + obj.base->alignM) / obj.base->templ->mb) %
                    obj.base->templ->nprows );
         *mycol = obj.base->ownerCol;
         break;
         
      case PLA_CONSTANT:
         *myrow = obj.base->ownerRow;
         *mycol = obj.base->ownerCol;
         break;
         
      default:
         PLA_Abort( "Undefined objtype", __LINE__, __FILE__ );
         break;
   }
   
   return PLA_SUCCESS;
}


/*----------------------------------------------------------------------------
  
   MPI uses row major ordering

----------------------------------------------------------------------------*/
#define ownerAll( row, col, obj )  ((row) * (obj).base->templ->npcols + (col))


int PLA_Obj_owner( PLA_Obj obj,
                   int     *myrow,
                   int     *mycol )                         
/*----------------------------------------------------------------------------

Purpose :   Find the owner row and column of the first element in the object

IN      obj          object
OUT     myrow        rank of owning row
OUT     mycol        rank of owning column

Return :    owner of object on all nodes in row major naming order

----------------------------------------------------------------------------*/
{
   int
      row, col;

   PLA_Obj_compute_owner( obj, myrow, mycol );
   
   if ( PLA_Obj_is_valid_state( obj ) )
   {      
      if ( *myrow == PLA_ALL_ROWS )
         row = obj.base->templ->myrow;
      else
         row = *myrow;
      
      if ( *mycol == PLA_ALL_COLS )
         col = obj.base->templ->mycol;
      else
         col = *mycol;
      
      return ownerAll( row, col, obj );
   }

   return PLA_UNDEFINED;
}


int PLA_Obj_owner_all( PLA_Obj obj ) 
/*----------------------------------------------------------------------------

Purpose :   Find the owner of object on all nodes

IN      obj          object

Return :    owner of object on all nodes in row major naming order

----------------------------------------------------------------------------*/
{
   int
      row, col;
   
   PLA_Obj_compute_owner( obj, &row, &col );

   if ( PLA_Obj_is_valid_state( obj ) )
   {
      if ( row == PLA_ALL_ROWS )
         row = obj.base->templ->myrow;
      
      if ( col == PLA_ALL_COLS )
         col = obj.base->templ->mycol;
      
      return ownerAll( row, col, obj );
   }

   return PLA_UNDEFINED;
}


PLA_Bool PLA_Obj_is_owner( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Determine is processor owns the object

IN      obj          object

Return :    if the processor owns the object:  TRUE or FALSE

----------------------------------------------------------------------------*/
{
   int
      row, col;
   
   PLA_Obj_compute_owner( obj, &row, &col );
   
   if ( PLA_Obj_is_valid_state( obj ) )
   {
      if ( ( row == obj.base->templ->myrow ||
             row == PLA_ALL_ROWS ) &&
           ( col == obj.base->templ->mycol ||
             col == PLA_ALL_COLS ) )
         return TRUE;
   }

   return FALSE;
}


int PLA_Obj_split_size( PLA_Obj  obj,
                        PLA_Side side )
/*----------------------------------------------------------------------------

Purpose :   Compute size of split to next block boundary

IN      obj       object
IN      side      side to find block boundary

Return :    length to next template sub-block split

----------------------------------------------------------------------------*/
{  
   switch ( PLA_Obj_objtype( obj ) )
   {
      case PLA_MATRIX:
         switch ( side )
         {
            case PLA_TOP:
               return min( obj.base->templ->mb - (obj.base->alignM + obj.offm) % obj.base->templ->mb, 
                           obj.m );
            case PLA_BOTTOM:
               return min( ( (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb > 0 ?
                             (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb : obj.base->templ->mb ), 
                           obj.m );
            case PLA_LEFT:
               return min( obj.base->templ->nb - (obj.base->alignN + obj.offn) % obj.base->templ->nb, 
                           obj.n );
            case PLA_RIGHT:
               return min( ( (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb > 0 ? 
                             (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb : obj.base->templ->nb ),
                           obj.n );
               
            case PLA_TL:
               return min( min( obj.base->templ->mb - (obj.base->alignM + obj.offm) % obj.base->templ->mb, 
                                obj.m ),
                           min( obj.base->templ->nb - (obj.base->alignN + obj.offn) % obj.base->templ->nb, 
                                obj.n ) );
            case PLA_TR:
               return min( min( obj.base->templ->mb - (obj.base->alignM + obj.offm) % obj.base->templ->mb, 
                                obj.m ),
                           min( ( (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb > 0 ? 
                                  (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb : obj.base->templ->nb ),
                                obj.n ) );
            case PLA_BL:
               return min( min( ( (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb > 0 ?
                                  (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb : obj.base->templ->mb ), 
                                obj.m ),
                           min( obj.base->templ->nb - (obj.base->alignN + obj.offn) % obj.base->templ->nb, 
                                obj.n ) );
            case PLA_BR:
               return min( min( ( (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb > 0 ?
                                  (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb : obj.base->templ->mb ), 
                                obj.m ),
                           min( ( (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb > 0 ? 
                                  (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb : obj.base->templ->nb ),
                                obj.n ) );
               
            default:
               PLA_Abort( "Side not yet implemented", __LINE__, __FILE__ );
               break;
         }
         break;
         
      case PLA_MROWVECTOR:
         switch ( side )
         {
            case PLA_TOP:
               return obj.m;
            case PLA_BOTTOM:
               return obj.m;
            case PLA_LEFT:
               return min( obj.base->templ->nb - (obj.base->alignN + obj.offn) % obj.base->templ->nb, 
                           obj.n );
            case PLA_RIGHT:
               return min( ( (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb > 0 ? 
                             (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb : obj.base->templ->nb ),
                           obj.n );
               
            case PLA_TL:
               return min( obj.m,
                           min( obj.base->templ->nb - (obj.base->alignN + obj.offn) % obj.base->templ->nb, 
                                obj.n ) );
            case PLA_TR:
               return min( obj.m,
                           min( ( (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb > 0 ? 
                                  (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb : obj.base->templ->nb ),
                                obj.n ) );
            case PLA_BL:
               return min( obj.m,
                           min( obj.base->templ->nb - (obj.base->alignN + obj.offn) % obj.base->templ->nb, 
                                obj.n ) );
            case PLA_BR:
               return min( obj.m,
                           min( ( (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb > 0 ? 
                                  (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb : obj.base->templ->nb ),
                                obj.n ) );
               
            default:
               PLA_Abort( "Side not yet implemented", __LINE__, __FILE__ );
               break;
         }
         break;
         
      case PLA_MCOLVECTOR:
         switch ( side )
         {
            case PLA_TOP:
               return min( obj.base->templ->mb - (obj.base->alignM + obj.offm) % obj.base->templ->mb, 
                           obj.m );
            case PLA_BOTTOM:
               return min( ( (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb > 0 ?
                             (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb : obj.base->templ->mb ), 
                           obj.m );
            case PLA_LEFT:
               return obj.n;
            case PLA_RIGHT:
               return obj.n;
               
            case PLA_TL:
               return min( min( obj.base->templ->mb - (obj.base->alignM + obj.offm) % obj.base->templ->mb, 
                                obj.m ),
                           obj.n );
            case PLA_TR:
               return min( min( obj.base->templ->mb - (obj.base->alignM + obj.offm) % obj.base->templ->mb, 
                                obj.m ),
                           obj.n );
            case PLA_BL:
               return min( min( ( (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb > 0 ?
                                  (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb : obj.base->templ->mb ), 
                                obj.m ),
                           obj.n );
            case PLA_BR:
               return min( min( ( (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb > 0 ?
                                  (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb : obj.base->templ->mb ), 
                                obj.m ),
                           obj.n );
               
            default:
               PLA_Abort( "Side not yet implemented", __LINE__, __FILE__ );
               break;
         }
         break;
         
      case PLA_MSCALAR:
         switch ( side )
         {
            case PLA_TOP:
               return obj.m;
            case PLA_BOTTOM:
               return obj.m;
            case PLA_LEFT:
               return obj.n;
            case PLA_RIGHT:
               return obj.n;
               
            case PLA_TL:
               return min( obj.m,
                           obj.n );
            case PLA_TR:
               return min( obj.m,
                           obj.n );
            case PLA_BL:
               return min( obj.m,
                           obj.n );
            case PLA_BR:
               return min( obj.m,
                           obj.n );
               
            default:
               PLA_Abort( "Side not yet implemented", __LINE__, __FILE__ );
               break;
         }
         break;
         
      case PLA_PMROWVECTOR:
         switch ( side )
         {
            case PLA_TOP:
               return obj.m;
            case PLA_BOTTOM:
               return obj.m;
            case PLA_LEFT:
               return min( obj.base->templ->nb - (obj.base->alignN + obj.offn) % obj.base->templ->nb, 
                           obj.n );
            case PLA_RIGHT:
               return min( ( (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb > 0 ? 
                             (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb : obj.base->templ->nb ),
                           obj.n );
               
            case PLA_TL:
               return min( obj.m,
                           min( obj.base->templ->nb - (obj.base->alignN + obj.offn) % obj.base->templ->nb, 
                                obj.n ) );
            case PLA_TR:
               return min( obj.m,
                           min( ( (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb > 0 ? 
                                  (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb : obj.base->templ->nb ),
                                obj.n ) );
            case PLA_BL:
               return min( obj.m,
                           min( obj.base->templ->nb - (obj.base->alignN + obj.offn) % obj.base->templ->nb, 
                                obj.n ) );
            case PLA_BR:
               return min( obj.m,
                           min( ( (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb > 0 ? 
                                  (obj.base->alignN + obj.offn + obj.n) % obj.base->templ->nb : obj.base->templ->nb ),
                                obj.n ) );
               
            default:
               PLA_Abort( "Side not yet implemented", __LINE__, __FILE__ );
               break;
         }
         break;
         
      case PLA_PMCOLVECTOR:
         switch ( side )
         {
            case PLA_TOP:
               return min( obj.base->templ->mb - (obj.base->alignM + obj.offm) % obj.base->templ->mb, 
                           obj.m );
            case PLA_BOTTOM:
               return min( ( (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb > 0 ?
                             (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb : obj.base->templ->mb ), 
                           obj.m );
            case PLA_LEFT:
               return obj.n;
            case PLA_RIGHT:
               return obj.n;
               
            case PLA_TL:
               return min( min( obj.base->templ->mb - (obj.base->alignM + obj.offm) % obj.base->templ->mb, 
                                obj.m ),
                           obj.n );
            case PLA_TR:
               return min( min( obj.base->templ->mb - (obj.base->alignM + obj.offm) % obj.base->templ->mb, 
                                obj.m ),
                           obj.n );
            case PLA_BL:
               return min( min( ( (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb > 0 ?
                                  (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb : obj.base->templ->mb ), 
                                obj.m ),
                           obj.n );
            case PLA_BR:
               return min( min( ( (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb > 0 ?
                                  (obj.base->alignM + obj.offm + obj.m) % obj.base->templ->mb : obj.base->templ->mb ), 
                                obj.m ),
                           obj.n );
               
            default:
               PLA_Abort( "Side not yet implemented", __LINE__, __FILE__ );
               break;
         }
         break;
         
      case PLA_CONSTANT:    
         switch ( side )
         {
            case PLA_TOP:
               return obj.m;
            case PLA_BOTTOM:
               return obj.m;
            case PLA_LEFT:
               return obj.n;
            case PLA_RIGHT:
               return obj.n;
               
            case PLA_TL:
               return min( obj.m,
                           obj.n );
            case PLA_TR:
               return min( obj.m,
                           obj.n );
            case PLA_BL:
               return min( obj.m,
                           obj.n );
            case PLA_BR:
               return min( obj.m,
                           obj.n );
               
            default:
               PLA_Abort( "Side not yet implemented", __LINE__, __FILE__ );
               break;
         }
         break;
         
      default:
         PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
         break;      
   }
   
   return 0;
}
