/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Part_2x2( PLA_Obj A,  PLA_Obj *ATL, PLA_Obj *ATR,
			            PLA_Obj *ABL, PLA_Obj *ABR, 
			int mb, int nb, PLA_Quadrant quadrant )
/*----------------------------------------------------------------------------

Purpose :   Partition matrix A into a top and bottom side where the side
            indicated by 'side' has 'mb' rows and 'nb' columns

-----------    -----------
|         |    |ATL | ATR|
|    A    | => |----|----|
|         |    |ABL | ABR|
-----------    -----------

IN      A          old matrix view
OUT     ATL        new top left matrix view
OUT     ATR        new top right matrix view
OUT     ABR        new bottom left matrix view
OUT     ABL        new bottom right matrix view
IN      mb         block size of new matrix indicated by 'size'
IN      nb         block size of new matrix indicated by 'size'
IN      quadrant   matrix specified to make if dimension mb x nb

----------------------------------------------------------------------------*/
{
   // Set mb and nb to be the dimensions of ATL
   if ( quadrant == PLA_BL || quadrant == PLA_BR ) 
      mb = A.m - mb;
   if ( quadrant == PLA_TR || quadrant == PLA_BR ) 
      nb = A.n - nb;
   
   ATL->m      = mb;
   ATL->n      = nb;
   ATL->offm   = A.offm;
   ATL->offn   = A.offn;
   ATL->base   = A.base;
   
   ABL->m      = A.m - mb;
   ABL->n      = nb;
   ABL->offm   = A.offm + mb;
   ABL->offn   = A.offn;
   ABL->base   = A.base;
   
   ATR->m      = mb;
   ATR->n      = A.n - nb;
   ATR->offm   = A.offm;
   ATR->offn   = A.offn + nb;
   ATR->base   = A.base;
   
   ABR->m      = A.m - mb;
   ABR->n      = A.n - nb;
   ABR->offm   = A.offm + mb;
   ABR->offn   = A.offn + nb;
   ABR->base   = A.base;
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Merge_2x2( PLA_Obj ATL, PLA_Obj ATR,
			 PLA_Obj ABL, PLA_Obj ABR, 
			 PLA_Obj *A )
/*----------------------------------------------------------------------------

Purpose :   Merge a 2 x 2 partitioning of a matrix into a single view

-----------    -----------
|ATL | ATR|    |         |
|----|----| => |    A    |
|ABL | ABR|    |         |
-----------    -----------

IN      ATL        old top left matrix view
IN      ATR        old top right matrix view
IN      ABR        old bottom left matrix view
IN      ABL        old bottom right matrix view
OUT     A          new matrix view

----------------------------------------------------------------------------*/
{
   if ( A == NULL )
      PLA_Abort( "PLA_Merge: NULL object!", __LINE__, __FILE__ );

   if ( ATL.base->buffer != ABL.base->buffer ||
	ATR.base->buffer != ABR.base->buffer ||
	ATL.base->buffer != ATR.base->buffer )
      PLA_Abort( "PLA_Merge: buffer mismatch!", __LINE__, __FILE__ );
   
   if ( ATL.m != ATR.m || ABL.m != ABR.m || 
	ATL.n != ABL.n || ATR.n != ABR.n )
      PLA_Abort( "PLA_Merge: dimension mismatch!", __LINE__, __FILE__ );

   if ( ATL.offm != ABL.offm + ABL.m ||
	ATR.offm != ABR.offm + ABL.m ||
	ATL.offm != ATR.offm         ||
	ABL.offm != ABR.offm )
      PLA_Abort( "PLA_Merge: Objects not vertically adjacent!", __LINE__, __FILE__ );
   
   if ( ATL.offn != ABL.offn         ||
	ATR.offn != ABR.offn         ||
	ATL.offn != ATR.offn + ATR.n ||
	ABL.offn != ABR.offn + ATR.n )
      PLA_Abort( "PLA_Merge: Objects not horizontally adjacent!", __LINE__, __FILE__ );
   
   A->m      = ATL.m + ABL.m;
   A->n      = ATL.n + ATR.n;
   A->offm   = ATL.offm;
   A->offn   = ATL.offn;
   A->base   = ATL.base;
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Repart_2x2_to_3x3( 
   PLA_Obj ATL, PLA_Obj ATR,  PLA_Obj *A00, PLA_Obj *A01, PLA_Obj *A02,
                              PLA_Obj *A10, PLA_Obj *A11, PLA_Obj *A12,
   PLA_Obj ABL, PLA_Obj ABR,  PLA_Obj *A20, PLA_Obj *A21, PLA_Obj *A22,
   int mb, int nb, PLA_Quadrant quadrant )
/*----------------------------------------------------------------------------

Purpose :   Repartition a 2 x 2 partitioning of matrix A into 
            a 3 x 3 partitioning where 'mb' x 'nb' submatrix A11 
            split from the quadrant indicated by 'quadrant'

-------------    -------------
|ATL  |  ATR|    |A00|A01|A02| 
|     |     |    |------------
|-----------| => |A10|A11|A12|
|     |     |    |------------
|ABL  |  ABR|    |A20|A21|A22|
-------------    -------------

IN      ATL             old top left matrix view
IN      ATR             old top right matrix view
IN      ABR             old bottom left matrix view
IN      ABL             old bottom right matrix view
OUT     A00, A01, A02   new matrix views
        A10, A11, A12   
        A20, A21, A22
IN      mb              block size of new matrix indicated by 'size'
IN      nb              block size of new matrix indicated by 'size'
IN      quadrant        matrix specified to make if dimension mb x nb

----------------------------------------------------------------------------*/
{
   if ( quadrant == PLA_BR )
   { 
      A00->m      = ATL.m;
      A00->n      = ATL.n;
      A00->offm   = ATL.offm;
      A00->offn   = ATL.offn;
      A00->base   = ATL.base;
      
      A01->m      = ATR.m;
      A01->n      = nb;
      A01->offm   = ATR.offm;
      A01->offn   = ATR.offn;
      A01->base   = ATR.base;
   
      A02->m      = ATR.m;
      A02->n      = ATR.n - nb;
      A02->offm   = ATR.offm;
      A02->offn   = ATR.offn + nb;
      A02->base   = ATR.base;
      
      A10->m      = mb;
      A10->n      = ABL.n;
      A10->offm   = ABL.offm;
      A10->offn   = ABL.offn;
      A10->base   = ABL.base;
      
      A11->m      = mb;
      A11->n      = nb;
      A11->offm   = ABR.offm;
      A11->offn   = ABR.offn;
      A11->base   = ABR.base;
      
      A12->m      = mb;
      A12->n      = ABR.n - nb;
      A12->offm   = ABR.offm;
      A12->offn   = ABR.offn + nb;
      A12->base   = ABR.base;
      
      A20->m      = ABL.m - mb;
      A20->n      = ABL.n;
      A20->offm   = ABL.offm + mb;
      A20->offn   = ABL.offn;
      A20->base   = ABL.base;
      
      A21->m      = ABR.m - mb;
      A21->n      = nb;
      A21->offm   = ABR.offm + mb;
      A21->offn   = ABR.offn;
      A21->base   = ABR.base;
      
      A22->m      = ABR.m - mb;
      A22->n      = ABR.n - nb;
      A22->offm   = ABR.offm + mb;
      A22->offn   = ABR.offn + nb;
      A22->base   = ABR.base;
   }
   else if ( quadrant == PLA_TL )
   {
      PLA_Part_2x2( ATL,   A00, /**/ A01,
		        /* ************* */
		           A10, /**/ A11,
		    /* with */ mb, /* by */ nb, /* submatrix */ PLA_BR );
      
      PLA_Part_1x2( ABL,   A20, /**/ A21,
		    /* with width */ nb, /* submatrix */ PLA_RIGHT );
      
      PLA_Part_2x1 ( ATR,   A02, 
                         /* *** */
		            A12,
		     /* with length */ mb, /* submatrix */ PLA_BOTTOM );
      
      A22->m      = ABR.m;
      A22->n      = ABR.n;
      A22->offm   = ABR.offm;
      A22->offn   = ABR.offn;
      A22->base   = ABR.base;
   }
   else if ( quadrant == PLA_TR )
   {
      PLA_Part_2x2( ATR,   A01, /**/ A02,
		        /* ************* */
		           A11, /**/ A12,
		    /* with */ mb, /* by */ nb, /* submatrix */ PLA_BL );
      
      PLA_Part_2x1( ATL,   A00, 
		        /* *** */
                           A10, 
		    mb, PLA_BOTTOM );
      
      PLA_Part_1x2( ABR,   A21, /* ** */ A22, 
		    nb, PLA_LEFT );
      
      A20->m      = ABL.m;
      A20->n      = ABL.n;
      A20->offm   = ABL.offm;
      A20->offn   = ABL.offn;
      A20->base   = ABR.base;
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Cont_with_3x3_to_2x2( 
   PLA_Obj *ATL, PLA_Obj *ATR,  PLA_Obj A00, PLA_Obj A01, PLA_Obj A02,
                                PLA_Obj A10, PLA_Obj A11, PLA_Obj A12,
   PLA_Obj *ABL, PLA_Obj *ABR,  PLA_Obj A20, PLA_Obj A21, PLA_Obj A22,
   PLA_Quadrant quadrant )
/*----------------------------------------------------------------------------

Purpose :   Update the 2 x 2 partitioning of matrix A by moving the boundaries
            so that A11 is added to the quadrant indicated by 'quadrant'

-------------    -------------
|A00|A01|A02|    |ATL  |  ATR|
|------------    |     |     |
|A10|A11|A12| => |-----------|
|------------    |     |     |
|A20|A21|A22|    |ABL  |  ABR|
-------------    -------------

OUT     ATL             new top left matrix view
OUT     ATR             new top right matrix view
OUT     ABR             new bottom left matrix view
OUT     ABL             new bottom right matrix view
IN      A00, A01, A02   old matrix views
        A10, A11, A12
        A20, A21, A22
IN      quadrant        matrix specified to make

----------------------------------------------------------------------------*/
{
   if ( quadrant == PLA_TL )
   {
      ATL->m      = A00.m + A10.m;
      ATL->n      = A00.n + A01.n;
      ATL->offm   = A00.offm;
      ATL->offn   = A00.offn;
      ATL->base   = A00.base;
      
      ATR->m      = A02.m + A12.m;
      ATR->n      = A02.n;
      ATR->offm   = A02.offm;
      ATR->offn   = A02.offn;
      ATR->base   = A02.base;
      
      ABL->m      = A20.m;
      ABL->n      = A20.n + A21.n;
      ABL->offm   = A20.offm;
      ABL->offn   = A20.offn;
      ABL->base   = A20.base;
      
      ABR->m      = A22.m;
      ABR->n      = A22.n;
      ABR->offm   = A22.offm;
      ABR->offn   = A22.offn;
      ABR->base   = A22.base;
   }
   else if ( quadrant == PLA_BR )
   {
      ATL->m      = A00.m;
      ATL->n      = A00.n;
      ATL->offm   = A00.offm;
      ATL->offn   = A00.offn;
      ATL->base   = A00.base;
      
      ATR->m      = A01.m;
      ATR->n      = A01.n + A02.n;
      ATR->offm   = A01.offm;
      ATR->offn   = A01.offn;
      ATR->base   = A01.base;
      
      ABL->m      = A10.m + A20.m;
      ABL->n      = A10.n;
      ABL->offm   = A10.offm;
      ABL->offn   = A10.offn;
      ABL->base   = A10.base;
      
      ABR->m      = A11.m + A21.m;
      ABR->n      = A11.n + A12.n;
      ABR->offm   = A11.offm;
      ABR->offn   = A11.offn;
      ABR->base   = A11.base;
   }
   else if ( quadrant == PLA_BL )
   {
      ATL->m      = A00.m;
      ATL->n      = A00.n + A01.n;
      ATL->offm   = A00.offm;
      ATL->offn   = A00.offn;
      ATL->base   = A00.base;
      
      ATR->m      = A02.m;
      ATR->n      = A02.n;
      ATR->offm   = A02.offm;
      ATR->offn   = A02.offn;
      ATR->base   = A02.base;
      
      ABL->m      = A10.m + A20.m;
      ABL->n      = A10.n + A11.n;
      ABL->offm   = A10.offm;
      ABL->offn   = A10.offn;
      ABL->base   = A10.base;
      
      ABR->m      = A12.m + A22.m;
      ABR->n      = A12.n ;
      ABR->offm   = A12.offm;
      ABR->offn   = A12.offn;
      ABR->base   = A12.base;
   }
   else
      PLA_Abort( "PLA_TR not yet implemented", __LINE__, __FILE__ );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Part_2x1( PLA_Obj A,  PLA_Obj *AT, 
			            PLA_Obj *AB,
			int mb, PLA_Side side )
/*----------------------------------------------------------------------------

Purpose :   Partition matrix A into a top and bottom side where the side
            indicated by 'side' has 'mb' rows

-----------    -----------
|         |    |   AT    |
|    A    | => |---------|
|         |    |   AB    |
-----------    -----------

IN      A        old matrix view
OUT     AT       new top matrix view
OUT     AB       new bottom matrix view
IN      mb       block size of new matrix indicated by 'size'
IN      side     matrix specified to make if dimension mb x n

----------------------------------------------------------------------------*/
{  
   // Set mb to be the dimension of AT
   if ( side == PLA_BOTTOM ) 
      mb = A.m - mb;
   
   AT->m      = mb;
   AT->n      = A.n;
   AT->offm   = A.offm;
   AT->offn   = A.offn;
   AT->base   = A.base;
   
   AB->m      = A.m - mb;
   AB->n      = A.n;
   AB->offm   = A.offm + mb;
   AB->offn   = A.offn;
   AB->base   = A.base;
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Merge_2x1( PLA_Obj AT,
			 PLA_Obj AB,
			 PLA_Obj *A )
/*----------------------------------------------------------------------------

Purpose :   Merge a 2 x 1 partitioning of a matrix into a single view

-----------    -----------
|   AT    |    |         |
|---------| => |    A    |
|   AB    |    |         |
-----------    ----------

IN      AT       old top matrix view
IN      AB       old bottom matrix view
OUT     A        new matrix view

----------------------------------------------------------------------------*/
{
   if ( A == NULL )
      PLA_Abort( "PLA_Merge: NULL object!", __LINE__, __FILE__ );

   if ( AT.base->buffer != AB.base->buffer )
      PLA_Abort( "PLA_Merge: buffer mismatch!", __LINE__, __FILE__ );
   
   if ( AT.n != AB.n )
      PLA_Abort( "PLA_Merge: dimension mismatch!", __LINE__, __FILE__ );
   
   if ( AB.offm != AT.offm + AT.m )
      PLA_Abort( "PLA_Merge: Objects not vertically adjacent!", __LINE__, __FILE__ );

   if ( AB.offn != AT.offn )
      PLA_Abort( "PLA_Merge: Objects not horizontally adjacent!", __LINE__, __FILE__ );

   A->m      = AT.m + AB.m;
   A->n      = AT.n;
   A->offm   = AT.offm;
   A->offn   = AT.offn;
   A->base   = AT.base;
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Repart_2x1_to_3x1( PLA_Obj AT,  PLA_Obj *A0,
                                              PLA_Obj *A1,
				 PLA_Obj AB,  PLA_Obj *A2,
				 int mb, PLA_Side side )
/*----------------------------------------------------------------------------

Purpose :   Repartition a 2 x 1 partitioning of matrix A into 
            a 3 x 1 partitioning where submatrix A1 with 'mb' rows 
            is split from the side indicated by 'side'

-----------    -----------
|   AT    |    |   A0    |
|         |    |---------|
|---------| => |   A1    |
|         |    |---------|
|   AB    |    |   A2    |
-----------    -----------

IN      AT       old top matrix view
IN      AB       old bottom matrix view
OUT     A0       new top matrix view
OUT     A1       new middle matrix view
OUT     A2       new bottom matrix view
IN      mb       block size of new matrix indicated by 'size'
IN      side     matrix specified to make if dimension mb x n

----------------------------------------------------------------------------*/
{
   if ( side == PLA_TOP )
   {
      PLA_Part_2x1( AT,   A0, 
		       /* ** */
		          A1,
		    /* with length */ mb, /* submatrix */ PLA_BOTTOM );
    
      A2->m      = AB.m;
      A2->n      = AB.n;
      A2->offm   = AB.offm;
      A2->offn   = AB.offn;
      A2->base   = AB.base;
   }
   else
   {
      A0->m      = AT.m;
      A0->n      = AT.n;
      A0->offm   = AT.offm;
      A0->offn   = AT.offn;
      A0->base   = AT.base;
      
      PLA_Part_2x1( AB,  A1, 
                      /* *** */
                         A2,
		    /* with length */ mb, /* submatrix */ PLA_TOP );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Cont_with_3x1_to_2x1( PLA_Obj *AT,  PLA_Obj A0,
                                                  PLA_Obj A1,
				    PLA_Obj *AB,  PLA_Obj A2,
				    PLA_Side side )
/*----------------------------------------------------------------------------

Purpose :   Update the 2 x 1 partitioning of matrix A into 
            a 3 x 1 partitioning where submatrix A1 with 'mb' rows 
	    is split from the side indicated by 'side'

-----------    -----------
|   A0    |    |   AT    |
|---------|    |         |
|   A1    | => |---------|
|---------|    |         |
|   A2    |    |   AB    |
-----------    -----------

OUT     AT       new top matrix view
OUT     AB       new bottom matrix view
IN      A0       old top matrix view
IN      A1       old middle matrix view
IN      A2       old bottom matrix view
IN      side     matrix specified to make

----------------------------------------------------------------------------*/
{
   if ( side == PLA_TOP ) 
   {
      AT->m      = A0.m + A1.m;
      AT->n      = A0.n;
      AT->offm   = A0.offm;
      AT->offn   = A0.offn;
      AT->base   = A0.base;
      
      AB->m      = A2.m;
      AB->n      = A2.n;
      AB->offm   = A2.offm;
      AB->offn   = A2.offn;
      AB->base   = A2.base;
   }
   else
   {
      AT->m      = A0.m;
      AT->n      = A0.n;
      AT->offm   = A0.offm;
      AT->offn   = A0.offn;
      AT->base   = A0.base;
      
      AB->m      = A1.m + A2.m;
      AB->n      = A1.n;
      AB->offm   = A1.offm;
      AB->offn   = A1.offn;
      AB->base   = A1.base;
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Part_1x2( PLA_Obj A, 
			PLA_Obj *AL, PLA_Obj *AR,
			int nb, PLA_Side side )
/*----------------------------------------------------------------------------

Purpose :   Partition matrix A into a top and bottom side where the side
            indicated by 'side' has 'nb' columns

-----------    -----------
|         |    |    |    |
|    A    | => | AL | AR |
|         |    |    |    |
-----------    -----------

IN      A        old matrix view
OUT     AL       new left matrix view
OUT     AR       new right matrix view
IN      nb       block size of new matrix indicated by 'size'
IN      side     matrix specified to make if dimension m x nb

----------------------------------------------------------------------------*/
{
   if ( side == PLA_LEFT )
   {
      AL->m      = A.m;
      AL->n      = nb;
      AL->offm   = A.offm;
      AL->offn   = A.offn;
      AL->base   = A.base;
      
      AR->m      = A.m;
      AR->n      = A.n - nb;
      AR->offm   = A.offm;
      AR->offn   = A.offn + nb;
      AR->base   = A.base;
   }
   else
   {
      AL->m      = A.m;
      AL->n      = A.n - nb;
      AL->offm   = A.offm;
      AL->offn   = A.offn;
      AL->base   = A.base;
      
      AR->m      = A.m;
      AR->n      = nb;
      AR->offm   = A.offm;
      AR->offn   = A.offn + (A.n - nb);
      AR->base   = A.base;
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Merge_1x2( PLA_Obj AL, PLA_Obj AR,
			 PLA_Obj *A )
/*----------------------------------------------------------------------------

Purpose :   Merge a 2 x 1 partitioning of a matrix into a single view

-----------    -----------
|    |    |    |         |
| AL | AR | => |    A    |
|    |    |    |         |
-----------    ----------

IN      AL       old left matrix view
IN      AR       old right matrix view
OUT     A        new matrix view

----------------------------------------------------------------------------*/
{
   if ( A == NULL )
      PLA_Abort( "PLA_Merge: NULL object!", __LINE__, __FILE__ );
   
   if ( AL.base->buffer != AR.base->buffer )
      PLA_Abort( "PLA_Merge: buffer mismatch!", __LINE__, __FILE__ );
   
   if ( AL.m != AR.m )
      PLA_Abort( "PLA_Merge: dimension mismatch!", __LINE__, __FILE__ );

   if ( AL.offm != AR.offm )
      PLA_Abort( "PLA_Merge: Objects not vertically adjacent!", __LINE__, __FILE__ );

   if ( AL.offn != AR.offn + AR.n )
      PLA_Abort( "PLA_Merge: Objects not horizontally adjacent!", __LINE__, __FILE__ );
   
   A->m      = AL.m;
   A->n      = AL.n + AR.n;
   A->offm   = AL.offm;
   A->offn   = AL.offn;
   A->base   = AL.base;
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Repart_1x2_to_1x3( PLA_Obj  AL,              PLA_Obj  AR,
				 PLA_Obj *A0, PLA_Obj *A1, PLA_Obj *A2,
				 int nb, PLA_Side side )
/*----------------------------------------------------------------------------

Purpose :   Repartition a 2 x 1 partitioning of matrix A into 
            a 3 x 1 partitioning where submatrix A1 with 'nb' columns 
	    is split from the side indicated by 'side'

-----------    ----------
|    |    |    |  |  |  |
|    |    |    |  |  |  |
| AL | AR | => |A0|A1|A2|
|    |    |    |  |  |  |
|    |    |    |  |  |  |
-----------    ----------

IN      AL       old left matrix view
IN      AR       old right matrix view
OUT     A0       new left matrix view
OUT     A1       new middle matrix view
OUT     A2       new right matrix view
IN      mb       block size of new matrix indicated by 'size'
IN      side     matrix specified to make if dimension m x nb

----------------------------------------------------------------------------*/
{
   if ( side == PLA_LEFT )
   {
      A0->m      = AL.m;
      A0->n      = AL.n - nb;
      A0->offm   = AL.offm;
      A0->offn   = AL.offn;
      A0->base   = AL.base;
      
      A1->m      = AL.m;
      A1->n      = nb;
      A1->offm   = AL.offm;
      A1->offn   = AL.offn + (AL.n - nb);
      A1->base   = AL.base;
      
      A2->m      = AR.m;
      A2->n      = AR.n;
      A2->offm   = AR.offm;
      A2->offn   = AR.offn;
      A2->base   = AR.base;
   }
   else // if ( side == PLA_RIGHT )
   {
      A0->m      = AL.m;
      A0->n      = AL.n;
      A0->offm   = AL.offm;
      A0->offn   = AL.offn;
      A0->base   = AL.base;
      
      A1->m      = AR.m;
      A1->n      = nb;
      A1->offm   = AR.offm;
      A1->offn   = AR.offn;
      A1->base   = AR.base;
      
      A2->m      = AR.m;
      A2->n      = AR.n - nb;
      A2->offm   = AR.offm;
      A2->offn   = AR.offn + nb;
      A2->base   = AR.base;
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Cont_with_1x3_to_1x2( PLA_Obj *AL,              PLA_Obj *AR,
				    PLA_Obj  A0, PLA_Obj  A1, PLA_Obj  A2,
				    PLA_Side side )
/*----------------------------------------------------------------------------

Purpose :   Update the 1 x 2 partitioning of matrix A into 
            a 1 x 3 partitioning where submatrix A1 with 'nb' columns 
	    is split from the side indicated by 'side'

----------    -----------
|  |  |  |    |    |    |
|  |  |  |    |    |    |
|A0|A1|A2| => | AL | AR |
|  |  |  |    |    |    |
|  |  |  |    |    |    |
----------    -----------

OUT     AL       new left matrix view
OUT     AR       new right matrix view
IN      A0       old left matrix view
IN      A1       old middle matrix view
IN      A2       old right matrix view
IN      side     matrix specified to make

----------------------------------------------------------------------------*/
{
   if ( side == PLA_LEFT )
   {
      AL->m      = A0.m;
      AL->n      = A0.n + A1.n;
      AL->offm   = A0.offm;
      AL->offn   = A0.offn;
      AL->base   = A0.base;
      
      AR->m      = A2.m;
      AR->n      = A2.n;
      AR->offm   = A2.offm;
      AR->offn   = A2.offn;
      AR->base   = A2.base;
   }
   else
   {
      AL->m      = A0.m;
      AL->n      = A0.n;
      AL->offm   = A0.offm;
      AL->offn   = A0.offn;
      AL->base   = A0.base;
      
      AR->m      = A1.m;
      AR->n      = A1.n + A2.n;
      AR->offm   = A1.offm;
      AR->offn   = A1.offn;
      AR->base   = A1.base;
   }
   
   return PLA_SUCCESS;
}
