/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Local_asum( PLA_Obj x, 
			  PLA_Obj asum_x )
/*----------------------------------------------------------------------------

   PLA_Local_asum

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_asum_check( x, asum_x );
   
   if ( PLA_Obj_is_owner( x ) && 
	PLA_Obj_is_owner( asum_x ) )
      FLA_Asum( x, asum_x );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_axpy( PLA_Obj alpha, 
			  PLA_Obj A, 
			  PLA_Obj B )
/*----------------------------------------------------------------------------

   PLA_Local_axpy

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_axpy_check( alpha, A, B );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B ) )
      FLA_Axpy( alpha, A, B );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_axpy_x( PLA_Trans trans,
			    PLA_Obj   alpha, 
			    PLA_Obj   A, 
			    PLA_Obj   B )
/*----------------------------------------------------------------------------

   PLA_Local_axpy_x

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_axpy_x_check( trans, alpha, A, B );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B ) )
      FLA_Axpy_t( trans, alpha, A, B );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_axpy_y( PLA_Trans trans,
			    PLA_Obj   alpha, 
			    PLA_Obj   A, 
			    PLA_Obj   beta, 
			    PLA_Obj   B )
/*----------------------------------------------------------------------------

   PLA_Local_axpy_y

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_axpy_y_check( trans, alpha, A, beta, B );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( B ) )
      FLA_Axpy_y( trans, alpha, A, beta, B );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_copy( PLA_Obj A,
			  PLA_Obj B )
/*----------------------------------------------------------------------------

   PLA_Local_copy

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_copy_check( A, B );
   
   if ( PLA_Obj_is_owner( A ) &&
	PLA_Obj_is_owner( B ) )
      FLA_Copy( A, B );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_copy_x( PLA_Trans trans,
			    PLA_Obj   A,
			    PLA_Obj   B )
/*----------------------------------------------------------------------------

   PLA_Local_copy_x

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_copy_x_check( trans, A, B );
   
   if ( PLA_Obj_is_owner( A ) &&
	PLA_Obj_is_owner( B ) )
      FLA_Copy_t( trans, A, B );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_dot( PLA_Obj x,
			 PLA_Obj y,
			 PLA_Obj rho )
/*----------------------------------------------------------------------------

   PLA_Local_dot

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_dot_check( x, y, rho );
   
   if ( PLA_Obj_is_owner( x ) &&
	PLA_Obj_is_owner( y ) &&
	PLA_Obj_is_owner( rho ) )
      FLA_Dot( x, y, rho );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_dot_x( PLA_Obj alpha,
			   PLA_Obj x,
			   PLA_Obj y,
			   PLA_Obj beta,
			   PLA_Obj rho )
/*----------------------------------------------------------------------------

   PLA_Local_dot_x

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_dot_x_check( alpha, x, y, beta, rho );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( y )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( rho ) )
      FLA_Dot_s( alpha, x, y, beta, rho );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_dotc( PLA_Conj conj,
			  PLA_Obj x,
			  PLA_Obj y,
			  PLA_Obj rho )
/*----------------------------------------------------------------------------

   PLA_Local_dotc

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_dotc_check( conj, x, y, rho );
   
   if ( PLA_Obj_is_owner( x ) &&
	PLA_Obj_is_owner( y ) &&
	PLA_Obj_is_owner( rho ) )
      FLA_Dot_c( conj, x, y, rho );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_iamax( PLA_Obj x,
			   PLA_Obj index )
/*----------------------------------------------------------------------------

   PLA_Local_iamax

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_iamax_check( x, index );
   
   if ( PLA_Obj_is_owner( x ) &&
	PLA_Obj_is_owner( index ))
      FLA_Iamax( x, index );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_inv_scal( PLA_Obj alpha,
			      PLA_Obj A )
/*----------------------------------------------------------------------------

   PLA_Local_inv_scal

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_inv_scal_check( alpha, A );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A ) )
      FLA_Inv_scal( alpha, A );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_nrm2( PLA_Obj x,
			  PLA_Obj nrm_x )
/*----------------------------------------------------------------------------

   PLA_Local_nrm2

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_nrm2_check( x, nrm_x );
   
   if ( PLA_Obj_is_owner( x ) &&
	PLA_Obj_is_owner( nrm_x ) )
      FLA_Nrm2( x, nrm_x );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_scal( PLA_Obj alpha,
			  PLA_Obj A )
/*----------------------------------------------------------------------------

   PLA_Local_scal

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_scal_check( alpha, A );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A ) )
      FLA_Scal( alpha, A );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_scal_x( PLA_Uplo uplo,
			    PLA_Obj  alpha,
			    PLA_Obj  A )
/*----------------------------------------------------------------------------

   PLA_Local_scal_x

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_scal_x_check( uplo, alpha, A );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A ) )
      FLA_Scal_r( uplo, alpha, A );
   
   return PLA_SUCCESS; 
}


PLA_Error PLA_Local_swap( PLA_Obj A,
			  PLA_Obj B )
/*----------------------------------------------------------------------------

   PLA_Local_swap

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_swap_check( A, B );
   
   if ( PLA_Obj_is_owner( A ) &&
	PLA_Obj_is_owner( B ) )
      FLA_Swap( A, B );
   
   return PLA_SUCCESS; 
}
