/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Local_gemv( PLA_Trans transa,
			  PLA_Obj   alpha,
			  PLA_Obj   A,
			  PLA_Obj   x,
			  PLA_Obj   beta,
			  PLA_Obj   y )
/*----------------------------------------------------------------------------

   PLA_Local_gemv

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_gemv_check( transa, alpha, A, x, beta, y );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( y ) )
      FLA_Gemv( transa, alpha, A, x, beta, y );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_ger( PLA_Obj alpha,
			 PLA_Obj x,
			 PLA_Obj y,
			 PLA_Obj A )
/*----------------------------------------------------------------------------

   PLA_Local_ger

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_ger_check( alpha, x, y, A );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( y )     &&
	PLA_Obj_is_owner( A ) )
      FLA_Ger( alpha, x, y, A );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_gerc( PLA_Obj alpha,
			  PLA_Obj x,
			  PLA_Obj y,
			  PLA_Obj A )
/*----------------------------------------------------------------------------

   PLA_Local_gerc
    
----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_gerc_check( alpha, x, y, A );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( y )     &&
	PLA_Obj_is_owner( A ) )
      FLA_Ger_c( FLA_NO_CONJUGATE, FLA_CONJUGATE, alpha, x, y, A );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_hemv( PLA_Uplo uplo,
			  PLA_Obj  alpha,
			  PLA_Obj  A,
			  PLA_Obj  x,
			  PLA_Obj  beta,
			  PLA_Obj  y )
/*----------------------------------------------------------------------------

   PLA_Local_hemv

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_hemv_check( uplo, alpha, A, x, beta, y );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( y ) )
      FLA_Hemv( uplo, alpha, A, x, beta, y );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_her( PLA_Uplo uplo,
			 PLA_Obj  alpha,
			 PLA_Obj  x,
			 PLA_Obj  A )
/*----------------------------------------------------------------------------

   PLA_Local_her

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_her_check( uplo, alpha, x, A );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( A ) )
      FLA_Her( uplo, alpha, x, A );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_her2( PLA_Uplo uplo,
			  PLA_Obj  alpha,
			  PLA_Obj  x,
			  PLA_Obj  y,
			  PLA_Obj  A )
/*----------------------------------------------------------------------------

   PLA_Local_her2

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_her2_check( uplo, alpha, x, y, A );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( y )     &&
	PLA_Obj_is_owner( A ) )
      FLA_Her2( uplo, alpha, x, y, A );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_symv( PLA_Uplo uplo,
			  PLA_Obj  alpha,
			  PLA_Obj  A,
			  PLA_Obj  x,
			  PLA_Obj  beta,
			  PLA_Obj  y )
/*----------------------------------------------------------------------------

   PLA_Local_symv

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_symv_check( uplo, alpha, A, x, beta, y );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( y ) )
      FLA_Symv( uplo, alpha, A, x, beta, y );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_syr( PLA_Uplo uplo,
			 PLA_Obj  alpha,
			 PLA_Obj  x,
			 PLA_Obj  A )
/*----------------------------------------------------------------------------

   PLA_Local_syr

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_syr_check( uplo, alpha, x, A );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( A ) )
      FLA_Syr( uplo, alpha, x, A );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_syr2( PLA_Uplo uplo,
			  PLA_Obj  alpha,
			  PLA_Obj  x,
			  PLA_Obj  y,
			  PLA_Obj  A )
/*----------------------------------------------------------------------------

   PLA_Local_syr2

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_syr2_check( uplo, alpha, x, y, A );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( y )     &&
	PLA_Obj_is_owner( A ) )
      FLA_Syr2( uplo, alpha, x, y, A );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_trmv( PLA_Uplo  uplo,
			  PLA_Trans transa,
			  PLA_Diag  diag,
			  PLA_Obj   A,
			  PLA_Obj   x )
/*----------------------------------------------------------------------------

   PLA_Local_trmv

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_trmv_check( uplo, transa, diag, A, x );
   
   if ( PLA_Obj_is_owner( A ) &&
	PLA_Obj_is_owner( x ) )
      FLA_Trmv( uplo, transa, diag, A, x );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_trmv_sx( PLA_Uplo  uplo,
			    PLA_Trans transa,
			    PLA_Diag  diag,
			    PLA_Obj   alpha,
			    PLA_Obj   A,
			    PLA_Obj   x,
			    PLA_Obj   beta,
			    PLA_Obj   y )
/*----------------------------------------------------------------------------

   PLA_Local_trmv_sx

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_trmv_sx_check( uplo, transa, diag, alpha, A, x, beta, y );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( y ) )
      FLA_Trmv_sx( uplo, transa, diag, alpha, A, x, beta, y );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_trsv( PLA_Uplo  uplo,
			  PLA_Trans transa,
			  PLA_Diag  diag,
			  PLA_Obj   A,
			  PLA_Obj   x )
/*----------------------------------------------------------------------------

   PLA_Local_trsv

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_trsv_check( uplo, transa, diag, A, x );
   
   if ( PLA_Obj_is_owner( A ) &&
	PLA_Obj_is_owner( x ) )
      FLA_Trsv( uplo, transa, diag, A, x );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_trsv_sx( PLA_Uplo  uplo,
			    PLA_Trans transa,
			    PLA_Diag  diag,
			    PLA_Obj   alpha,
			    PLA_Obj   A,
			    PLA_Obj   x,
			    PLA_Obj   beta,
			    PLA_Obj   y )
/*----------------------------------------------------------------------------

   PLA_Local_trsv_sx

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_trsv_sx_check( uplo, transa, diag, alpha, A, x, beta, y );
   
   if ( PLA_Obj_is_owner( alpha ) && 
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( x )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( y ) )
      FLA_Trsv_sx( uplo, transa, diag, alpha, A, x, beta, y );
   
   return PLA_SUCCESS;
}
