/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Local_gemm( PLA_Trans transa,
			  PLA_Trans transb,
			  PLA_Obj   alpha,
			  PLA_Obj   A,
			  PLA_Obj   B,
			  PLA_Obj   beta,
			  PLA_Obj   C )
/*----------------------------------------------------------------------------

   PLA_Local_gemm

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_gemm_check( transa, transb, alpha, A, B, beta, C );  
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( C ) )
      FLA_Gemm( transa, transb, alpha, A, B, beta, C );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_hemm( PLA_Side side,
			  PLA_Uplo uplo,
			  PLA_Obj  alpha,
			  PLA_Obj  A,
			  PLA_Obj  B,
			  PLA_Obj  beta,
			  PLA_Obj  C )
/*----------------------------------------------------------------------------

   PLA_Local_hemm

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_hemm_check( side, uplo, alpha, A, B, beta, C );  
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( C ) )
      FLA_Hemm( side, uplo, alpha, A, B, beta, C );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_her2k( PLA_Uplo  uplo,
			   PLA_Trans trans,
			   PLA_Obj   alpha,
			   PLA_Obj   A,
			   PLA_Obj   B,
			   PLA_Obj   beta,
			   PLA_Obj   C )
/*----------------------------------------------------------------------------

   PLA_Local_her2k

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_her2k_check( uplo, trans, alpha, A, B, beta, C );  
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( C ) )
      FLA_Her2k( uplo, trans, alpha, A, B, beta, C );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_herk( PLA_Uplo  uplo,
			  PLA_Trans trans,
			  PLA_Obj   alpha,
			  PLA_Obj   A,
			  PLA_Obj   beta,
			  PLA_Obj   C )
/*----------------------------------------------------------------------------

   PLA_Local_herk

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_herk_check( uplo, trans, alpha, A, beta, C );  
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( C ) )
      FLA_Herk( uplo, trans, alpha, A, beta, C );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_symm( PLA_Side side,
			  PLA_Uplo uplo,
			  PLA_Obj  alpha,
			  PLA_Obj  A,
			  PLA_Obj  B,
			  PLA_Obj  beta,
			  PLA_Obj  C )
/*----------------------------------------------------------------------------

   PLA_Local_symm

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_symm_check( side, uplo, alpha, A, B, beta, C );  
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( C ) )
      FLA_Symm( side, uplo, alpha, A, B, beta, C );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_syr2k( PLA_Uplo  uplo,
			   PLA_Trans trans,
			   PLA_Obj   alpha,
			   PLA_Obj   A,
			   PLA_Obj   B,
			   PLA_Obj   beta,
			   PLA_Obj   C )
/*----------------------------------------------------------------------------

   PLA_Local_syr2k

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_syr2k_check( uplo, trans, alpha, A, B, beta, C );  
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( C ) )
      FLA_Syr2k( uplo, trans, alpha, A, B, beta, C );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_syrk( PLA_Uplo  uplo,
			  PLA_Trans trans,
			  PLA_Obj   alpha,
			  PLA_Obj   A,
			  PLA_Obj   beta,
			  PLA_Obj   C )
/*----------------------------------------------------------------------------

   PLA_Local_syrk

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_syrk_check( uplo, trans, alpha, A, beta, C );  
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( C ) )
      FLA_Syrk( uplo, trans, alpha, A, beta, C );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_trmm( PLA_Side  side,
			  PLA_Uplo  uplo,
			  PLA_Trans trans,
			  PLA_Diag  diag,
			  PLA_Obj   alpha,
			  PLA_Obj   A,
			  PLA_Obj   B )
/*----------------------------------------------------------------------------

   PLA_Local_trmm

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_trmm_check( side, uplo, trans, diag, alpha, A, B );  
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B ) )
      FLA_Trmm( side, uplo, trans, diag, alpha, A, B );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_trmm_sx( PLA_Side  side,
			    PLA_Uplo  uplo,
			    PLA_Trans trans,
			    PLA_Diag  diag,
			    PLA_Obj   alpha,
			    PLA_Obj   A,
			    PLA_Obj   B,
			    PLA_Obj   beta,
			    PLA_Obj   C )
/*----------------------------------------------------------------------------

   PLA_Local_trmm_sx

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_trmm_sx_check( side, uplo, trans, diag, alpha, A, B, beta, C );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( C ) )
      FLA_Trmm_sx( side, uplo, trans, diag, alpha, A, B, beta, C );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_trsm( PLA_Side  side,
			  PLA_Uplo  uplo,
			  PLA_Trans trans,
			  PLA_Diag  diag,
			  PLA_Obj   alpha,
			  PLA_Obj   A,
			  PLA_Obj   B )
/*----------------------------------------------------------------------------

   PLA_Local_trsm

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_trsm_check( side, uplo, trans, diag, alpha, A, B );  
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B ) )
      FLA_Trsm( side, uplo, trans, diag, alpha, A, B );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Local_trsm_sx( PLA_Side  side,
			    PLA_Uplo  uplo,
			    PLA_Trans trans,
			    PLA_Diag  diag,
			    PLA_Obj   alpha,
			    PLA_Obj   A,
			    PLA_Obj   B,
			    PLA_Obj   beta,
			    PLA_Obj   C )
/*----------------------------------------------------------------------------

   PLA_Local_trsm_sx

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Local_trsm_sx_check( side, uplo, trans, diag, alpha, A, B, beta, C );
   
   if ( PLA_Obj_is_owner( alpha ) &&
	PLA_Obj_is_owner( A )     &&
	PLA_Obj_is_owner( B )     &&
	PLA_Obj_is_owner( beta )  &&
	PLA_Obj_is_owner( C ) )
      FLA_Trsm_sx( side, uplo, trans, diag, alpha, A, B, beta, C );  
   
   return PLA_SUCCESS;
}
