*
*  libflame
*  An object-based infrastructure for developing high-performance
*  dense linear algebra libraries.
*
*  Copyright (C) 2008, The University of Texas
*
*  libflame is free software; you can redistribute it and/or modify
*  it under the terms of the GNU Lesser General Public License as
*  published by the Free Software Foundation; either version 2.1 of
*  the License, or (at your option) any later version.
*
*  libflame is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
*  Lesser General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public
*  License along with libflame; if you did not receive a copy, see
*  http://www.gnu.org/licenses/.
*
*  For more information, please contact us at flame@cs.utexas.edu or
*  send mail to:
*
*  Field G. Van Zee and/or
*  Robert A. van de Geijn
*  The University of Texas at Austin
*  Department of Computer Sciences
*  1 University Station C0500
*  Austin TX 78712
*
      SUBROUTINE DGEMM ( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB,
     $                   BETA, C, LDC )
*     .. Scalar Arguments ..
      CHARACTER*1        TRANSA, TRANSB
      INTEGER            M, N, K, LDA, LDB, LDC
      DOUBLE PRECISION   ALPHA, BETA
*     .. Array Arguments ..
      DOUBLE PRECISION   A( LDA, * ), B( LDB, * ), C( LDC, * )
*     ..
*
*  Purpose
*  =======
*
*  DGEMM  performs one of the matrix-matrix operations
*
*     C := alpha*op( A )*op( B ) + beta*C,
*
*  where  op( X ) is one of
*
*     op( X ) = X   or   op( X ) = X',
*
*  alpha and beta are scalars, and A, B and C are matrices, with op( A )
*  an m by k matrix,  op( B )  a  k by n matrix and  C an m by n matrix.
*
*  Parameters
*  ==========
*
*  TRANSA - CHARACTER*1.
*           On entry, TRANSA specifies the form of op( A ) to be used in
*           the matrix multiplication as follows:
*
*              TRANSA = 'N' or 'n',  op( A ) = A.
*
*              TRANSA = 'T' or 't',  op( A ) = A'.
*
*              TRANSA = 'C' or 'c',  op( A ) = A'.
*
*           Unchanged on exit.
*
*  TRANSB - CHARACTER*1.
*           On entry, TRANSB specifies the form of op( B ) to be used in
*           the matrix multiplication as follows:
*
*              TRANSB = 'N' or 'n',  op( B ) = B.
*
*              TRANSB = 'T' or 't',  op( B ) = B'.
*
*              TRANSB = 'C' or 'c',  op( B ) = B'.
*
*           Unchanged on exit.
*
*  M      - INTEGER.
*           On entry,  M  specifies  the number  of rows  of the  matrix
*           op( A )  and of the  matrix  C.  M  must  be at least  zero.
*           Unchanged on exit.
*
*  N      - INTEGER.
*           On entry,  N  specifies the number  of columns of the matrix
*           op( B ) and the number of columns of the matrix C. N must be
*           at least zero.
*           Unchanged on exit.
*
*  K      - INTEGER.
*           On entry,  K  specifies  the number of columns of the matrix
*           op( A ) and the number of rows of the matrix op( B ). K must
*           be at least  zero.
*           Unchanged on exit.
*
*  ALPHA  - DOUBLE PRECISION.
*           On entry, ALPHA specifies the scalar alpha.
*           Unchanged on exit.
*
*  A      - DOUBLE PRECISION array of DIMENSION ( LDA, ka ), where ka is
*           k  when  TRANSA = 'N' or 'n',  and is  m  otherwise.
*           Before entry with  TRANSA = 'N' or 'n',  the leading  m by k
*           part of the array  A  must contain the matrix  A,  otherwise
*           the leading  k by m  part of the array  A  must contain  the
*           matrix A.
*           Unchanged on exit.
*
*  LDA    - INTEGER.
*           On entry, LDA specifies the first dimension of A as declared
*           in the calling (sub) program. When  TRANSA = 'N' or 'n' then
*           LDA must be at least  max( 1, m ), otherwise  LDA must be at
*           least  max( 1, k ).
*           Unchanged on exit.
*
*  B      - DOUBLE PRECISION array of DIMENSION ( LDB, kb ), where kb is
*           n  when  TRANSB = 'N' or 'n',  and is  k  otherwise.
*           Before entry with  TRANSB = 'N' or 'n',  the leading  k by n
*           part of the array  B  must contain the matrix  B,  otherwise
*           the leading  n by k  part of the array  B  must contain  the
*           matrix B.
*           Unchanged on exit.
*
*  LDB    - INTEGER.
*           On entry, LDB specifies the first dimension of B as declared
*           in the calling (sub) program. When  TRANSB = 'N' or 'n' then
*           LDB must be at least  max( 1, k ), otherwise  LDB must be at
*           least  max( 1, n ).
*           Unchanged on exit.
*
*  BETA   - DOUBLE PRECISION.
*           On entry,  BETA  specifies the scalar  beta.  When  BETA  is
*           supplied as zero then C need not be set on input.
*           Unchanged on exit.
*
*  C      - DOUBLE PRECISION array of DIMENSION ( LDC, n ).
*           Before entry, the leading  m by n  part of the array  C must
*           contain the matrix  C,  except when  beta  is zero, in which
*           case C need not be set on entry.
*           On exit, the array  C  is overwritten by the  m by n  matrix
*           ( alpha*op( A )*op( B ) + beta*C ).
*
*  LDC    - INTEGER.
*           On entry, LDC specifies the first dimension of C as declared
*           in  the  calling  (sub)  program.   LDC  must  be  at  least
*           max( 1, m ).
*           Unchanged on exit.
*
*
*  Level 3 Blas routine.
*
*  -- Written on 8-February-1989.
*     Jack Dongarra, Argonne National Laboratory.
*     Iain Duff, AERE Harwell.
*     Jeremy Du Croz, Numerical Algorithms Group Ltd.
*     Sven Hammarling, Numerical Algorithms Group Ltd.
*
*
*     .. External Functions ..
      LOGICAL            LSAME
      EXTERNAL           LSAME
*     .. External Subroutines ..
      EXTERNAL           XERBLA
*     .. Intrinsic Functions ..
      INTRINSIC          MAX
*     .. Local Scalars ..
      LOGICAL            NOTA, NOTB
      INTEGER            I, INFO, J, NCOLA, NROWA, NROWB, NCOLB
*     .. Parameters ..
      DOUBLE PRECISION   ONE         , ZERO
      PARAMETER        ( ONE = 1.0D+0, ZERO = 0.0D+0 )
*     ..       
*     .. Declare FLAME related stuff
      include "FLAMEf.h"
*
      INTEGER            A_obj( FLA_OBJ_SIZE ),
     $                   B_obj( FLA_OBJ_SIZE ),
     $                   C_obj( FLA_OBJ_SIZE ),
     $                   alpha_obj( FLA_OBJ_SIZE ), 
     $                   beta_obj( FLA_OBJ_SIZE ),
     $                   itransa,
     $                   itransb,
     $                   datatype
      PARAMETER        ( datatype = FLA_DOUBLE )
*     ..
*     .. Executable Statements ..
*
*     Initialize FLAME.
      CALL FLA_INIT_F( INFO )
*
*
*     Set  NOTA  and  NOTB  as  true if  A  and  B  respectively are not
*     transposed and set  NROWA, NCOLA and  NROWB  as the number of rows
*     and  columns of  A  and the  number of  rows  of  B  respectively.
*
      NOTA  = LSAME( TRANSA, 'N' )
      NOTB  = LSAME( TRANSB, 'N' )
      IF( NOTA )THEN
         NROWA = M
         NCOLA = K
      ELSE
         NROWA = K
         NCOLA = M
      END IF
      IF( NOTB )THEN
         NROWB = K
         NCOLB = N
      ELSE
         NROWB = N
         NCOLB = K
      END IF
*
*     Test the input parameters.
*
      INFO = 0
      IF(      ( .NOT.NOTA                 ).AND.
     $         ( .NOT.LSAME( TRANSA, 'C' ) ).AND.
     $         ( .NOT.LSAME( TRANSA, 'T' ) )      )THEN
         INFO = 1
      ELSE IF( ( .NOT.NOTB                 ).AND.
     $         ( .NOT.LSAME( TRANSB, 'C' ) ).AND.
     $         ( .NOT.LSAME( TRANSB, 'T' ) )      )THEN
         INFO = 2
      ELSE IF( M  .LT.0               )THEN
         INFO = 3
      ELSE IF( N  .LT.0               )THEN
         INFO = 4
      ELSE IF( K  .LT.0               )THEN
         INFO = 5
      ELSE IF( LDA.LT.MAX( 1, NROWA ) )THEN
         INFO = 8
      ELSE IF( LDB.LT.MAX( 1, NROWB ) )THEN
         INFO = 10
      ELSE IF( LDC.LT.MAX( 1, M     ) )THEN
         INFO = 13
      END IF
      IF( INFO.NE.0 )THEN
         CALL XERBLA( 'DGEMM ', INFO )
         RETURN
      END IF
*
*     Quick return if possible.
*
      IF( ( M.EQ.0 ).OR.( N.EQ.0 ).OR.
     $    ( ( ( ALPHA.EQ.ZERO ).OR.( K.EQ.0 ) ).AND.( BETA.EQ.ONE ) ) )
     $   RETURN
*
*     And if  alpha.eq.zero.
*
      IF( ALPHA.EQ.ZERO )THEN
         IF( BETA.EQ.ZERO )THEN
            DO 20, J = 1, N
               DO 10, I = 1, M
                  C( I, J ) = ZERO
   10          CONTINUE
   20       CONTINUE
         ELSE
            DO 40, J = 1, N
               DO 30, I = 1, M
                  C( I, J ) = BETA*C( I, J )
   30          CONTINUE
   40       CONTINUE
         END IF
         RETURN
      END IF
*
*     Set FLAME transposition argument for A.
*
      IF     ( LSAME( TRANSA, 'N' ) )THEN
         itransa = FLA_NO_TRANSPOSE
      ELSE IF( LSAME( TRANSA, 'T' ) )THEN
         itransa = FLA_TRANSPOSE
      ELSE IF( LSAME( TRANSA, 'C' ) )THEN
         itransa = FLA_CONJ_TRANSPOSE
      END IF
*
*     Set FLAME transposition argument for B.
*
      IF     ( LSAME( TRANSB, 'N' ) )THEN
         itransb = FLA_NO_TRANSPOSE
      ELSE IF( LSAME( TRANSB, 'T' ) )THEN
         itransb = FLA_TRANSPOSE
      ELSE IF( LSAME( TRANSB, 'C' ) )THEN
         itransb = FLA_CONJ_TRANSPOSE
      END IF
*
*     Create base object for A_obj and attach matrix buffer.
*
      CALL FLA_OBJ_CREATE_W_OUT_BUFF_F( datatype, NROWA, NCOLA, A_obj,
     $                                  INFO )
      CALL FLA_OBJ_ATTACH_BUFF_F( A, LDA, A_obj, INFO )
*
*     Create base object for B_obj and attach matrix buffer.
*
      CALL FLA_OBJ_CREATE_W_OUT_BUFF_F( datatype, NROWB, NCOLB, B_obj,
     $                                  INFO )
      CALL FLA_OBJ_ATTACH_BUFF_F( B, LDB, B_obj, INFO )
*
*     Create base object for C_obj and attach matrix buffer.
*
      CALL FLA_OBJ_CREATE_W_OUT_BUFF_F( datatype, M, N, C_obj, 
     $                                  INFO )
      CALL FLA_OBJ_ATTACH_BUFF_F( C, LDC, C_obj, INFO )
*
*     Create base object for alpha_obj and attach scalar buffer.
*
      CALL FLA_OBJ_CREATE_W_OUT_BUFF_F( datatype, 1, 1, alpha_obj, 
     $                                  INFO )
      CALL FLA_OBJ_ATTACH_BUFF_F( ALPHA, 1, alpha_obj, INFO )
*
*     Create base object for beta_obj and attach scalar buffer.
*
      CALL FLA_OBJ_CREATE_W_OUT_BUFF_F( datatype, 1, 1, beta_obj,
     $                                  INFO )
      CALL FLA_OBJ_ATTACH_BUFF_F( BETA, 1, beta_obj, INFO )
*
*     Call FLAME gemm code.
*
      CALL FLA_GEMM_F( itransa, itransb, alpha, A, B, beta, C, INFO )
*
*     Free A_obj base object (but leave A buffer alone).
*
      CALL FLA_OBJ_FREE_W_OUT_BUFF_F( A_obj, INFO )
*
*     Free B_obj base object (but leave B buffer alone).
*
      CALL FLA_OBJ_FREE_W_OUT_BUFF_F( B_obj, INFO )
*
*     Free C_obj base object (but leave C buffer alone).
*
      CALL FLA_OBJ_FREE_W_OUT_BUFF_F( C_obj, INFO )
*
*     Free alpha_obj base object (but leave alpha buffer alone).
*
      CALL FLA_OBJ_FREE_W_OUT_BUFF_F( alpha_obj, INFO )
*
*     Free beta_obj base object (but leave beta buffer alone).
*
      CALL FLA_OBJ_FREE_W_OUT_BUFF_F( beta_obj, INFO )
*
      RETURN
*
*     End of DGEMM .
*
      END
