/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

extern fla_syrk_t*      fla_syrk_cntl_ip;
extern fla_syrk_t*      fla_syrk_cntl_op;
extern fla_syrk_t*      fla_syrk_cntl_sq;
extern fla_blocksize_t* fla_syrk_var2_bsize;
extern fla_blocksize_t* fla_syrk_var5_bsize;

FLA_Error FLA_Syrk( FLA_Uplo uplo, FLA_Trans trans, FLA_Obj alpha, FLA_Obj A, FLA_Obj beta, FLA_Obj C )
{
  FLA_Error    r_val = FLA_SUCCESS;
#ifdef FLA_ENABLE_BLAS3_FRONT_END_CNTL_TREES
  FLA_Datatype datatype;
  int          m_C, k_A;
  int          FLA_SYRK_VAR2_BLOCKSIZE;
  int          FLA_SYRK_VAR5_BLOCKSIZE;
#endif

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Bool     check_params;

  // Check parameters.
  if ( FLA_Check_parameters() )
    FLA_Syrk_check( uplo, trans, alpha, A, beta, C );

  // Temporarily disable parameter checking.
  check_params = FLA_Check_parameters_set( FALSE );
#endif
  
#ifdef FLA_ENABLE_BLAS3_FRONT_END_CNTL_TREES
  // Determine the datatype of the operation.
  datatype = FLA_Obj_datatype( A );

  // Extract the appropriate blocksize for the given datatype.
  FLA_SYRK_VAR2_BLOCKSIZE = FLA_Blocksize_extract( datatype, fla_syrk_var2_bsize );
  FLA_SYRK_VAR5_BLOCKSIZE = FLA_Blocksize_extract( datatype, fla_syrk_var5_bsize );

  // Determine the dimension of C.
  m_C = FLA_Obj_length( C );

  // Determine the k dimension of A after transposition.
  if ( trans == FLA_NO_TRANSPOSE )
    k_A = FLA_Obj_width( A );
  else
    k_A = FLA_Obj_length( A );

  // Invoke FLA_Syrk_internal() with the appropriate control tree.
  if      ( FLA_SYRK_VAR2_BLOCKSIZE < m_C && FLA_SYRK_VAR5_BLOCKSIZE < k_A )
  {
    r_val = FLA_Syrk_internal( uplo, trans, alpha, A, beta, C, fla_syrk_cntl_sq );
  }
  else if ( FLA_SYRK_VAR2_BLOCKSIZE < m_C && k_A <= FLA_SYRK_VAR5_BLOCKSIZE )
  {
    r_val = FLA_Syrk_internal( uplo, trans, alpha, A, beta, C, fla_syrk_cntl_op );
  }
  else if ( m_C <= FLA_SYRK_VAR2_BLOCKSIZE && FLA_SYRK_VAR5_BLOCKSIZE < k_A )
  {
    r_val = FLA_Syrk_internal( uplo, trans, alpha, A, beta, C, fla_syrk_cntl_ip );
  }
  else if ( m_C <= FLA_SYRK_VAR2_BLOCKSIZE && k_A <= FLA_SYRK_VAR5_BLOCKSIZE )
  {
    r_val = FLA_Syrk_external( uplo, trans, alpha, A, beta, C );
  }
#else
  //r_val = FLA_Syrk_internal( uplo, trans, alpha, A, beta, C, fla_syrk_cntl_op );
  r_val = FLA_Syrk_external( uplo, trans, alpha, A, beta, C );
#endif
  
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  // Restore parameter checking to its previous status.
  FLA_Check_parameters_set( check_params );
#endif
  
  return r_val;
}

void FLA_F2C( fla_syrk_f ) 
     ( F_INT *uplo, F_INT *trans, F_INT *alpha, F_INT *A, F_INT *beta, F_INT *C, F_INT *IERROR )
{
  *IERROR = FLA_Syrk( *( ( FLA_Uplo  * ) uplo  ),
                      *( ( FLA_Trans * ) trans ),
                      *( ( FLA_Obj   * ) alpha ),
                      *( ( FLA_Obj   * ) A     ),
                      *( ( FLA_Obj   * ) beta  ),
                      *( ( FLA_Obj   * ) C     ) );
}

