/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_QR_Accum_S_unb_var2( FLA_Obj A, FLA_Obj S )
{
  FLA_Obj ATL,   ATR,      A00,  a01,     A02, 
          ABL,   ABR,      a10t, alpha11, a12t,
                           A20,  a21,     A22;

  FLA_Obj STL,   STR,      S00,  s01,     S02, 
          SBL,   SBR,      s10t, sigma11, s12t,
                           S20,  s21,     S22;

  FLA_Obj w01;

  FLA_Part_2x2( A,    &ATL, &ATR,
                      &ABL, &ABR,     0, 0, FLA_TL );

  FLA_Part_2x2( S,    &STL, &STR,
                      &SBL, &SBR,     0, 0, FLA_TL );

  while ( FLA_Obj_width ( ATL ) != FLA_Obj_width ( A )   &&
	  FLA_Obj_length( ATL ) != FLA_Obj_length( A )){

    FLA_Repart_2x2_to_3x3( ATL, /**/ ATR,       &A00,  /**/ &a01,     &A02,
                        /* ************* */   /* ************************** */
                                                &a10t, /**/ &alpha11, &a12t,
                           ABL, /**/ ABR,       &A20,  /**/ &a21,     &A22,
                           1, 1, FLA_BR );

    FLA_Repart_2x2_to_3x3( STL, /**/ STR,       &S00,  /**/ &s01,     &S02,
                        /* ************* */   /* ************************** */
                                                &s10t, /**/ &sigma11, &s12t,
                           SBL, /**/ SBR,       &S20,  /**/ &s21,     &S22,
                           1, 1, FLA_BR );

    /*------------------------------------------------------------*/

    /* Update the current column by applying ( I - U S' U' ) to it
    where U and S are the parts of % these matrices that have been
    accumulated so far.
    
    w01 = - triu( S00 )' * ( trilu( A00 )' * a01 + alpha11 * a10t' + A20' * a21 );  
    
    a01     = a01     + trilu( A00 ) * w01;
    alpha11 = alpha11 + a10t         * w01;
    a21     = a21     + A20          * w01;
    */

    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, a01, &w01 );
    FLA_Copy( a01, w01 );
    FLA_Trmv( FLA_LOWER_TRIANGULAR, FLA_TRANSPOSE, FLA_UNIT_DIAG, 
	      A00, w01 );
    FLA_Axpy_t( FLA_TRANSPOSE, alpha11, a10t, w01 );
    FLA_Gemv( FLA_TRANSPOSE, FLA_ONE, A20, a21, FLA_ONE, w01 );
    FLA_Trmv( FLA_UPPER_TRIANGULAR, FLA_TRANSPOSE, FLA_NONUNIT_DIAG,
	      S00, w01 );
    FLA_Scal( FLA_MINUS_ONE, w01 );
    FLA_Dot_s( FLA_ONE, a10t, w01, FLA_ONE, alpha11 );
    FLA_Gemv( FLA_NO_TRANSPOSE, FLA_ONE, A20, w01, FLA_ONE, a21 );
    FLA_Trmv( FLA_LOWER_TRIANGULAR, FLA_NO_TRANSPOSE, FLA_UNIT_DIAG,
	      A00, w01 );
    FLA_Axpy( FLA_ONE, w01, a01 );
    FLA_Obj_free( &w01 );

    /* Compute the current Householder vector 

    [ alpha11, ...
      a21, sigma11 ] = Househ2( alpha11, ...
                                a21 );
    */

    FLA_Househ2( alpha11,
                 a21, sigma11 );

    /* Update S
    s01  = - triu( S00 ) * ( a10t' + A20' * a21 ) * sigma11;
    */

    FLA_Copy_t( FLA_TRANSPOSE, a10t, s01 );
    FLA_Gemv( FLA_TRANSPOSE, FLA_ONE, A20, a21, FLA_ONE, s01 );
    FLA_Trmv( FLA_UPPER_TRIANGULAR, FLA_NO_TRANSPOSE, FLA_NONUNIT_DIAG,
	      S00, s01 );
    FLA_Scal( sigma11, s01 );
    FLA_Scal( FLA_MINUS_ONE, s01 );

    /*------------------------------------------------------------*/

    FLA_Cont_with_3x3_to_2x2( &ATL, /**/ &ATR,       A00,  a01,     /**/ A02,
                                                     a10t, alpha11, /**/ a12t,
                            /* ************** */  /* ************************ */
                              &ABL, /**/ &ABR,       A20,  a21,     /**/ A22,
                              FLA_TL );

    FLA_Cont_with_3x3_to_2x2( &STL, /**/ &STR,       S00,  s01,     /**/ S02,
                                                     s10t, sigma11, /**/ s12t,
                            /* ************** */  /* ************************ */
                              &SBL, /**/ &SBR,       S20,  s21,     /**/ S22,
                              FLA_TL );

  }

  return FLA_SUCCESS;
}



